(require 'persia)
(require 'headhand)

(ps:dlopen "collision.so")     ; enable collision detection
(ps:dlopen "mouse_down.so")    ; mouse-button routines
(ps:dlopen "frame_capture.so") ; ps:Delete function

;;; world origin is on the floor under the flock transmitter
;;; with positive-y axis towards curved wall of lab
;;; with positive-x axis towards master-box, 
;;; with positive-z axis up

;;; board origin is bottom left-hand corner
;;; units are in meters

(define doing-vr #t)

;;; temp funcs
(define m-display
  (lambda args
    (for-each display args)))

(define disp-pfmatrix
  (lambda (mat)
    (letrec ((helper
	      (lambda (num)
		(if (= num 16)
		    (newline)
		    (begin
		      (display (ps:pfmatrix-ref mat (quotient num 4)
						(modulo num 4)))
		      (display #\space)
		      (display
		       (ps:pfmatrix-ref mat (quotient num 4)
					(modulo (+ 1 num) 4)))
		      (display #\space)
		      (display
		       (ps:pfmatrix-ref mat (quotient num 4)
					(modulo (+ 2 num) 4)))
		      (display #\space)
		      (display
		       (ps:pfmatrix-ref mat (quotient num 4)
					(modulo (+ 3 num) 4)))
		      (newline)
		      (helper (+ num 4)))))))
      (helper 0))))


;;; temp declarations

; hand sensitivity to mouse-movement
(if (not doing-vr)
    (define mouse-sensitivity .01))

; units of meters
(define board-width 1.50)
(define board-height 1.0)
(define board-y-distance-from-transmitter .4)
(define board-z-distance-from-floor .675) ; from the floor to bottom of board
(define block-width 0.03)
(define block-height 0.02)
(define person-height 1.80)
(define room-width 3.50)
(define room-height 2.65)
(define room-depth 7.60)
(define frontwall-dist-from-transmitter 3.20)
(define eyes-dist-from-board .40)
(define eyes-dist-from-floor 1.18)
(define hand-dist-from-floor-init .28)
(define hand-x-dist-from-eyes-init .25)
(define hand-width .015)
(define hand-length (if doing-vr .087 .173))
(define hand-thickness .005)  ; estimate

; debug
(define temp-off-head-y -1.)
(define temp-off-hand-z .73)

(if doing-vr
    (ps:position-windows-for-vr-eyes 0 1)
    (begin
      (ps:position-window 0 50 750 200 200)
      (ps:position-window 1 250 750 200 200)))

(ps:init-cameras 2)           ; must come before any graphics code

;;; define some useful colors
(define red     (ps:make-gstate '(color . #(0.8 0.1 0.1))))  ; | block colors
(define green   (ps:make-gstate '(color . #(0.1 0.8 0.1))))  ; |
(define blue    (ps:make-gstate '(color . #(0.1 0.1 0.8))))  ; |
(define yellow  (ps:make-gstate '(color . #(0.8 0.8 0.1))))  ; |
(define purple  (ps:make-gstate '(color . #(0.8 0.1 0.8))))  ; |

(define white   (ps:make-gstate '(color . #(.8 .8 .8))))     ; | room colors
(define pink    (ps:make-gstate '(color . #(.8 .5 .5))))     ; |
(define brown   (ps:make-gstate '(color . #(1. .42 .28))))   ; |

(define t-peach (ps:make-gstate '(color . #(1. .6 .4 .05)))) ; | hand color



(if doing-vr
    (begin
      (ps:headhand-start)
      (ps:headhand-set-num-receivers 2)))

;;; create and place a room environment
(define left-wall               ; create the left side wall
  (ps:place-node
   (ps:make-geode-plane
    room-depth room-height (cons 'gstate pink))
   (- (/ room-width 2)) (- frontwall-dist-from-transmitter room-depth) 0.))
(pf:dcs-rot left-wall 90. 90. 0.)
(define right-wall             ; create the right side wall
  (ps:place-node
   (ps:make-geode-plane
    room-depth room-height (cons 'gstate pink))
 ;     (disp-pfmatrix tmp-mat)
  (/ room-width 2) frontwall-dist-from-transmitter 0.))
(pf:dcs-rot right-wall -90. 90. 0.)
(define back-wall              ; create the back wall
  (ps:place-node
   (ps:make-geode-plane
    room-width room-height (cons 'gstate brown))
   (/ room-width 2) (- frontwall-dist-from-transmitter room-depth) 0.
   ps:*scene-group*))
(pf:dcs-rot back-wall 180. 90. 0.)
(define front-wall            ; create the front wall
  (ps:place-node
   (ps:make-geode-plane
    room-width room-height (cons 'gstate brown))
   (- (/ room-width 2)) frontwall-dist-from-transmitter 0.
   ps:*scene-group*))
(pf:dcs-rot front-wall 0. 90. 0.)

;;; create and place the board
(define bl:*board-plane*
   (ps:make-geode-plane board-width board-height
			(cons 'gstate (ps:make-gstate '(color . #(.1 .4 .1))))))
(define board
  (ps:place-node
   bl:*board-plane*
   (- (/ board-width 2)) 
   board-y-distance-from-transmitter 
   board-z-distance-from-floor
   ps:*scene-group*))
(pf:dcs-rot board 0.0 90.0 0.0)

;;; deal with the viewpoint
(if doing-vr
    (begin
      (define head (ps:place-vr-eyes-in-head-dcs 0 1 0.04))
      (pf:add-child ps:*scene-group* ps:*head-dcs*)
      (pf:add-child ps:*scene-group* ps:*hand-dcs*))
    (begin
      (define bl:*head-dcs* (pf:new-dcs))
      (define bl:*hand-dcs* (pf:new-dcs))
      (pf:add-child ps:*scene-group* bl:*head-dcs*)
      (pf:add-child ps:*scene-group* bl:*hand-dcs*)
      (pf:dcs-trans
       bl:*head-dcs*
       0.
       (+ temp-off-head-y
	  (- board-y-distance-from-transmitter eyes-dist-from-board))
       eyes-dist-from-floor)
      (pf:dcs-trans             ; put hand into initial position
       bl:*hand-dcs*
       hand-x-dist-from-eyes-init
       (- board-y-distance-from-transmitter eyes-dist-from-board)
       (+ hand-dist-from-floor-init temp-off-hand-z))
      (pf:dcs-rot bl:*hand-dcs* 0. 90. 0.)
      (if (not doing-vr) (ps:attach-camera 1 bl:*head-dcs*))))


;;; Some useful functions for making the board layout
(define (make-geode-block gstate-color)
  (ps:make-geode-cube block-width block-width block-height 
		      (cons 'gstate gstate-color)))

(define (make-and-place-block x z gstate-color)
  (ps:place-node (make-geode-block gstate-color)
		   x z 0.0 board))


;;; Attach a box your hand (the box emulates your hand's appearance)
; bl:*hand-box* is the box geode, and bl:*hand-box-dcs* is the dcs in which it
;  resides.
(define bl:*hand-box* (ps:make-geode-cube hand-width hand-length hand-thickness
					  (cons 'gstate t-peach)))
(if doing-vr
    (define bl:*hand-box-dcs*
      (ps:place-node bl:*hand-box*
		     (- (/ hand-width 2.)) 0. 0. ps:*hand-dcs*))
    (define bl:*hand-box-dcs*
      (ps:place-node bl:*hand-box*
		     (- (/ hand-width 2.)) 0. 0. bl:*hand-dcs*)))

;;; Set up collision-detection
(ps:init-collision)                            ; every node has mask of class 0
;(ps:set-I-class board 1)                      ; the board and all children
;(ps:set-all-classes)                          ; always follow ps:set-I-class
;(ps:get-I-class board) ; debug


;;; Some board-lego handling functions ;;;

; * board->map : converts coord on board into integer coord in board-map.  Note:
;  no error-checking for now; coord may be off board.
(define (board->map real)
  (inexact->exact (truncate (/ real block-width))))

; * block-in-pos? : returns #t if there is a block recorded in board-map at pos
;  (x, y) and #f otherwise.
(define (block-in-pos? x y)
  (begin
    (m-display x #\, y #\space)
    (vector-ref (vector-ref board-map x) y)))

; * place-block : takes the boardx & boardy coordinates and locates the
;  corresponding position in the board-map grid.  If no block currently exists
;  in that position, the block, "block", is put into the position and the dcs
;  that translates the block relative to the board is returned.
(define (place-block block boardx boardy)
  (let ((x-pos (board->map boardx))
	(y-pos (board->map boardy)))
    ;; stay within the borders
    (cond
     ((> x-pos (vector-length board-map))
      (set! x-pos (vector-length board-map)))
     ((> 0 x-pos)
      (set! x-pos 0)))
    (cond
     ((> y-pos (vector-length (vector-ref board-map 0)))
      (set! y-pos (vector-ref board-map 0)))
     ((> 0 y-pos)
      (set! y-pos 0)))

    (if (not (block-in-pos? x-pos y-pos))
	(begin                                 ; place block, update board-map
	  (vector-set! (vector-ref board-map x-pos) y-pos #t)
	  (ps:place-node
	   block (* x-pos block-width) (* y-pos block-width) 0. board))
	#f)))                                  ; return false: block not placed

;;; Globals
(define blk-in-hnd #f)
(define bl:hand-pos-x hand-x-dist-from-eyes-init)
(define bl:hand-pos-y
  (- board-y-distance-from-transmitter eyes-dist-from-board))
(define bl:hand-pos-z (+ hand-dist-from-floor-init temp-off-hand-z))
(define bl:hand-pos-p 90.)
(define frame-counter 0) ; debug
(define block-pickup-start 0.)                ; time cushion for ea blockmove so
(define block-place-start 0.)                 ;  that block doesn't jump back &
                                              ;  forth between hand and board


(define board-map                             ; board is a "lego" board w/ slots
  (make-vector (inexact->exact (truncate (/ board-width block-width))) '()))
(let ((board-map-col-len
       (inexact->exact (truncate (/ board-height block-width)))))
  (letrec
   ((iterator
     (lambda (board-map-index)
       (cond
	((= 0 board-map-index)
	 (vector-set! board-map 0 (make-vector board-map-col-len #f)))
	(else
	 (vector-set! board-map board-map-index
		      (make-vector board-map-col-len #f))
	 (iterator (- board-map-index 1)))))))
   (iterator (- (vector-length board-map) 1))))

;;; Put some blocks on the board;
(display board-map)
(display "starting red... ")
(if (place-block (make-geode-block red) .50 .50)
    (m-display "red placed at .50 .50" #\newline))
(display board-map)
(display "starting red... ")
(if (place-block (make-geode-block red) .65 .55)
    (m-display "red placed at .65 .55" #\newline))
(display "starting blue... ")
(if (place-block (make-geode-block blue) .55 .50)
    (m-display "blue placed at .55 .50" #\newline))
(display "starting blue... ")
(if (place-block (make-geode-block blue) .50 .45)
    (m-display "blue placed at .50 .45" #\newline))
(display "starting green... ")
(if (place-block (make-geode-block green) .55 .55)
    (m-display "green placed at .55 .55" #\newline))
(display "starting red... ")
(if (place-block (make-geode-block red) .65 .50)
    (m-display "red placed at .65 .50" #\newline))
(display "starting purple... ")
(if (place-block (make-geode-block purple) .50 .60)
    (m-display "purple placed at .50 .60" #\newline))
(display "starting purple... ")
(if (place-block (make-geode-block purple) .60 .45)
    (m-display "purple placed at .50 .60" #\newline))


;;; Deal with movement
(ps:add-pre-frame-hook
 (lambda ()
   (let
       ((dcs-org 
	 (if doing-vr
	     (ps:dcs_to_world_coords ps:*hand-dcs*)
	     (ps:dcs_to_world_coords bl:*hand-dcs*)))
	(tmp-mat (ps:make-pfmatrix))
	; hnd-pts are the 2 pts in the hand that form the "hot spot".  The
	;  segment runs outward from the tip of the middle finger and from dorsal
	;  to ventral (palm) side if there is no block in hand.  If there is a
	;  block in the hand, the hot segment runs diagonally through the block
	;  from the hand outward and towards the palm side of the block.
	(hnd-pt1 (make-vector 3 0.))
	(hnd-pt2 (make-vector 3 0.)))
;     (vector-set! hnd-pt1 1 hand-length)
     (vector-set! hnd-pt1 2 (if blk-in-hnd block-width hand-thickness))
     (vector-set! hnd-pt2 1 (+ (if blk-in-hnd block-height .01) hand-length))
     (vector-set! hnd-pt2 2 (- hand-thickness))

     (set! hnd-pt1 (ps:vector-transform-by-mat hnd-pt1 dcs-org))
     (set! hnd-pt2 (ps:vector-transform-by-mat hnd-pt2 dcs-org))
     (m-display "blk-in-hnd = " blk-in-hnd " at start of frame " frame-counter
		#\space (pf:get-num-child board) #\c #\newline) ; debug
     (set! frame-counter (+ 1 frame-counter)) ; debug
;     (pf:dcs-trans bl:*blk-in-hand-dcs* (vector-ref hnd-pt1 0)
;		   (vector-ref hnd-pt1 1) (vector-ref hnd-pt1 2))
;     (disp-pfmatrix dcs-org)
;     (m-display #\newline hnd-pt1 #\newline #\newline)
;     (m-display (ps:seg-node-collision board '#(0. 1.95 1.) '#(0. 2.05 1.) 0)
;		#\newline)
;     (m-display hnd-pt1 hnd-pt2 #\newline)
     (define intersected-geode (ps:seg-node-collision board hnd-pt1
						      hnd-pt2 0))
;     (if intersected-geode ; debug
;	 (m-display intersected-geode #\newline))))) 

     ;;; move hand special
     (if (not doing-vr)
	 (let
	     ((init-x hand-x-dist-from-eyes-init)
	      (init-y (- board-y-distance-from-transmitter eyes-dist-from-board))
	      (init-z (+ hand-dist-from-floor-init temp-off-hand-z)))
	   ; don't let hand move through board
	   (if (equal? bl:*board-plane* intersected-geode)
	       (set! bl:hand-pos-y (- bl:hand-pos-y .05))
	       (set! bl:hand-pos-y
		     (+ init-y (* mouse-sensitivity
				  (- (ps:mouse-y) (/ (ps:mouse-y-max) 2))))))
	   (set! bl:hand-pos-x (+ init-x
				  (* mouse-sensitivity
				     (- (ps:mouse-x) (/ (ps:mouse-x-max) 2)))))
	   (pf:dcs-trans bl:*hand-dcs*
			 bl:hand-pos-x bl:hand-pos-y bl:hand-pos-z)
	   (cond
	    ((ps:mouse-check-left)
	     (pf:dcs-rot bl:*hand-dcs* 0.
			 (+ 1. (set! bl:hand-pos-p (+ 1. bl:hand-pos-p))) 0.))
	    ((ps:mouse-check-mid)
	     (pf:dcs-rot bl:*hand-dcs* 0.
			 (- (set! bl:hand-pos-p (- bl:hand-pos-p 1.)) 1.) 0.))))
	 (if (equal? bl:*board-plane* intersected-geode)
	     (let ((tmp-mat (ps:make-pfmatrix)))
	       (pf:get-dcs-mat ps:*hand-dcs* tmp-mat)
	       (pf:dcs-trans ps:*hand-dcs*
			     (ps:pfmatrix-ref tmp-mat 3 0)
			     (- (ps:pfmatrix-ref tmp-mat 3 1) .05)
			     (ps:pfmatrix-ref tmp-mat 3 2)))))

     ;;; move blocks
     (if blk-in-hnd
	 ;;; place blocks
	 (if (equal? intersected-geode bl:*board-plane*)
	     (if (< 1. (- (ps:second-count) block-pickup-start))
		 (begin
		   (set! block-place-start (ps:second-count))
		   (if (place-block
			(pf:get-child blk-in-hnd 0)
			(+ (/ (+ (vector-ref hnd-pt1 0) (vector-ref hnd-pt2 0)) 2)
			   (/ board-width 2))
			(- (/ (+ (vector-ref hnd-pt1 2) (vector-ref hnd-pt2 2)) 2)
			   board-z-distance-from-floor))
		       (begin                    ; remove block from hand
			 (pf:remove-child
			  (if doing-vr ps:*hand-dcs* bl:*hand-dcs*)
			  blk-in-hnd) ; debug
			 (pf:remove-child blk-in-hnd (pf:get-child blk-in-hnd 0))
			 (if (not (ps:Delete blk-in-hnd))
			     (m-display "   Error: delete failed" #\newline))
			 (set! blk-in-hnd #f))
		       (m-display "position taken - block not placed" #\newline)))
		 (m-display (- (ps:second-count) block-pickup-start)
			    "picking-up-block, so board i-section doesn't count"
			    #\newline))) ; debug
	 ;;; pick-up blocks
	 (if (and intersected-geode
		  (not (equal? intersected-geode bl:*board-plane*))
		  (< 1. (- (ps:second-count) block-place-start)))
	     (begin
	       (m-display "picking up block now" #\newline)
	       (set! block-pickup-start (ps:second-count))
	       (m-display
		" "
		(pf:remove-child board
				 (pf:get-parent intersected-geode 0))) ; debug
	       (m-display " " (pf:get-parent intersected-geode 0)
			  " removed from " board #\newline)
	       (let ((dcs-matrix (ps:make-pfmatrix)))
		 (pf:get-dcs-mat (pf:get-parent intersected-geode 0) dcs-matrix)
		 (let ((x (board->map (+ (/ block-width 2)
					 (ps:pfmatrix-ref dcs-matrix 3 0))))
		       (y (board->map (+ (/ block-width 2)
					 (ps:pfmatrix-ref dcs-matrix 3 1)))))
		   (m-display #\newline #\newline #\newline #\newline #\newline
			      x y)
		   (vector-set! (vector-ref board-map x) y #f)))
	       (let ((temp-node (pf:get-parent intersected-geode 0)))
		 ;;; remove connection for intersected-geode & dcs parent
		 (pf:remove-child temp-node intersected-geode)
		 (if (not (ps:Delete temp-node))
		     (m-display "   Error: delete failed" #\newline)))
 	       (set! blk-in-hnd                       ; make blk-in-hnd new parent
		     (ps:place-node intersected-geode
				    (- (/ block-width 2))
				    (+ hand-length block-height)
				    (- (/ block-width 2))
				    (if doing-vr ps:*hand-dcs* bl:*hand-dcs*)))
	       (pf:dcs-rot blk-in-hnd 0. 90. 0.)
	       (m-display "    " blk-in-hnd " added to "
			  (if doing-vr ps:*hand-dcs* bl:*hand-dcs*) #\newline))
	     (m-display "no blocks hit" #\newline))))))


;     (m-display (pf:remove-child board intersected-geode) #\newline)
;     (pf:dcs-trans bl:*blk-in-hand-dcs* (vector-ref hnd-pt2 0)
;		   (vector-ref hnd-pt2 1) (vector-ref hnd-pt2 2)))))
;     (m-display (ps:vector-transform-by-mat hnd-pt2 dcs-org) #\newline)
;     (ps:vector-transform-by-mat hnd-pt1 bl:*hand-box-dcs*) ; debug
;     (set! hnd-pt2 (ps:vector-transform-by-DCS hnd-pt1 bl:*hand-box-dcs*)))))
;   (display (ps:mouse-check-left))
;   (display (ps:mouse-check-mid))
;   (display (ps:mouse-check-right))

;;; Begin draw loop
(ps:run)