/* Cubes and Planes
   written by:  Chandler Chao from SGI provided code.

   Description:  Herein are provided some basic routines to construct a
   cube.

   Cubes are create by:  MakeCube(height, width, depth, red, green,
                                  blue, transparency, geostate)
             it returns a geoset.
*/

#include "persia-scm.h"
/* #include "cube.h" */

/* Creates a cube, returns a geoset */
pfGeoSet *MakeCube(float width, float depth, float height)
{
    pfGeoSet *gset;
    void *arena;
    pfVec3 *scoords;
    pfVec4 *scolor;
    ushort *scindex;
    pfVec2 *tcoords;
    pfVec3 *snorms;
    ushort *tindex;
    ushort *svindex;
    ushort *snindex;
    /* These data arrays are static so that they will exist after
     * this routine exits. Nope... got initialized from arena because 
     * used in a dynamic object... (MH)
     */

    /*
     * The pfGeoSet has been allocated out of shared memory, but notice that
     * the geometry data (above) has not. Since the data is static, in an
     * MP application, all processes will have a copy of the data as
     * initialized before they are forked.
     * However, changes made to the geometry data will only be visible to 
     * the process the changes are made in.
     */
    arena = pfGetSharedArena();
    gset = pfNewGSet(arena);

    scoords = (pfVec3 *) pfMalloc (sizeof(pfVec3) * 8, arena);
    scolor = (pfVec4 *) pfMalloc (sizeof(pfVec4), arena);
    scindex = (ushort *) pfMalloc (sizeof(ushort), arena);
    tcoords = (pfVec2 *) pfMalloc (sizeof(pfVec2) * 12, arena);
    snorms = (pfVec3 *) pfMalloc (sizeof(pfVec3) * 6, arena);
    tindex = (ushort *) pfMalloc (sizeof(ushort) * 24, arena);
    svindex = (ushort *) pfMalloc (sizeof(ushort) * 24, arena);
    snindex = (ushort *) pfMalloc (sizeof(ushort) * 6, arena);

    snorms[0][0] = 0.0; snorms[0][1] = 0.0; snorms[0][2] = 1.0;
    snorms[1][0] = 0.0; snorms[1][1] = 0.0; snorms[1][2] = -1.0;
    snorms[2][0] = 0.0; snorms[2][1] = 1.0; snorms[2][2] = 0.0;
    snorms[3][0] = 0.0; snorms[3][1] = -1.0; snorms[3][2] = 0.0;
    snorms[4][0] = 1.0; snorms[4][1] = 0.0; snorms[4][2] = 0.0;
    snorms[5][0] = -1.0; snorms[5][1] = 0.0; snorms[5][2] = 0.0;

    /* scoords depend on height, width, and depth */ 
    scoords[0][0] = 0.0; scoords[0][1] = 0.0; scoords[0][2] = height;
    scoords[1][0] = width; scoords[1][1] = 0.0; scoords[1][2] = height;
    scoords[2][0] = width; scoords[2][1] = depth; scoords[2][2] = height;
    scoords[3][0] = 0.0; scoords[3][1] = depth; scoords[3][2] = height;
    scoords[4][0] = 0.0; scoords[4][1] = 0.0; scoords[4][2] = 0.0;
    scoords[5][0] = width; scoords[5][1] = 0.0; scoords[5][2] = 0.0;
    scoords[6][0] = width; scoords[6][1] = depth; scoords[6][2] = 0.0;
    scoords[7][0] = 0.0; scoords[7][1] = depth; scoords[7][2] = 0.0;
    
    /* tcoords are set to equal the size of a side due to the
       fact that we want textures repeated for the whole size of the
       geometry */

#if 0
    /* Use this to look at grass texture before calling cube-scale-texture
       width *= .2; depth *= .2; height *= .2; */
    tcoords[0][0] = 0.0f; tcoords[0][1] = 0.0f;
    tcoords[1][0] = width; tcoords[1][1] = 0.0f;
    tcoords[2][0] = width; tcoords[2][1] = height;
    tcoords[3][0] = 0.0f; tcoords[3][1] = height;
    tcoords[4][0] = depth; tcoords[4][1] = 0.0f;
    tcoords[5][0] = depth; tcoords[5][1] = height;
    tcoords[6][0] = depth; tcoords[6][1] = width;
    tcoords[7][0] = 0.0f; tcoords[7][1] = width;
#elseif 0
    tcoords[0][0] = 0.0f; tcoords[0][1] = 0.0f;
    tcoords[1][0] = 1.0f; tcoords[1][1] = 0.0f;
    tcoords[2][0] = 1.0f; tcoords[2][1] = 1.0f;
    tcoords[3][0] = 0.0f; tcoords[3][1] = 1.0f;
#else
    /* Use this to look at grass texture before calling cube-scale-texture
       width *= .2; depth *= .2; height *= .2; */
    tcoords[0][0] = 0.0f; tcoords[0][1] = 0.0f;
    tcoords[1][0] = width; tcoords[1][1] = 0.0f;
    tcoords[2][0] = width; tcoords[2][1] = height;
    tcoords[3][0] = 0.0f; tcoords[3][1] = height;
    tcoords[4][0] = 0.0f; tcoords[4][1] = 0.0f;
    tcoords[5][0] = width; tcoords[5][1] = 0.0f;
    tcoords[6][0] = width; tcoords[6][1] = depth;
    tcoords[7][0] = 0.0f; tcoords[7][1] = depth;
    tcoords[8][0] = 0.0f; tcoords[8][1] = 0.0f;
    tcoords[9][0] = depth; tcoords[9][1] = 0.0f;
    tcoords[10][0] = depth; tcoords[10][1] = height;
    tcoords[11][0] = 0.0f; tcoords[11][1] = height;
#endif

#if 0
    tindex[0] = 6; tindex[1] = 4; tindex[2] = 0; tindex[3] = 7;
    tindex[4] = 5; tindex[5] = 3; tindex[6] = 0; tindex[7] = 4;
    tindex[8] = 6; tindex[9] = 4; tindex[10] = 0; tindex[11] = 7;
    tindex[12] = 5; tindex[13] = 3; tindex[14] = 0; tindex[15] = 4;
    tindex[16] = 3; tindex[17] = 2; tindex[18] = 1; tindex[19] = 0;
    tindex[20] = 3; tindex[21] = 2; tindex[22] = 1; tindex[23] = 0;
#elseif 0
    tindex[0] = 0; tindex[1] = 1; tindex[2] = 2; tindex[3] = 3;
    tindex[4] = 0; tindex[5] = 1; tindex[6] = 2; tindex[7] = 3;
    tindex[8] = 0; tindex[9] = 1; tindex[10] = 2; tindex[11] = 3;
    tindex[12] = 0; tindex[13] = 1; tindex[14] = 2; tindex[15] = 3;
    tindex[16] = 0; tindex[17] = 1; tindex[18] = 2; tindex[19] = 3;
    tindex[20] = 0; tindex[21] = 1; tindex[22] = 2; tindex[23] = 3;
#else
    tindex[0] = 4; tindex[1] = 5; tindex[2] = 6; tindex[3] = 7; /* top */
    tindex[4] = 8; tindex[5] = 9; tindex[6] = 10; tindex[7] = 11;
    tindex[8] = 4; tindex[9] = 5; tindex[10] = 6; tindex[11] = 7;
    tindex[12] = 8; tindex[13] = 9; tindex[14] = 10; tindex[15] = 11;
    tindex[16] = 0; tindex[17] = 1; tindex[18] = 2; tindex[19] = 3; /* back */
    tindex[20] = 0; tindex[21] = 1; tindex[22] = 2; tindex[23] = 3; /* front */
#endif

#if 1
    svindex[0] = 0; svindex[1] = 1; svindex[2] = 2; svindex[3] = 3;
    svindex[4] = 0; svindex[5] = 3; svindex[6] = 7; svindex[7] = 4;
    svindex[8] = 7; svindex[9] = 6; svindex[10] = 5; svindex[11] = 4;
    svindex[12] = 5; svindex[13] = 6; svindex[14] = 2; svindex[15] = 1;
    svindex[16] = 3; svindex[17] = 2; svindex[18] = 6; svindex[19] = 7;
    svindex[20] = 4; svindex[21] = 5; svindex[22] = 1; svindex[23] = 0;
#else
    svindex[0] = 0; svindex[1] = 1; svindex[2] = 2; svindex[3] = 3;
    svindex[4] = 0; svindex[5] = 3; svindex[6] = 7; svindex[7] = 4;
    svindex[8] = 4; svindex[9] = 7; svindex[10] = 6; svindex[11] = 5;
    svindex[12] = 1; svindex[13] = 5; svindex[14] = 6; svindex[15] = 2;
    svindex[16] = 3; svindex[17] = 2; svindex[18] = 6; svindex[19] = 7;
    svindex[20] = 0; svindex[21] = 4; svindex[22] = 5; svindex[23] = 1;
#endif

    snindex[0] = 0; snindex[1] = 5; snindex[2] = 1;
    snindex[3] = 4; snindex[3] = 2; snindex[5] = 3;

/*    scolor[0][0] = red; scolor[0][1] = green;
    scolor[0][2] = blue; scolor[0][3] = trans;*/
    scindex[0] = 1;

    pfGSetAttr(gset, PFGS_NORMAL3, PFGS_PER_PRIM, snorms, snindex);

    pfGSetAttr(gset, PFGS_COORD3, PFGS_PER_VERTEX, scoords, svindex);
    pfGSetAttr(gset, PFGS_COLOR4, PFGS_OVERALL, scolor, scindex);
    pfGSetAttr(gset, PFGS_TEXCOORD2, PFGS_PER_VERTEX, tcoords, tindex);
    pfGSetPrimType(gset, PFGS_QUADS);
    pfGSetNumPrims(gset, 6);
    
    return gset;
}


Object P_MakeCube(Object width,
		  Object depth,
		  Object height)
{
  pfGeode *geode;
  pfGeoSet *c;
  Check_Type(width, T_Flonum);
  Check_Type(depth, T_Flonum);
  Check_Type(height, T_Flonum);

  c = MakeCube(FLONUM(width)->val,
	       FLONUM(depth)->val,
	       FLONUM(height)->val);
  geode = pfNewGeode(); 
  pfAddGSet(geode, c);
  return Make_Pfo((pfObject*)geode);
}

Object
P_CubeSetTexScale(Object geode, Object scale)
{

#define IN_PLACE 1

  pfVec2 *tcoords;
  ushort *tindex;
  float scal;
  pfGeode *g;
  pfGeoSet *geoset;
  void *arena;
#if IN_PLACE
  pfVec3 *scoords;
  ushort *svindex;
  float width, depth, height;
#endif

  Check_Pfo_Class(geode, PFTYPE_GEODE);
  Check_Type(scale, T_Flonum);

  g = (pfGeode*) PFO(geode);
  geoset = pfGetGSet(g, 0);
  scal = 1.0f / FLONUM(scale)->val;

#if IN_PLACE
  pfGetGSetAttrLists(geoset, PFGS_TEXCOORD2, (void**)&tcoords, &tindex);
  pfGetGSetAttrLists(geoset, PFGS_COORD3, (void**)&scoords, &svindex);
#if 0
  {
    int i;
    for (i = 0; i < 8; i++)
      fprintf(stderr, 
	      "scoords[%d][0] = %f scoords[%d][1] = %f scoords[%d][1] = %f\n",
	      i, scoords[i][0], i, scoords[i][1], i, scoords[i][2]);
  }
#endif
  width = scoords[1][0];
  depth = scoords[2][1];
  height = scoords[0][2];
#else
  arena = pfGetSharedArena();
  tcoords = (pfVec2 *) pfMalloc (sizeof(pfVec2) * 8, arena);
  tindex = (ushort *) pfMalloc (sizeof(ushort) * 24, arena);
#endif /* IN_PLACE */

#if IN_PLACE
#if 0
  {
    int i;
    for (i = 1; i < 8; i++)
      fprintf(stderr, "tcoords[%d][0] = %f tcoords[%d][1] = %f\n",
	      i, tcoords[i][0], i, tcoords[i][1]);
  }
#endif
#if 0
  tcoords[1][0] = width*scal;
  tcoords[2][0] = width*scal; tcoords[2][1] = height*scal;
  tcoords[3][1] = height*scal;
  tcoords[4][0] = depth*scal;
  tcoords[5][0] = depth*scal; tcoords[5][1] = height*scal;
  tcoords[6][0] = depth*scal; tcoords[6][1] = width*scal;
  tcoords[7][1] = width*scal;
#endif /* 0 */
  tcoords[1][0] = width*scal;
  tcoords[2][0] = width*scal; tcoords[2][1] = height*scal;
  tcoords[3][1] = height*scal;
  tcoords[5][0] = width*scal;
  tcoords[6][0] = width*scal; tcoords[6][1] = depth*scal;
  tcoords[7][1] = depth*scal;
  tcoords[9][0] = depth*scal;
  tcoords[10][0] = depth*scal; tcoords[10][1] = height*scal;
  tcoords[11][1] = height*scal;
#if 0
  {
    int i;
    for (i = 1; i < 8; i++)
      fprintf(stderr, "tcoords[%d][0] = %f tcoords[%d][1] = %f\n",
	      i, tcoords[i][0], i, tcoords[i][1]);
  }
#endif
#else
  tcoords[0][0] = 0.0f; tcoords[0][1] = 0.0f;
  tcoords[1][0] = scal; tcoords[1][1] = 0.0f;
  tcoords[2][0] = scal; tcoords[2][1] = scal;
  tcoords[3][0] = 0.0f; tcoords[3][1] = scal;
  tcoords[4][0] = scal; tcoords[4][1] = 0.0f;
  tcoords[5][0] = scal; tcoords[5][1] = scal;
  tcoords[6][0] = scal; tcoords[6][1] = scal;
  tcoords[7][0] = 0.0f; tcoords[7][1] = scal;
  /* We should figure out how to just make this NULL */
  tindex[0] = 6; tindex[1] = 4; tindex[2] = 0; tindex[3] = 7;
  tindex[4] = 5; tindex[5] = 3; tindex[6] = 0; tindex[7] = 4;
  tindex[8] = 6; tindex[9] = 4; tindex[10] = 0; tindex[11] = 7;
  tindex[12] = 5; tindex[13] = 3; tindex[14] = 0; tindex[15] = 4;
  tindex[16] = 3; tindex[17] = 2; tindex[18] = 1; tindex[19] = 0;
  tindex[20] = 3; tindex[21] = 2; tindex[22] = 1; tindex[23] = 0;
  pfGSetAttr(geoset, PFGS_TEXCOORD2, PFGS_PER_VERTEX, tcoords, tindex);
#endif

  return geode;
}

Object
P_CubeSetTexCoords(Object geode, Object scw, Object scd, Object sch)
{
  pfVec2 *tcoords;
  ushort *tindex;
  pfGeode *g;
  pfGeoSet *geoset;
  void *arena;
  float width, depth, height;

  Check_Pfo_Class(geode, PFTYPE_GEODE);
  Check_Type(scw, T_Flonum);
  Check_Type(scd, T_Flonum);
  Check_Type(sch, T_Flonum);

  g = (pfGeode*) PFO(geode);
  geoset = pfGetGSet(g, 0);

  pfGetGSetAttrLists(geoset, PFGS_TEXCOORD2, (void**)&tcoords, &tindex);

  width = 1.0f / FLONUM(scw)->val;
  depth = 1.0f / FLONUM(scd)->val;
  height = 1.0f / FLONUM(sch)->val;

  tcoords[1][0] = width;
  tcoords[2][0] = width; tcoords[2][1] = height;
  tcoords[3][1] = height;
  tcoords[5][0] = width;
  tcoords[6][0] = width; tcoords[6][1] = depth;
  tcoords[7][1] = depth;
  tcoords[9][0] = depth;
  tcoords[10][0] = depth; tcoords[10][1] = height;
  tcoords[11][1] = height;

  return geode;
}

void cube_init ()
{
  Define_Primitive(P_MakeCube, "ps:make-geode-cube1", 3, 3, EVAL); 
  Define_Primitive(P_CubeSetTexScale, "ps:cube-scale-texture", 2, 2, EVAL);
  Define_Primitive(P_CubeSetTexCoords, "ps:cube-set-tcoords", 4, 4, EVAL);
}
