/* typedef int sigset_t		/* quick fix */
#include "persia-scm.h"
#include "persia.h"

#include <sys/types.h>
#include <dlfcn.h>
#include <string.h>

long elk_window_positions[MAX_CHANS][4];

static int elk_no_pre_frame = 0;

/* pfObject Scheme type */

int T_Pfo;

int Pfo_Sizeof(Object a)
{
  return sizeof(struct S_Pfo);
}

int Pfo_Equal(Object a, Object b)
{
  return (PFO(a) == PFO(b));
}

int Pfo_Print(Object o, Object port, int raw, int depth, int len)
{
  Printf(port, "#[pfobject %s 0x%x]", 
	 pfGetTypeName(PFO(o)),
	 PFO(o));
}

void Check_Pfo(Object x)
{
  Check_Type(x, T_Pfo);
}

void Check_Pfo_Class(Object x, long class)
{
  Check_Type(x, T_Pfo);
  if (!(pfGetType(PFO(x)) & class))
    Primitive_Error("pfo type mismatch on pfo: ~s", x);
}

/* xxx we should use a hashtable to make sure we don't create more
   than one of these for pfObject?? */
Object Make_Pfo(pfObject *o)
{
  Object d;

  d = Alloc_Object(sizeof(struct S_Pfo), T_Pfo, 0);
  PFO(d) = o;
  return d;
}

Object Define_Pfo(pfObject *o, const char *var_name)
{
  Object p = Make_Pfo(o);
  Object var;
  Define_Variable(&var, var_name, p);
  return p;
}


Object P_Is_Pfo(Object x)
{
  return TYPE(x) == T_Pfo ? True : False;
}

/* pfMatrix Scheme type */

int T_pfMatrix;

static int pfMatrix_Sizeof(Object a)
{
  return sizeof(struct S_pfMatrix);
}

static int pfMatrix_Equal(Object a, Object b)
{
  /* xxx fix this to compare contents of matrix */
  return (PFM(a) == PFM(b));
}

static int pfMatrix_Print(Object o, Object port, int raw, int depth, int len)
{
  Printf(port, "#[pfMatrix 0x%x]",
	 PFM(o));
}

void Check_pfMatrix(Object x)
{
  Check_Type(x, T_pfMatrix);
}

Object Make_pfMatrix(pfMatrix m)
{
  Object d;
  int i, j;

  d = Alloc_Object(sizeof(struct S_pfMatrix), T_pfMatrix, 0);
  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      PFM(d)[i][j] = m[i][j];
  return d;
}

Object Define_pfMatrix(pfMatrix m, const char *var_name)
{
  Object p = Make_pfMatrix(m);
  Object var;
  Define_Variable(&var, var_name, p);
  return p;
}


Object P_Is_pfMatrix(Object x)
{
  return TYPE(x) == T_pfMatrix ? True : False;
}


Object P_Make_pfMatrix()
{
  pfMatrix m;
  pfMakeIdentMat(m);
  return Make_pfMatrix(m);
}

Object P_pfMatrix2Vector(Object m)
{
  Object vector;
  Object matrix;
  int i, j;

  Check_Type(m, T_pfMatrix);
  matrix = Make_Vector(4, Null);
  for (i = 0; i < 4; i++)
    {
      vector = Make_Vector(4, Null);
      for (j = 0; j < 4; j++)
	P_Vector_Set(vector, V_INTS(j), Make_Flonum(PFM(m)[i][j]));
      P_Vector_Set(matrix, V_INTS(i), vector);
    }
  return matrix;
}

Object P_Vector2pfMatrix(Object v)
{
  pfMatrix m;
  Object entry;
  int i, j;

  Check_Type(v, T_Vector);
  for (i = 0; i < 4; i++)
    {
      for (j = 0; j < 4; j++)
	{
	  entry = P_Vector_Ref(P_Vector_Ref(v, V_INTS(i)), V_INTS(j));
	  Check_Type(entry, T_Flonum);
	  m[i][j] = FLONUM(entry)->val;
	}
    }
  return Make_pfMatrix(m);
}

Object P_pfMatrixRef(Object m, Object i, Object j)
{
  Check_Type(m, T_pfMatrix);
  Check_Type(i, T_Fixnum);
  Check_Type(j, T_Fixnum);
  return Make_Flonum(PFM(m)[FIXNUM(i)][FIXNUM(j)]);
}

Object P_pfMatrixSet(Object m, Object i, Object j, Object value)
{
  Check_Type(m, T_pfMatrix);
  Check_Type(i, T_Fixnum);
  Check_Type(j, T_Fixnum);
  Check_Type(value, T_Flonum);
  PFM(m)[FIXNUM(i)][FIXNUM(j)] = FLONUM(value)->val;
  return Void;
}

/* Performer functions */

Object P_NewDCS()
{
  return Make_Pfo((pfObject*)pfNewDCS());
}

Object P_DCSRot (Object x, Object h, Object p, Object f)
{
  Check_Pfo_Class(x, PFTYPE_DCS);
  Check_Type(h, T_Flonum);
  Check_Type(p, T_Flonum);
  Check_Type(f, T_Flonum);
  pfDCSRot(PFD(x), 
	   FLONUM(h)->val,
	   FLONUM(p)->val,
	   FLONUM(f)->val);
  return Void;
}

Object P_DCSRotBy(Object x, Object h, Object p, Object f)
{
  pfMatrix m;
  float head_angle_x;
  float head_angle_y;
  pfCoord c;

  Check_Pfo_Class(x, PFTYPE_DCS);
  Check_Type(h, T_Flonum);
  Check_Type(p, T_Flonum);
  Check_Type(f, T_Flonum);
  pfGetDCSMat(ViewState->eye_head_dcs, m);
  pfGetOrthoMatCoord(m, &c);
  head_angle_x = c.hpr[0];
  head_angle_y = c.hpr[1];
  pfDCSRot(PFD(x), 
	   c.hpr[0] + FLONUM(h)->val,
	   c.hpr[1] + FLONUM(p)->val,
	   c.hpr[2] + FLONUM(4)->val);
  return Void;
}

Object P_DCSTrans(Object o, Object x, Object y, Object z)
{
  Check_Pfo_Class(o, PFTYPE_DCS);
  Check_Type(x, T_Flonum);
  Check_Type(y, T_Flonum);
  Check_Type(z, T_Flonum);
  pfDCSTrans(PFD(o), 
	   FLONUM(x)->val,
	   FLONUM(y)->val,
	   FLONUM(z)->val);
  return o;
}

Object P_DCSScale(Object x, Object s)
{
  Check_Pfo_Class(x, PFTYPE_DCS);
  Check_Type(s, T_Flonum);
  pfDCSScale(PFD(x), 
	     FLONUM(s)->val);
  return x;
}

#if 0
void *malloc(size_t s)
{
  return Safe_Malloc(s);
}
void *realloc(void *p, size_t s)
{
  return Safe_Realloc(p,s);
}
#endif

Object P_ReturnDCSMat(Object d)
{
  pfMatrix pf_matrix;
  Object vector;
  Object matrix;
  int i, j;

  Check_Pfo_Class(d, PFTYPE_DCS);
  pfGetDCSMat(PFD(d), pf_matrix);

  matrix = Make_Vector(4, Null);
  for (i = 0; i < 4; i++)
    {
      vector = Make_Vector(4, Null);
      for (j = 0; j < 4; j++)
	P_Vector_Set(vector, V_INTS(j), Make_Flonum(pf_matrix[i][j]));
      P_Vector_Set(matrix, V_INTS(i), vector);
    }
  return matrix;
}

Object P_GetDCSMat(Object d, Object m)
{
  Check_Pfo_Class(d, PFTYPE_DCS);
  Check_Type(m, T_pfMatrix);
  pfGetDCSMat(PFD(d), PFM(m));
  return Void;
}

Object P_DCSMat(Object d, Object m)
{
  Check_Pfo_Class(d, PFTYPE_DCS);
  Check_Type(m, T_pfMatrix);
  pfDCSMat(PFD(d), PFM(m));
  return Void;
}

Object P_ChanView(Object chan,
		  Object x, Object y, Object z, 
		  Object h, Object p, Object r)
{
  pfCoord view;
  Check_Pfo_Class(chan, PFTYPE_CHANNEL);
  Check_Type(x, T_Flonum);
  Check_Type(y, T_Flonum);
  Check_Type(z, T_Flonum);
  Check_Type(h, T_Flonum);
  Check_Type(p, T_Flonum);
  Check_Type(r, T_Flonum);
  pfSetVec3(view.xyz, 
	    FLONUM(x)->val,
	    FLONUM(y)->val,
	    FLONUM(z)->val);
  pfSetVec3(view.hpr, 
	    FLONUM(h)->val,
	    FLONUM(p)->val,
	    FLONUM(r)->val);
  pfChanView((pfChannel*)PFO(chan), view.xyz, view.hpr);
  return chan;
}

Object P_WindowPosition(Object wn,
			Object x, Object y, Object width, Object height)
{
  int window_num;
  Check_Type(wn, T_Fixnum);
  Check_Type(x, T_Fixnum);
  Check_Type(y, T_Fixnum);
  Check_Type(width, T_Fixnum);
  Check_Type(height, T_Fixnum);
  window_num = FIXNUM(wn);
  if (window_num >= MAX_CHANS)
    Primitive_Error("too many window positions");
  elk_window_positions[window_num][0] = FIXNUM(x);
  elk_window_positions[window_num][1] = FIXNUM(y);
  elk_window_positions[window_num][2] = FIXNUM(x) + FIXNUM(width) - 1;
  elk_window_positions[window_num][3] = FIXNUM(y) + FIXNUM(height) - 1;
  return Void;
}

Object P_StartGraphics(Object num_cameras)
{
  static int done = 0;
  static char *argv[] = {"persia"};
  if (done)
    return Void;
  done = 1;

  Check_Type(num_cameras, T_Fixnum);
  start_graphics(1, argv, FIXNUM(num_cameras));

  /* Some globals variables */
  /* xxx clean up some of these names */
  Define_Pfo((pfObject*)ViewState->scene, "ps:*scene*");
  Define_Pfo((pfObject*)ViewState->sceneGroup, "ps:*scene-group*");
  Define_Pfo((pfObject*)ViewState->sceneGroup, "ps:*default-parent*");
  Define_Pfo((pfObject*)ViewState->sceneDCS, "ps:*scene-dcs*");
  Define_Pfo((pfObject*)ViewState->masterChan, "ps:*channel*");

  /* Call the Scheme hook functions */
  {
    Object proc;
    proc = P_Cdr(Lookup_Symbol(Intern("ps:*post-init-cameras-function*"), 1));
    Funcall(proc, Null, 0);
  }

  return Void;
}

Object P_PositionCamera(Object camera_num,
			Object x, Object y, Object z, 
			Object h, Object p, Object r)
{
  pfCoord view;
  Check_Type(camera_num, T_Fixnum);
  if (FIXNUM(camera_num) >= ViewState->num_cameras)
    Primitive_Error("camera index out of range");
  Check_Type(x, T_Flonum);
  Check_Type(y, T_Flonum);
  Check_Type(z, T_Flonum);
  Check_Type(h, T_Flonum);
  Check_Type(p, T_Flonum);
  Check_Type(r, T_Flonum);
  pfSetVec3(view.xyz, 
	    FLONUM(x)->val,
	    FLONUM(y)->val,
	    FLONUM(z)->val);
  pfSetVec3(view.hpr, 
	    FLONUM(h)->val,
	    FLONUM(p)->val,
	    FLONUM(r)->val);
  pfChanView(ViewState->chans[FIXNUM(camera_num)], view.xyz, view.hpr);
  return Void;
}

Object P_AttachCamera(Object camera_num, Object a_dcs)
{
  int cn;
  Check_Type(camera_num, T_Fixnum);
  cn = FIXNUM(camera_num);
  if (a_dcs == False)
    {
      ViewState->camera_pos[cn] = NULL;
    }
  else
    {
      Check_Pfo_Class(a_dcs, PFTYPE_DCS);
      if (cn < 0 || cn >= MAX_CHANS)
	Primitive_Error("camera index out of range");
      ViewState->camera_pos[cn] = (pfDCS*) PFO(a_dcs);
    }
  return Void;
}

Object P_ChanFOV(Object camera_num, Object horiz, Object vert)
{
  Check_Type(camera_num, T_Fixnum);
  Check_Type(horiz, T_Flonum);
  Check_Type(vert, T_Flonum);
  pfChanFOV(ViewState->chans[FIXNUM(camera_num)],
	    FLONUM(horiz)->val,
	    FLONUM(vert)->val);
  return Void;
}

Object P_ShowGUI(Object t_or_f)
{
  if (t_or_f == False)
    updateGUI(0);
  else
    updateGUI(1);
  return Void;
}

#if 0
Object P_MakeCamera()
{
  int n = make_camera(0,0,0,0);
  if (n > 0)
    return Make_Pfo((pfObject*)ViewState->chans[n]);
  else
    return Void;
}
#endif

Object P_Flatten(Object n)
{
  Check_Pfo_Class(n, PFCLASS_NODE);
  pfFlatten(PFN(n),0);
  return Void;
}

Object P_AddGSet(Object geode, Object gset)
{
  Check_Pfo_Class(geode, PFCLASS_GEODE);
  Check_Pfo_Class(gset, PFCLASS_NODE);
  pfAddGSet(PFN(geode), (pfGeoSet*)PFN(gset));
  return Void;
}

Object P_SetFog (Object t_or_f, 
		 Object fog_type,
		 Object near_range, Object far_range)
{
  if (t_or_f == False)
    {
      ViewState->fog = PFFOG_OFF;
    }
  else
    {
      Check_Type(fog_type, T_Fixnum);
      Check_Type(near_range, T_Flonum);
      Check_Type(far_range, T_Flonum);
#if 0
      ViewState->fog = PFFOG_VTX_LIN;
      ViewState->fog = PFFOG_VTX_EXP;
      ViewState->fog = PFFOG_VTX_EXP2;
#endif
      if (FIXNUM(fog_type) < 0)
	ViewState->fog = PFFOG_VTX_LIN;
      else
	ViewState->fog = FIXNUM(fog_type);
	
      if (FLONUM(far_range)->val < 0)
	ViewState->farFogRange = 6800.0f;
      else
	ViewState->farFogRange = FLONUM(far_range)->val;

      if (FLONUM(near_range)->val < 0)
	ViewState->nearFogRange = 8 * 4.0f;
      else
	ViewState->nearFogRange = FLONUM(near_range)->val;
      /* this was default ViewState->nearFogRange = 16000.0f; */
    }
  return Void;
}

Object P_NewBboard()
{
  pfBillboard *b = pfNewBboard();
  return Make_Pfo((pfObject*)b);
}

Object P_LoadFile(Object filename, Object gstate)
{
  char *fn;
  pfNode *o;

  Check_Type(filename, T_String);
  Make_C_String(filename, fn);
  /* fprintf(stdout,"Blah! %x",(pfGeoState*)PFO(gstate)); */
  if (gstate != Null)
    {
      /* fprintf(stdout,"Blah! %x",(pfGeoState*)PFO(gstate)); */
      Check_Pfo_Class(gstate, PFTYPE_GEOSTATE);
      o = LoadFile(fn, (pfGeoState*)PFO(gstate));
    }
  else
      o = LoadFile(fn, NULL);

  if (o == NULL)
    Primitive_Error("couldn't load file: ~s", filename);
  return Make_Pfo((pfObject*)o);
}


Object P_Frame()
{
  pfFrame();
  return Void;
}

Object P_Tick()
{
  one_loop();
  return Void;
}

Object P_RunLoop()
{
  run_loop();
  return Void;
}

Object P_mouse_x_screen ()
{
  return Make_Long(ViewState->mouse.xpos);
}

Object P_mouse_y_screen ()
{
  return Make_Long(ViewState->mouse.ypos);
}

Object P_mouse_x_screen_max ()
{
  return Make_Long(persia_mouse_x_max);
}

Object P_mouse_y_screen_max ()
{
  return Make_Long(persia_mouse_y_max);
}

Object P_SetDoPreFrame (Object t_or_f)
{
  if (t_or_f == False)
    elk_no_pre_frame = 1;
  else
    elk_no_pre_frame = 0;
  return Void;
}

Object P_dlopen(Object filename)
{
  void *handle;

  Check_Type(filename, T_String);

  /* Dynamically link the shared object. */
  handle = dlopen(STRING(filename)->data, RTLD_NOW);
  if (!handle)
    {
      char *e = dlerror();
      Primitive_Error("dlopen() error: ~s", 
		      Make_Const_String(e,strlen(e)));
      return False;
    }

  /* Call the initialization function, assumed to be named
     after the filename, with the ".so" extention removed, and
     "_init" appended. */
  {
    void (*init_func)();
    char *fn;
    int len;
    char *init_func_name, *tmp;
  
    len = strlen(STRING(filename)->data);
    
    init_func_name = (char*) alloca(len+4);
    strcpy(init_func_name, STRING(filename)->data);
    strcpy(init_func_name+len-3, "_init");
    tmp = strrchr(init_func_name, '/');
    if (tmp)
      init_func_name = tmp + 1;
    init_func = (void(*)()) dlsym(handle, init_func_name);
    if (!init_func)
      {
	char *e = dlerror();
	Primitive_Error("dlsym() error: ~s", 
			Make_Const_String(e,strlen(e)));
	dlclose(handle);
	return False;
      }
    (*init_func)();
  }
  
  return True;
}

/* Avoid GC; I should store these values in memory and provide
   functions for accessing them. */
void elk_pre_frame_procedure()
{
  if (elk_no_pre_frame)
    return;

  /* Call the Scheme hook functions */
  {
    Object proc;
    proc = P_Cdr(Lookup_Symbol(Intern("ps:*pre-frame-function*"), 1));
    Funcall(proc, Null, 0);
  }
}

static Object V_PFFOG_VTX_LIN;
static Object V_PFFOG_VTX_EXP;
static Object V_PFFOG_VTX_EXP2;
static Object V_PFFOG_PIX_LIN;
static Object V_PFFOG_PIX_EXP;
static Object V_PFFOG_PIX_EXP2;

void init_new()
{
  /* Set load-path variable for this version of persia. */
  Init_Loadpath(".:" PERSIA_SCMDIR ":/usr/grads/lib/elk/runtime/scm:/usr/grads/lib/elk/runtime/obj");

  /* The Performer Object type */
  T_Pfo = Define_Type(0, "pfobject", 
		      NOFUNC, sizeof(struct S_Pfo),
		      Pfo_Equal, Pfo_Equal, Pfo_Print, NOFUNC);

  T_pfMatrix = Define_Type(0, "pfmatrix", 
			   NOFUNC, sizeof(struct S_pfMatrix),
			   pfMatrix_Equal, pfMatrix_Equal, pfMatrix_Print,
			   NOFUNC);

  /* Some functions */
  Define_Primitive(P_Is_Pfo, "pfobject?", 1, 1, EVAL);
  Define_Primitive(P_Is_pfMatrix, "pfmatrix?", 1, 1, EVAL);

  Define_Variable(&V_PFFOG_VTX_LIN, "pf:*fog-vtx-lin*",
		  Make_Fixnum(PFFOG_VTX_LIN));
  Define_Variable(&V_PFFOG_VTX_EXP, "pf:*fog-vtx-exp*",
		  Make_Fixnum(PFFOG_VTX_EXP));
  Define_Variable(&V_PFFOG_VTX_EXP2, "pf:*fog-vtx-exp2*",
		  Make_Fixnum(PFFOG_VTX_EXP2));

  Define_Primitive(P_Make_pfMatrix, "ps:make-pfmatrix", 0, 0, EVAL);
  Define_Primitive(P_pfMatrix2Vector, "ps:pfmatrix->vector", 1, 1, EVAL);
  Define_Primitive(P_Vector2pfMatrix, "ps:vector->pfmatrix", 1, 1, EVAL);
  Define_Primitive(P_pfMatrixRef, "ps:pfmatrix-ref", 3, 3, EVAL);
  Define_Primitive(P_pfMatrixSet, "ps:pfmatrix-set!", 4, 4, EVAL);

  Define_Primitive(P_Flatten, "pf:flatten", 1, 1, EVAL);
  Define_Primitive(P_AddGSet, "pf:add-gset", 1, 1, EVAL);

  Define_Primitive(P_NewDCS, "pf:new-dcs", 0, 0, EVAL);
  Define_Primitive(P_DCSRot, "pf:dcs-rot", 4, 4, EVAL);
  Define_Primitive(P_DCSRotBy, "pf:dcs-rot-by", 4, 4, EVAL);
  Define_Primitive(P_DCSTrans, "pf:dcs-trans", 4, 4, EVAL);
  Define_Primitive(P_DCSScale, "pf:dcs-scale", 2, 2, EVAL);

  Define_Primitive(P_GetDCSMat, "pf:get-dcs-mat", 2, 2, EVAL);
  Define_Primitive(P_DCSMat, "pf:dcs-mat", 2, 2, EVAL);

  Define_Primitive(P_WindowPosition, "ps:position-window", 5, 5, EVAL);
  Define_Primitive(P_StartGraphics, "ps:init-cameras", 1, 1, EVAL);
  Define_Primitive(P_AttachCamera, "ps:attach-camera", 2, 2, EVAL);
  Define_Primitive(P_PositionCamera, "ps:position-camera", 7, 7, EVAL);
  Define_Primitive(P_ChanFOV, "pf:chan-fov", 3, 3, EVAL);
  
  /* Define_Primitive(P_CameraOff, "camera-off", 1, 1, EVAL);  */
  /* Define_Primitive(P_CameraOn, "camera-on", 1, 1, EVAL);  */

  Define_Primitive(P_ShowGUI, "ps:show-gui", 1, 1, EVAL);
  Define_Primitive(P_SetDoPreFrame, "ps:set-do-pre-frame", 1, 1, EVAL);

  Define_Primitive(P_ChanView, "pf:chan-view", 7, 7, EVAL);

  Define_Primitive(P_LoadFile, "ps:make-node-from-file1", 2, 2, EVAL);

  Define_Primitive(P_dlopen, "ps:dlopen", 1, 1, EVAL);   

  Define_Primitive(P_mouse_x_screen, "ps:mouse-x", 0, 0, EVAL);
  Define_Primitive(P_mouse_y_screen, "ps:mouse-y", 0, 0, EVAL);
  Define_Primitive(P_mouse_x_screen_max, "ps:mouse-x-max", 0, 0, EVAL);
  Define_Primitive(P_mouse_y_screen_max, "ps:mouse-y-max", 0, 0, EVAL);

  Define_Primitive(P_Frame, "pf:frame", 0, 0, EVAL);
  Define_Primitive(P_Tick, "ps:tick", 0, 0, EVAL);
  Define_Primitive(P_RunLoop, "ps:run", 0, 0, EVAL);

  Define_Primitive(P_SetFog, "ps:set-fog", 4, 4, EVAL);

  /* initialize camera window positions to indicate "no preference" */
  /* xxx Fix this so that "no preference is indicated in a separate array */
  {
    int i;
    for (i = 0; i < MAX_CHANS; i++)
      elk_window_positions[i][0] = 0;
  }
}
