/* eye.c - Scheme interface to eye-agent functions */

#include "persia-scm.h"
#include "persia.h"

#if 0
#include "persia.h"
#include <gl/gl.h>
#include <Performer/pf.h>
#include "pfsgi.h"
#include <gl/device.h>
#endif

/* The eye camera 
static pfDCS *head_dcs;
static int head_camera_num = 0; */

/*unsigned long head_camera_image[];
  retina
  vc
  mask*/

#define RETINA_WIDTH 100
#define RETINA_HEIGHT 100
#define NUMPIXELS_RETINA_ARRAY ((RETINA_WIDTH + 1) * (RETINA_HEIGHT + 1))
#define SIZEOF_RETINA_ARRAY (NUMPIXELS_RETINA_ARRAY * sizeof(unsigned long))

#define ATTENTION_REGION_COLOR 0x0
#define ATTENTION_POINT_COLOR 0x00ffff
#define ATTENTION_MASK_COLOR 0xb0b0b0
#define RETINA_BACKGROUND_COLOR 0x808080

#ifndef MAX
#define MAX(a,b) \
       ({typedef _ta = (a), _tb = (b);  \
         _ta _a = (a); _tb _b = (b);     \
         _a > _b ? _a : _b; })
#endif
#ifndef MIN
#define MIN(a,b) \
       ({typedef _ta = (a), _tb = (b);  \
         _ta _a = (a); _tb _b = (b);     \
         _a < _b ? _a : _b; })
#endif


#if 0

#define SCENE_SIZE_RADIANS_Y (DEG2RAD(frame.size.z))
#define SCENE_SIZE_RADIANS_X (DEG2RAD(frame.size.z)*frame.size.x/frame.size.y)
#define MAX_SCENE_RADIANS_Y (SCENE_SIZE_RADIANS_Y / 2)
#define MIN_SCENE_RADIANS_Y (-MAX_SCENE_RADIANS_Y)
#define MAX_SCENE_RADIANS_X (SCENE_SIZE_RADIANS_X / 2)
#define MIN_SCENE_RADIANS_X (-MAX_SCENE_RADIANS_X)

/* This converts relative sizes */
#define RADIANS2PIXELS(R) ((R) * frame.size.y / SCENE_SIZE_RADIANS_Y)
#define PIXELS2RADIANS(P) ((P) * SCENE_SIZE_RADIANS_Y / frame.size.y)

/* This converts absolute coords, like foveaAngleX to a win coord */
/* Here pixels are scene pixel coords */
/* RADIAN is the distance from center, measured in radians */
/* PIXEL is the distance from lower left, measured in pixels */
#define RADIAN2PIXEL_Y(R) (RADIANS2PIXELS(R) + frame.size.y/2)
#define RADIAN2PIXEL_X(R) (RADIANS2PIXELS(R) + frame.size.x/2)
#define PIXEL2RADIAN_Y(P) (PIXELS2RADIANS((P) - frame.size.y/2))
#define PIXEL2RADIAN_X(P) (PIXELS2RADIANS((P) - frame.size.x/2))

#define ATTN_RADIAN2PIXEL_X(R) (RADIANS2PIXELS(R) + RETINA_WIDTH/2)
#define ATTN_RADIAN2PIXEL_Y(R) (RADIANS2PIXELS(R) + RETINA_HEIGHT/2)

#define FOVEA_SIZE_RADIANS_Y (PIXELS2RADIANS(RETINA_HEIGHT))
#define FOVEA_SIZE_RADIANS_X (PIXELS2RADIANS(RETINA_WIDTH))
#define MAX_foveaAngleX (MAX_SCENE_RADIANS_X - FOVEA_SIZE_RADIANS_X/2)
#define MIN_foveaAngleX (-MAX_foveaAngleX)
#define MAX_foveaAngleY (MAX_SCENE_RADIANS_Y - FOVEA_SIZE_RADIANS_Y/2)
#define MIN_foveaAngleY (-MAX_foveaAngleY)

#define MAX_attentionAngleX (FOVEA_SIZE_RADIANS_X/2)
#define MIN_attentionAngleX (-MAX_attentionAngleX)
#define MAX_attentionAngleY (FOVEA_SIZE_RADIANS_Y/2)
#define MIN_attentionAngleY (-MAX_attentionAngleY)

#endif /* 0 */


static unsigned long retina_array[NUMPIXELS_RETINA_ARRAY];
static unsigned long vc_array[NUMPIXELS_RETINA_ARRAY];
static unsigned long mask_array[NUMPIXELS_RETINA_ARRAY];

/* offsets from center of head-camera (window) */
static float eye_angle_x = 0.0f;
static float eye_angle_y = 0.0f;

/* offsets from center of fovea (box) */
static float attn_angle_x = 0.0f;
static float attn_angle_y = 0.0f;

void draw_square(float centerx, float centery,
		float width, float height)
{
  const int lineWidth = 1;
  static unsigned long colorArray[] = {0,0,0,0,0,0,0,0};

  rectzoom(lineWidth,height);
  lrectwrite(centerx - width/2,
	     centery - height/2,
	     centerx - width/2,
	     centery - height/2,
	    colorArray);
  lrectwrite(centerx + width/2,
	     centery - height/2,
	     centerx + width/2,
	     centery - height/2,
	     colorArray);
  rectzoom(width,lineWidth);
  lrectwrite(centerx - width/2,
	     centery - height/2,
	     centerx - width/2,
	     centery - height/2,
	    colorArray);
  lrectwrite(centerx - width/2,
	     centery + height/2,
	     centerx - width/2,
	     centery + height/2,
	     colorArray);
}


Object P_ShowVcArray()
{
  ulong xsize, ysize;

  getsize(&xsize, &ysize);
  /* winset()... */
  drawmode(NORMALDRAW);
  rectzoom(1.0, 1.0);
  lrectwrite(0, ysize-RETINA_HEIGHT,
	     RETINA_WIDTH,
	     ysize,
	     vc_array);
  return Void;
}

Object P_StartEye()
{
  ViewState->doing_eye_processing = 1;
  return Void;
}

Object P_StopEye()
{
  ViewState->doing_eye_processing = 0;
  return Void;
}

Object P_EyeSetCamera (Object camnum, Object dcs)
{
  Check_Type(camnum, T_Fixnum);
  Check_Pfo_Class(dcs, PFTYPE_DCS);
  ViewState->eye_camera_num = FIXNUM(camnum);
  ViewState->eye_head_dcs = PFD(dcs);
  return Void;
}

Object P_CaptureCamera ()
{
  long xsize, ysize, xorg, yorg;
  float fov_h, fov_v;

  if (persia_frame_count < 3)
    return Void;

/* pfPushState();  pfBasicState(); */

#if 1
  getsize(&xsize, &ysize);
  /* pfGetChanFOV(ViewState->masterChan, &fov_h, &fov_v); */
  xorg = xsize / 2 - RETINA_WIDTH/2 + (eye_angle_x);
  yorg = ysize / 2 - RETINA_HEIGHT/2 + (eye_angle_y);
  draw_square(xsize/2 + (eye_angle_x),
	      ysize/2 + (eye_angle_y),
	      RETINA_WIDTH, RETINA_HEIGHT);
#elif 0
  if ((camnum < 0) || (camnum >= ViewState->num_cameras))
    return Void;
  pfGetChanOrigin(Chan[camnum], &xorg, &yorg);
  pfGetChanSize(Chan[camnum], &xsize, &ysize);
#else
  getorigin( &xorg, &yorg); 
  getsize( &xsize, &ysize);
#endif

  /* winset()... */
  drawmode(NORMALDRAW);
  rectzoom(1.0, 1.0);
    
  lrectread(xorg, yorg,
	    xorg + RETINA_WIDTH, yorg + RETINA_HEIGHT, 
	    retina_array);

#if 1 /* Reproduce in upper right corner */
  lrectwrite(xsize-RETINA_WIDTH, ysize-RETINA_HEIGHT,
	     xsize,
	     ysize,
	     retina_array);
#elif 0 /* Reproduce in upper left corner */
  lrectwrite(0, ysize-RETINA_HEIGHT,
	     RETINA_WIDTH,
	     ysize,
	     retina_array);
#else
  memcpy(vc_array, retina_array, SIZEOF_RETINA_ARRAY);

  /* Show VC array */
  lrectwrite(0, ysize-RETINA_HEIGHT,
	     RETINA_WIDTH,
	     ysize,
	     vc_array);
#endif


/*  pfPopState(); */

  return Void;
}

Object P_EyeSetAngle (Object x, Object y)
{
  Check_Type(x, T_Flonum);
  Check_Type(y, T_Flonum);
  eye_angle_x = FLONUM(x)->val;
  eye_angle_y = FLONUM(y)->val;
  return Void;
}

Object P_CenterHeadOnEye (Object gain)
{
  pfMatrix m;
  float head_angle_x;
  float head_angle_y;
  pfCoord c;

  Check_Type(gain, T_Flonum);
  pfGetDCSMat(ViewState->eye_head_dcs, m);
  pfGetOrthoMatCoord(m, &c);
  head_angle_x = c.hpr[0];
  head_angle_y = c.hpr[1];
  pfDCSRot(ViewState->eye_head_dcs, 
	   head_angle_x + (eye_angle_x - head_angle_x) * gain,
	   head_angle_y + (eye_angle_y - head_angle_y) * gain,
	   0.0f);
  /* VOR reflex */
  eye_angle_x += (eye_angle_x - head_angle_x) * gain;
  eye_angle_y += (eye_angle_y - head_angle_y) * gain;
  return Void;
}

Object P_EyeAttendRed()
{
  return Void;
}

Object P_EyeAttendGreen()
{
  return Void;
}

Object P_EyeFoveateAttention()
{
  return Void;
}

Object P_EyeAttentionMaskAll()
{
  return Void;
}

Object P_EyeAttentionUnmaskAll()
{
  return Void;
}

Object P_EyeAttentionUnmask(Object rad_x_low,
			    Object rad_y_low,
			    Object rad_x_high,
			    Object rad_y_high)
{
  return Void;
}

Object P_EyeAngleX()
{
  return Make_Flonum(eye_angle_x);
}

Object P_EyeAngleY()
{
  return Make_Flonum(eye_angle_y);
}

Object P_EyeAttnAngleX()
{
  return Make_Flonum(attn_angle_x);
}

Object P_EyeAttnAngleY()
{
  return Make_Flonum(attn_angle_y);
}


/* draw callback, called in generic.c:PostDraw() */
void eye_processing()
{
  pfPushState();
  pfBasicState();
  P_CaptureCamera();
  pfPopState();
}


void
init_eye() 
{
  fprintf(stderr, "init_eye()\n");
  Define_Primitive(P_StartEye, "ps:eye-start", 0, 0, EVAL);
  Define_Primitive(P_StopEye, "ps:eye-stop", 0, 0, EVAL);

  Define_Primitive(P_EyeSetAngle, "ps:eye-set-angle", 2, 2, EVAL);
  Define_Primitive(P_CaptureCamera, "ps:eye-snap", 0, 0, EVAL);
  Define_Primitive(P_CenterHeadOnEye, "ps:eye-center-head", 1, 1, EVAL);
  Define_Primitive(P_EyeSetCamera, "ps:eye-set-camera", 2, 2, EVAL);
}

/* #endif */
