/* Scheme wrappers for dealing with Performer GeoState's */

#include "persia-scm.h"
#include <string.h>

/* A very cheap (ugly) texture file lookup */

#define MAX_NUM_TEXTURES 50

struct tf {
  char *filename;
  pfTexture *texture;
};

static struct tf textures[MAX_NUM_TEXTURES];
static int num_textures = 0;

pfTexture*
gstate_get_texture (const char *file)
{
  int i;
  for (i = 0; i < num_textures; i++)
    if ( ! strcmp(textures[i].filename, file) )
      return textures[i].texture;
  return NULL;
}

void
gstate_add_texture (const char *file, pfTexture *tex)
{
  char *s = (char*) malloc(strlen(file)+1);
  strcpy(s, file);
  textures[num_textures].filename = s;
  textures[num_textures].texture = tex;
}

/* End of very cheap (ugly texture file lookup */


/* Pass a texture name into this routine and it returns a geostate
   set with that texture */
pfGeoState *create_gstate_with_texture(char *texture)
{
  pfColortable* ctab;
  pfGeoState *gst;
  pfTexture *tex;
  pfTexEnv *tenv;
  void *arena;

  arena = pfGetSharedArena();

  gst = pfNewGState(arena);
  pfGStateMode(gst, PFSTATE_ENTEXTURE, 1);

  ctab = pfNewCtab(2, arena);
  pfGStateAttr(gst, PFSTATE_COLORTABLE, ctab);
  pfGStateMode(gst, PFSTATE_ENCOLORTABLE, PF_ON);

#if 1
  tex = pfuNewSharedTex (texture, pfGetSharedArena());
#else
  tex = gstate_get_texture (texture);
  if (!tex)
    {
      tex = pfNewTex(arena);
      pfLoadTexFile(tex, texture);
      gstate_add_texture(texture, tex);
    }
#endif

  pfGStateAttr(gst, PFSTATE_TEXTURE, tex);
  tenv = pfNewTEnv(arena);
  pfTEnvMode(tenv, PFTE_MODULATE);
  pfGStateAttr(gst, PFSTATE_TEXENV, tenv);

  return gst;
}

/* 
 * This function will create a gstate with no texture,
 * however, it turns off the transparency of the gstate, as object can
 * no longer be presumed to have their own color values, and in order
 * for an object to be displayed when it inherits a color but not
 * texture, is for the gstate to have transparency turned off.
 * (MH)
 */
pfGeoState *create_gstate()
{
  pfColortable* ctab;
  pfGeoState *gst;
  void *arena;

  arena = pfGetSharedArena();

  gst = pfNewGState(arena);
  pfGStateMode(gst, PFSTATE_ENTEXTURE, 0);
  pfGStateMode(gst, PFSTATE_TRANSPARENCY, 0);

  ctab = pfNewCtab(2, arena);
  pfGStateAttr(gst, PFSTATE_COLORTABLE, ctab);
  pfGStateMode(gst, PFSTATE_ENCOLORTABLE, PF_ON);

  return gst;
}


/* This routine will swap a GeoSet's geoState with one passed into
   this routine.  It returns the old geoState. */
pfGeoState *swapGeoStates(pfGeoSet *gst, pfGeoState *to)
{
  pfGeoState *from;

  from = pfGetGSetGState(gst);
  pfGSetGState(gst, to);

  return from;
}

Object P_GSetSwapGeoState(Object gset, Object gstate)
{
  pfGeoState *temp;
  Check_Pfo_Class(gset, PFTYPE_GEOSET);
  Check_Type(gstate, PFTYPE_GEOSTATE);
  temp = swapGeoStates((pfGeoSet*)PFO(gset),(pfGeoState*)PFO(gstate));
  return(gset);
}


Object P_MakeTexGeoState(Object texture_filename)
{
  pfGeoState *g;
  char *fn;

  Make_C_String(texture_filename, fn);
  g = create_gstate_with_texture(fn);
  return Make_Pfo((pfObject*)g);
}

Object P_MakeGeoState()
{
  pfGeoState *g;
  g = create_gstate();
  return Make_Pfo((pfObject*)g);
}

/*
 * P_GStateShade will set the shade model of the gstate that is passed
 * into it.  The two options are PFSM_FLAT=0 and PFSM_GOURAUD=1.  This
 * should set the shading for all GeoSets that the GState is attached
 * to.  It checks input types, then calls pfGStateMode() with the
 * correct parameters.
 * (MH)
 */
Object P_GStateShade(Object gstate, Object new)
{
  Check_Pfo_Class(gstate, PFTYPE_GEOSTATE);
  Check_Type(new, T_Fixnum);
  pfGStateMode((pfGeoState*)PFO(gstate), PFSTATE_SHADEMODEL, FIXNUM(new));
  return(gstate);
}


/*
 * P_ShadeModel is an alternative to the above function, allowing the
 * user to specify a shade model for all GeoSets that don't have one
 * set, and don't have a GState attached that sets the shade model.
 * Could be useful for large shading changes, where no object have the
 * model set explicitly.
 * (MH)
 */
Object P_ShadeModel(Object value)
{
  Check_Type(value, T_Fixnum);
  pfShadeModel(FIXNUM(value));
  return(value);
}


/*
 * Even more specific than either of the two above functions,
 * P_GSetShade will set the shade model for a specific GeoSet.  Much
 * the same as P_GStateShade.  Can only be used to override the above
 * functions for a single geoset.  Will only allow you to use flat
 * shading, or wireframe.  Gouraud shading must be set at a higher
 * level (and is the default).
 * (MH)
 * PFGS_FLATSHADE = 2, PFGS_WIREFRAME = 3, PFGS_COMPILE_GL = 4
 */
Object P_GSetShade(Object gset, Object value)
{
  Check_Pfo_Class(gset, PFTYPE_GEOSET);
  Check_Type(value, T_Fixnum);
  pfGSetDrawMode((pfGeoSet*)PFO(gset), FIXNUM(value), PF_ON);
  return(gset);
}


/*
 * P_GeoStateColor will set the color of a gstate, thus setting the
 * colors of all color indexed objects that are attached to it.  This
 * results in the ability to set the color of an entire group of
 * objects at the same time, without a series of calls.  It should
 * also be more efficient than doing a single call for each object.
 * This function returns the gstate, so that it can be used in
 * embedded calls.
 * (MH)
 */
Object P_GeoStateColor(Object gstate, Object a)
{
  pfColortable *currentctab;
  pfVec4 *newctab;
  Object entry;
  int i;
  
  Check_Pfo_Class(gstate, PFTYPE_GEOSTATE);
  Check_Type(a, T_Vector);

  currentctab = pfGetGStateAttr((pfGeoState*)PFO(gstate),PFSTATE_COLORTABLE);

  newctab = pfGetCtabColors(currentctab);

  /*
   * "entry" is used here to store Vector length.  Will be reset
   * before being used for anything else.  If user includes an alpha
   * value for colors, that should be input.  Otherwise require that
   * they have at least three values in the color entry.
   * (MH)
   */

  entry = P_Vector_Length(a);
  if ((entry < 3) || (entry > 4))
    {
      /* The range error is detected via entry, but the error is in a. */
      Range_Error(a);
    }
  else {
    if (entry == 4)
      {
	/*
	 * Use the alpha value if the user provides it.  Otherwise,
	 * don't touch it.
	 * (MH)
	 */
	entry = P_Vector_Ref(a, V_INTS(3));
	Check_Type(entry, T_Flonum);
	newctab[0][3] = FLONUM(entry)->val;
      }
    else if (entry == 3)
      {
	/*
	 * Default to an alpha value of 1.0 if not provided.
	 * (MH)
	 */
	newctab[0][3] =1.0f;
      }
    for (i = 0; i < 3; i++)
      {
	entry = P_Vector_Ref(a, V_INTS(i));
	Check_Type(entry, T_Flonum);
	newctab[0][i] = FLONUM(entry)->val;
      }
    
    pfCtabColor(currentctab,1,newctab[0]);
  }
  return gstate;
}

Object P_GStateAttr (Object gstate, Object attr, Object value)
{
  Check_Pfo_Class(gstate, PFTYPE_GEOSTATE);
  Check_Type(attr, T_Fixnum);
  Check_Pfo_Class(value, PFCLASS_LIBPR);

  pfGStateAttr((pfGeoState*) PFO(gstate),
	       FIXNUM(attr),
	       PFO(value));
  return Void;
}

Object P_GStateMode (Object gstate, Object mode, Object value)
{
  Check_Pfo_Class(gstate, PFTYPE_GEOSTATE);
  Check_Type(mode, T_Fixnum);
  Check_Type(value, T_Fixnum);

  pfGStateMode ((pfGeoState*) PFO(gstate),
		FIXNUM(mode),
		FIXNUM(value));
  return Void;
}

static Object V_PFSTATE_FRONTMTL;
static Object V_PFSTATE_BACKMTL;
static Object V_PFSTATE_TEXTURE;
static Object V_PFSTATE_TEXENV;
static Object V_PFSTATE_FOG;
static Object V_PFSTATE_LIGHTMODEL;
static Object V_PFSTATE_LIGHTS;
static Object V_PFSTATE_COLORTABLE;
static Object V_PFSTATE_HIGHLIGHT;

static Object V_PFSTATE_TRANSPARENCY;
static Object V_PFSTATE_ANTIALIAS;
static Object V_PFSTATE_DECAL;
static Object V_PFSTATE_ALPHAFUNC;
static Object V_PFSTATE_ALPHAREF;
static Object V_PFSTATE_ENLIGHTING;
static Object V_PFSTATE_ENTEXTURE;
static Object V_PFSTATE_ENFOG;
static Object V_PFSTATE_CULLFACE;
static Object V_PFSTATE_ENWIREFRAME;
static Object V_PFSTATE_ENCOLORTABLE;
static Object V_PFSTATE_ENHIGHLIGHTING;

static Object V_PF_ON;
static Object V_PF_OFF;

void gstate_init ()
{
  Define_Variable(&V_PFSTATE_FRONTMTL, "pf:*state-frontmtl*", 
		  Make_Fixnum(PFSTATE_FRONTMTL));
  Define_Variable(&V_PFSTATE_BACKMTL, "pf:*state-backmtl*",
		  Make_Fixnum(PFSTATE_BACKMTL));
  Define_Variable(&V_PFSTATE_TEXTURE, "pf:*state-texture*", 
		  Make_Fixnum(PFSTATE_TEXTURE));
  Define_Variable(&V_PFSTATE_TEXENV, "pf:*state-texenv*", 
		  Make_Fixnum(PFSTATE_TEXENV));
  Define_Variable(&V_PFSTATE_FOG, "pf:*state-fog*", 
		  Make_Fixnum(PFSTATE_FOG));
  Define_Variable(&V_PFSTATE_LIGHTMODEL, "pf:*state-lightmodel*", 
		  Make_Fixnum(PFSTATE_LIGHTMODEL));
  Define_Variable(&V_PFSTATE_LIGHTS, "pf:*state-lights*", 
		  Make_Fixnum(PFSTATE_LIGHTS));
  Define_Variable(&V_PFSTATE_COLORTABLE, "pf:*state-colortable*", 
		  Make_Fixnum(PFSTATE_COLORTABLE));
  Define_Variable(&V_PFSTATE_HIGHLIGHT, "pf:*state-highlight*", 
		  Make_Fixnum(PFSTATE_HIGHLIGHT));

  Define_Variable(&V_PFSTATE_TRANSPARENCY, "pf:*state-transparency*", 
		  Make_Fixnum(PFSTATE_TRANSPARENCY));
  Define_Variable(&V_PFSTATE_ANTIALIAS, "pf:*state-antialias*", 
		  Make_Fixnum(PFSTATE_ANTIALIAS));
  Define_Variable(&V_PFSTATE_DECAL, "pf:*state-decal*", 
		  Make_Fixnum(PFSTATE_DECAL));
  Define_Variable(&V_PFSTATE_ALPHAFUNC, "pf:*state-alphafunc*", 
		  Make_Fixnum(PFSTATE_ALPHAFUNC));
  Define_Variable(&V_PFSTATE_ALPHAREF, "pf:*state-alpharef*", 
		  Make_Fixnum(PFSTATE_ALPHAREF));
  Define_Variable(&V_PFSTATE_ENLIGHTING, "pf:*state-enlighting*", 
		  Make_Fixnum(PFSTATE_ENLIGHTING));
  Define_Variable(&V_PFSTATE_ENTEXTURE, "pf:*state-entexture*", 
		  Make_Fixnum(PFSTATE_ENTEXTURE));
  Define_Variable(&V_PFSTATE_ENFOG, "pf:*state-enfog*", 
		  Make_Fixnum(PFSTATE_ENFOG));
  Define_Variable(&V_PFSTATE_CULLFACE, "pf:*state-cullface*", 
		  Make_Fixnum(PFSTATE_CULLFACE));
  Define_Variable(&V_PFSTATE_ENWIREFRAME, "pf:*state-enwireframe*", 
		  Make_Fixnum(PFSTATE_ENWIREFRAME));
  Define_Variable(&V_PFSTATE_ENCOLORTABLE, "pf:*state-encolortable*", 
		  Make_Fixnum(PFSTATE_ENCOLORTABLE));
  Define_Variable(&V_PFSTATE_ENHIGHLIGHTING, "pf:*state-enhighlighting*", 
		  Make_Fixnum(PFSTATE_ENHIGHLIGHTING));

  Define_Variable(&V_PF_ON, "pf:*on*", 
		  Make_Fixnum(PF_ON));
  Define_Variable(&V_PF_OFF, "pf:*off*", 
		  Make_Fixnum(PF_OFF));

  Define_Primitive(P_MakeGeoState, "ps:make-plain-gstate", 0, 0, EVAL);
  Define_Primitive(P_MakeTexGeoState, "ps:make-textured-gstate", 1,1,EVAL);
  Define_Primitive(P_GeoStateColor, "ps:gstate-color", 2,2, EVAL);
  Define_Primitive(P_GSetShade, "pf:gset-shade", 2,2, EVAL);
  Define_Primitive(P_GStateShade, "pf:gstate-shade", 2,2, EVAL);
  Define_Primitive(P_ShadeModel, "pf:shade-model", 1,1, EVAL);
  Define_Primitive(P_GSetSwapGeoState, "ps:gset-swap-gstate", 2, 2, EVAL);

  Define_Primitive(P_GStateAttr, "pf:gstate-attr", 3, 3, EVAL);
  Define_Primitive(P_GStateMode, "pf:gstate-mode", 3, 3, EVAL);
}
