/*
 * Copyright (c) 1993 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that (i) the above copyright notices and this
 * permission notice appear in all copies of the software and related
 * documentation, and (ii) the name of Silicon Graphics may not be
 * used in any advertising or publicity relating to the software
 * without the specific, prior written permission of Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */


/*
 *	    gui.c
 */

#include <stdio.h>
#include <string.h>

#include <Performer/pf.h>
#include "pfutil.h"
#include "main.h"
#include "perfly.h"
#include "gui.h"

	/*-----------------------------------------------------------*/

static pfuPanel *initPanel(void);

static pfuGUIString xformer_strs[] = {"Track", "Drive", "Fly"};
static long xformer_vals[] = {PFUXF_TRACKBALL, PFUXF_DRIVE, PFUXF_FLY};

static pfuGUIString ESky_strs[] = {"Sky/Clear","Sky/Grnd","Sky", "Clear","Tag"};
static long ESky_vals[] = { PFES_SKY_CLEAR,PFES_SKY_GRND,PFES_SKY,PFES_FAST,
		        PFES_TAG};
			
static pfuGUIString Cull_strs[] = {"All", 
				    "View",
				    "View, GSet",
				    "Sort", 
				    "ignore lsources", 
				    "View, Sort", 
				    "View, GSet, Sort", 
				    "View (no lsources)",
				    "Sort (ignore lsources)", 
				    "View, GSet (ignore lsources)",
				    "View, Sort (ignore lsources)", 
				    "View, GSet, Sort (ignore lsources)", 
				    "None"};
static long Cull_vals[] = { PFCULL_ALL,
			    PFCULL_VIEW,
			    PFCULL_VIEW|PFCULL_GSET,
			    PFCULL_SORT, 
			    PFCULL_IGNORE_LSOURCES, 
			    PFCULL_VIEW|PFCULL_SORT, 
			    PFCULL_VIEW|PFCULL_GSET|PFCULL_SORT, 
			    PFCULL_VIEW|PFCULL_IGNORE_LSOURCES,
			    PFCULL_SORT|PFCULL_IGNORE_LSOURCES,
			    PFCULL_VIEW|PFCULL_GSET|PFCULL_IGNORE_LSOURCES,
			    PFCULL_VIEW|PFCULL_SORT|PFCULL_IGNORE_LSOURCES,
			    PFCULL_VIEW|PFCULL_GSET|PFCULL_SORT|PFCULL_IGNORE_LSOURCES,
			    0};

static pfuGUIString fog_strs[] = 
{
    "Off", 
    "Vtx Lin",  
    "Pix Lin", 
    "Vtx Exp",  
    "Pix Exp", 
    "Vtx Exp2", 
    "Pix Exp2",
    "Spline"
};
static long fog_vals[] = 
{
    PFFOG_OFF,
    PFFOG_VTX_LIN,
    PFFOG_PIX_LIN,
    PFFOG_VTX_EXP,
    PFFOG_PIX_EXP,
    PFFOG_VTX_EXP2,
    PFFOG_PIX_EXP2,
    PFFOG_PIX_SPLINE
};

#define NUM_STATS_VALS 4
static pfuGUIString stats_strs[] = {"Dft", "Gfx", "CPU", "Fill"};
static long stats_vals[] = {PFFSTATS_ENPFTIMES, 
	    PFSTATS_ENGFX | PFFSTATS_ENDB | PFFSTATS_ENCULL | PFFSTATS_ENPFTIMES, 
	    PFSTATSHW_ENCPU | PFFSTATS_ENPFTIMES, 
	    PFSTATSHW_ENGFXPIPE_FILL | PFFSTATS_ENPFTIMES};

	/*-----------------------------------------------------------*/


void
initGUI(pfChannel *masterChan)
{
    float 	a, b, c, d;
    pfuPanel	*panel;

    /* Set up size and position of GUI control panel */
    pfGetChanViewport(masterChan, &a, &b, &c, &d);

    if (ViewState->guiFormat == GUI_VERTICAL)
    {
	pfuGUIViewport(a, a + 0.22 * (b-a), c, d);
	pfChanViewport(masterChan, a + 0.22 * (b-a), b, c, d);
    } 
    else 
    {
	pfuGUIViewport(a, b, c, c + 0.22 * (d-c));
	pfChanViewport(masterChan, a , b, c + 0.22 * (d-c), d);
    }
    
    panel = initPanel();
    pfuEnablePanel(panel, 1);
}


void
updateWidget(long widget, float val)
{
    if (widget == GUI_STATS)
	pfuWidgetOnOff(ViewState->guiWidgets[widget], val);
    else
	pfuWidgetValue(ViewState->guiWidgets[widget], val);
}


void
updateGUIViewMsg(void)
{
    static char     posBuf[256];
    pfCoord	    *viewpt;

    /* put in position */
    viewpt =  &ViewState->viewCoord;
    sprintf(posBuf," X%5.1f Y%5.1f Z%5.1f H%5.1f P%5.1f R%5.1f",
	    viewpt->xyz[PF_X],viewpt->xyz[PF_Y],viewpt->xyz[PF_Z],
	    viewpt->hpr[PF_H],viewpt->hpr[PF_P],viewpt->hpr[PF_R]);
    pfuWidgetLabel(ViewState->guiWidgets[GUI_MSG], posBuf);
}


/*
 *    updateGUI-
 *	    This routine is called when channels are updated,
 *	    i.e. the GUI is turned ON/OFF
 */
void
updateGUI(long guiOn)
{
    
    if( guiOn )
    {
	float l, r, b, t;
	float ca, cb, cc, cd;
	
	pfuEnableGUI(TRUE);
    	pfuGetGUIViewport(&l, &r, &b, &t);
	pfGetChanViewport(ViewState->masterChan,&ca,&cb,&cc,&cd);
	
	if(ViewState->guiFormat == GUI_VERTICAL)
	    pfChanViewport(ViewState->masterChan, r, cb, cc, cd);
	else
	    pfChanViewport(ViewState->masterChan, ca, cb, t, cd);
    }
    else    /* turn off GUI */
    {
	pfuEnableGUI(FALSE);

	if (Multipipe || (NumChans == 1))
	   pfChanViewport(ViewState->masterChan, 0.0f, 1.0f, 0.0f, 1.0f);
	else if (NumChans > 1)
	   pfuConfigMCO(Chan,NumChans);
    }
}

static void
controlPanel(pfuWidget *w)
{
    switch(pfuGetWidgetId(w))
    {
    case GUI_COLLIDE:
	xformerMode(ViewState->xformerMode, 
		    ViewState->collideMode = (long)pfuGetWidgetValue(w));
	break;

    case GUI_CULL:
	ViewState->cullMode = (long)pfuGetWidgetValue(w);
	break;

    case GUI_ESKY:
	ViewState->earthSkyMode = (long)pfuGetWidgetValue(w);
	break;

    case GUI_FAR:
	ViewState->far = pfuGetWidgetValue(w);
	pfChanNearFar(ViewState->masterChan, ViewState->near, ViewState->far);
	break;

    case GUI_XFORMER:
	xformerMode(ViewState->xformerMode = (long)pfuGetWidgetValue(w),
		    ViewState->collideMode);
	break;

    case GUI_FOG:
	ViewState->fog = pfuGetWidgetValue(w);
	break;

    /* Update Fog Far Range according to slider value */
    case GUI_FAR_FOG_RANGE:
	ViewState->farFogRange = pfuGetWidgetValue(w);
	if (ViewState->farFogRange < ViewState->nearFogRange)
	{
	    ViewState->farFogRange = 1.01f * ViewState->nearFogRange;
	    pfuWidgetValue(ViewState->guiWidgets[GUI_FAR_FOG_RANGE], 
		ViewState->farFogRange);
	}
	break;

    case GUI_NEAR_FOG_RANGE:
	ViewState->nearFogRange = pfuGetWidgetValue(w);
	if (ViewState->farFogRange < ViewState->nearFogRange)
	{
	    ViewState->nearFogRange = (1.0f/1.01f) * ViewState->farFogRange;
	    pfuWidgetValue(ViewState->guiWidgets[GUI_NEAR_FOG_RANGE], 
		ViewState->nearFogRange);
	}
	break;

    /* Update Field of View according to slider value */
    case GUI_FOV:  
	{
	    int		 c;
	    float	 fov	= pfuGetWidgetValue(w)/NumChans;
	    pfVec3	 xyz;
	    pfVec3	 hpr;

	    /* all channels share this new FOV */
	    pfChanFOV(ViewState->masterChan, fov, -1.0f);

#if ! MCCALLUM
	    /* all channels have a positional offset of zero */
	    pfSetVec3(xyz, 0.0f, 0.0f, 0.0f);

	    /* set each channel's rotational offset */
	    for (c = 0; c < NumChans; c++)
	    {
		pfSetVec3(hpr, (((NumChans - 1)*0.5f) - c)*fov, 0.0f, 0.0f);
		pfChanViewOffsets(Chan[ChanOrder[c]], xyz, hpr);
	    }
#endif
	}
	break;

    /* Turns off GUI- F1 to turn back on */
    case GUI_GUI: 
	ViewState->gui = 0;
	ViewState->updateChannels = 1;
	break;

    /* Toggle Lighting */
    case GUI_LIGHTING:  
	ViewState->lighting = (long)pfuGetWidgetValue(w);
	drawModesChanged();
	break;
	
    case GUI_AA:
    {
	static long prevFade = 1;
	ViewState->aa = (long)pfuGetWidgetValue(w);
	if (!(ViewState->aa))
	{
	    if (prevFade =  ViewState->fade)
	    {
		ViewState->fade = 0;
		pfChanLODAttr(ViewState->masterChan, PFLOD_FADE, 0.0f);
		pfuWidgetValue(ViewState->guiWidgets[GUI_LOD_FADE],0.0f);
	    }
	}
	else if (ViewState->aa && prevFade && getgdesc(GD_MULTISAMPLE))
	{
	    ViewState->fade = 1;
	    pfChanLODAttr(ViewState->masterChan, PFLOD_FADE, ViewState->fadeRange);
	    pfuWidgetValue(ViewState->guiWidgets[GUI_LOD_FADE],1.0f);
	}
	drawModesChanged();
	break;
    }
    /* Update LOD Fade Range according to slider value */
    case GUI_LOD_FADE:
	ViewState->fade =  pfuGetWidgetValue(w);
	if (ViewState->realityGraphics)
	{
	    if (ViewState->fade)
		pfChanLODAttr(ViewState->masterChan, PFLOD_FADE, ViewState->fadeRange);
	    else 
		pfChanLODAttr(ViewState->masterChan, PFLOD_FADE, 0.0f);
	}
	break;

    case GUI_LOD_FADE_RANGE:
	if (ViewState->realityGraphics)
	{
	    if (ViewState->fade)
		pfChanLODAttr(ViewState->masterChan, PFLOD_FADE, 
			ViewState->fadeRange = pfuGetWidgetValue(w));
	    else
		pfChanLODAttr(ViewState->masterChan, PFLOD_FADE, 0.0f);
	}
	break;

    /* Update LOD Scale according to slider value */
    case GUI_LOD_SCALE:
	ViewState->LODscale = pfuGetWidgetValue(w);
	break;

    /* Update Near Clip Plane according to slider value */
    case GUI_NEAR:
	ViewState->near = pfuGetWidgetValue(w);
	pfChanNearFar(ViewState->masterChan, ViewState->near, ViewState->far);
	break;

    case GUI_PHASE:
	pfPhase((long)pfuGetWidgetValue(w));
	break;

    case GUI_PIX_LIMIT:
	pfChanLODAttr(ViewState->masterChan, PFLOD_STRESS_PIX_LIMIT,
	    (long)pfuGetWidgetValue(w));
	break;

    case GUI_REPOSITION:
	resetPosition(POS_ORIG);
	break;

    case GUI_CENTER:
	resetPosition(POS_CENTER);
	break;

    case GUI_QUIT:
	ViewState->exitFlag = TRUE;
	break;

    case GUI_STATS:
	ViewState->stats = pfuIsWidgetOn(w) ? 1 : 0;
	ViewState->statsEnable = (ulong) pfuGetWidgetValue(w);
	break;

    case GUI_STRESS:
	ViewState->stress = (long)pfuGetWidgetValue(w);
	break;

    case GUI_STRESS_LOW:
	{
	    float low = pfuGetWidgetValue(w);

	    if (low < ViewState->highLoad)
		ViewState->lowLoad = low;
	    else
	    {
		ViewState->lowLoad = ViewState->highLoad -0.1f;
		pfuWidgetValue(w,ViewState->lowLoad);
	    }
	}
	break;

    case GUI_STRESS_HIGH:
	{
	    float  high = pfuGetWidgetValue(w);

	    if (high > ViewState->lowLoad)
		ViewState->highLoad = high;
	    else
	    {
		ViewState->highLoad = ViewState->lowLoad + 0.1f;
		pfuWidgetValue(w,ViewState->highLoad);
	    }
	}
	break;

    case GUI_STRESS_MAX:
	ViewState->stressMax = pfuGetWidgetValue(w);
	break;

    case GUI_STRESS_SCALE:
	ViewState->stressScale = pfuGetWidgetValue(w);
	break;

    case GUI_TEXTURE:
	ViewState->texture = (long)pfuGetWidgetValue(w);
	drawModesChanged();
	break;

    case GUI_TOD:
	updateTimeOfDay(pfuGetWidgetValue(w));
	break;

    case GUI_WIREFRAME:
	ViewState->wireframe = (long)pfuGetWidgetValue(w);
	drawModesChanged();
	break;
    
    case GUI_SNAP:
	ViewState->snapImage = 1;
	break;

    default:
	break;
    }
}


void resetAll(pfuWidget *w)
{
    w; /* suppress compiler warn */

    pfuResetGUI();
    resetPosition(POS_ORIG);
}


static pfuPanel *
initPanel(void)
{
    float  	xo, yo, ys, xs;
    long    	x, y, yTop;
    long   	xSize, ySize;
    pfuWidget 	*wid;
    pfuPanel 	*panel;

    /* init the GUI panel for perfly */
    panel = pfuNewPanel();
    pfuGetPanelSize(panel, &xo, &xs, &yo, &ys);

    x = (long)xo;
    yTop = (long)ys;
    y = yTop - PFUGUI_BUTTON_YINC;

    xSize = PFUGUI_BUTTON_VLONG_XSIZE +(PFUGUI_BUTTON_VLONG_XINC);
    ySize = PFUGUI_MESSAGE_YSIZE;

    /* the message bar with positions */
    wid = pfuNewWidget(panel, PFUGUI_MESSAGE, 0);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    ViewState->guiWidgets[GUI_MSG] = wid;

    xSize = PFUGUI_BUTTON_VLONG_XSIZE;
    ySize = PFUGUI_BUTTON_YSIZE;

    y -= PFUGUI_BUTTON_YINC;

    /* a quit button */
    wid = pfuNewWidget(panel, PFUGUI_BUTTON, GUI_QUIT);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Quit");
    pfuWidgetFunc(wid, controlPanel);

    x += PFUGUI_BUTTON_VLONG_XINC;

    /* kill the gui */
    wid = pfuNewWidget(panel, PFUGUI_BUTTON, GUI_GUI);
    pfuWidgetLabel(wid,"GUI Off");
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetFunc(wid, controlPanel);

    x = xo;
    y -= (PFUGUI_BUTTON_YINC);
    xSize = PFUGUI_BUTTON_VLONG_XSIZE;

    /* Reset to origin button */
    wid = pfuNewWidget(panel, PFUGUI_BUTTON, GUI_REPOSITION);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Reposition");
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_REPOSITION] = wid;

    x += (PFUGUI_BUTTON_VLONG_XINC);
    
    /* Reset to center */
    wid = pfuNewWidget(panel, PFUGUI_BUTTON, GUI_CENTER);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Center");
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_CENTER] = wid;

    x = xo;
    y -= (PFUGUI_BUTTON_YINC);

    /* Snap screen */
    wid = pfuNewWidget(panel, PFUGUI_BUTTON, GUI_SNAP);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Snap Screen");
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_SNAP] = wid;

    x += (PFUGUI_BUTTON_VLONG_XINC);
    
    /* Reset all */
    wid = pfuNewWidget(panel, PFUGUI_BUTTON, GUI_RESET_ALL);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Reset All");
    pfuWidgetFunc(wid, resetAll);
    ViewState->guiWidgets[GUI_RESET_ALL] = wid;

    x = xo;
    y -= (PFUGUI_BUTTON_YINC + 6);

    /* toggle Wireframe*/
    wid = pfuNewWidget(panel, PFUGUI_SWITCH, GUI_WIREFRAME);
    pfuWidgetLabel(wid,"Wireframe");
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetFunc(wid, controlPanel);
    pfuWidgetDefaultValue(wid,PF_OFF);
    ViewState->guiWidgets[GUI_WIREFRAME] = wid;

    x += (PFUGUI_BUTTON_VLONG_XINC);

    /* toggle Lighting*/
    wid = pfuNewWidget(panel, PFUGUI_SWITCH, GUI_LIGHTING);
    pfuWidgetLabel(wid,"Lighting");
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetFunc(wid, controlPanel);
    pfuWidgetDefaultValue(wid,ViewState->lighting);
    pfuWidgetValue(wid,ViewState->lighting);
    ViewState->guiWidgets[GUI_LIGHTING] = wid;

    x = xo;
    y -= (PFUGUI_BUTTON_YINC);

    /* toggle Texture */
    wid = pfuNewWidget(panel, PFUGUI_SWITCH, GUI_TEXTURE);
    pfuWidgetLabel(wid,"Texture");
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetFunc(wid, controlPanel);
    pfuWidgetDefaultValue(wid,ViewState->texture);
    ViewState->guiWidgets[GUI_TEXTURE] = wid;

    x += (PFUGUI_BUTTON_VLONG_XINC);

    /* toggle Collision */
    wid = pfuNewWidget(panel, PFUGUI_SWITCH, GUI_COLLIDE);
    pfuWidgetLabel(wid,"Collide");
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetFunc(wid, controlPanel);
    pfuWidgetDefaultValue(wid,ViewState->collideMode);
    ViewState->guiWidgets[GUI_COLLIDE] = wid;

    x = xo;
    y -= (PFUGUI_BUTTON_YINC);
    xSize = PFUGUI_BUTTON_VLONG_XSIZE +(PFUGUI_BUTTON_VLONG_XINC);

    /* Earth/Sky Model*/
    wid = pfuNewWidget(panel, PFUGUI_MENU_BUTTON, GUI_ESKY);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Earth/Sky");
    pfuWidgetFunc(wid, controlPanel);
    /* if on RE, add PFES_TAG option */
    if (ViewState->realityGraphics) 
	pfuWidgetSelections(wid, ESky_strs, ESky_vals, NULL, 5);
    else
	pfuWidgetSelections(wid, ESky_strs, ESky_vals, NULL, 4);
    pfuWidgetDefaultValue(wid, ViewState->earthSkyMode);
    ViewState->guiWidgets[GUI_ESKY] = wid;

    if (ViewState->guiFormat == GUI_HORIZONTAL)
    {
	x = xo += 2*(PFUGUI_BUTTON_VLONG_XINC);
	y = ys;
    } 
    
    x = xo;
    y -= (PFUGUI_BUTTON_YINC);
    xSize = PFUGUI_BUTTON_VLONG_XSIZE +(PFUGUI_BUTTON_VLONG_XINC);

    /* Stats mode */
    wid = pfuNewWidget(panel,
		PFUGUI_RADIO_BUTTON_TOGGLE | PFUGUI_HORIZONTAL, GUI_STATS);
    pfuWidgetLabel(wid, "Stats");
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetFunc(wid, controlPanel);
    pfuWidgetSelections(wid, stats_strs, stats_vals, NULL, NUM_STATS_VALS);
    ViewState->guiWidgets[GUI_STATS] = wid;
    
    xSize = PFUGUI_BUTTON_VLONG_XSIZE;
    ySize = PFUGUI_BUTTON_YSIZE;
    y -= (PFUGUI_BUTTON_YINC);
    
    /* Stress button */
    wid = pfuNewWidget(panel, PFUGUI_SWITCH, GUI_STRESS);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Stress");
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_STRESS] = wid;

    x += PFUGUI_BUTTON_VLONG_XINC;

    /* Xformer model */
    wid = pfuNewWidget(panel, PFUGUI_MENU_BUTTON, GUI_XFORMER);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Motion");
    pfuWidgetFunc(wid, controlPanel);
    pfuWidgetSelections(wid, xformer_strs, xformer_vals, NULL, 3);
    pfuWidgetDefaultValue(wid, ViewState->xformerMode);
    ViewState->guiWidgets[GUI_XFORMER] = wid;

    y -= (PFUGUI_SLIDER_YINC);
    x = xo;

    /* Stress Low Slider */
    xSize = PFUGUI_BUTTON_VLONG_XSIZE +(PFUGUI_BUTTON_VLONG_XINC);
    ySize = PFUGUI_SLIDER_YSIZE;
    wid = pfuNewWidget(panel, PFUGUI_SLIDER, GUI_STRESS_LOW);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Stress Low");
    pfuWidgetRange(wid, 1, 0.0f, 1.0f, ViewState->lowLoad);
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_STRESS_LOW] = wid;

    y -= (PFUGUI_SLIDER_YINC);

    /* Stress High Slider */
    wid = pfuNewWidget(panel, PFUGUI_SLIDER, GUI_STRESS_HIGH);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Stress High");
    pfuWidgetRange(wid, 1, 0.0f, 1.0f, ViewState->highLoad);
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_STRESS_HIGH] = wid;

    y -= (PFUGUI_SLIDER_YINC);

    /* Stress Max Slider */
    wid = pfuNewWidget(panel, PFUGUI_SLIDER, GUI_STRESS_MAX);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Stress Max");
    pfuWidgetRange(wid, 1, 0.0f, 100.0f, ViewState->stressMax);
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_STRESS_MAX] = wid;

    y -= (PFUGUI_SLIDER_YINC);

    /* Stress Scale Slider */
    wid = pfuNewWidget(panel, PFUGUI_SLIDER, GUI_STRESS_SCALE);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Stress Scale");
    pfuWidgetRange(wid, 1, 0.0f, 4.0f, ViewState->stressScale);
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_STRESS_SCALE] = wid;

    if (ViewState->guiFormat == GUI_HORIZONTAL)
    {
	xo += 2*(PFUGUI_BUTTON_VLONG_XINC);
	y = ys;
    }
    x = xo;
   
    ySize = PFUGUI_SLIDER_YSIZE;
    y -= (PFUGUI_SLIDER_YINC);

    /* Global LOD Scale Slider */
    wid = pfuNewWidget(panel, PFUGUI_SLIDER_LOG, GUI_LOD_SCALE);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Lod Scale");
    pfuWidgetRange(wid, 1, 0.0f, 500.0f, ViewState->LODscale);
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_LOD_SCALE] = wid;

    y -= (PFUGUI_SLIDER_YINC);

    /* Pixel Limit Slider */
    wid = pfuNewWidget(panel, PFUGUI_SLIDER, GUI_PIX_LIMIT);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Pixel Limit");
    pfuWidgetRange(wid, 1, -1.0f, 1280.0f, -1.0f);
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_PIX_LIMIT] = wid;

    y -= (PFUGUI_SLIDER_YINC );

     /* Slider for FOV */
    wid = pfuNewWidget(panel, PFUGUI_SLIDER, GUI_FOV);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Field of View");
    switch (NumChans)
    {
    case 1:
	pfuWidgetRange(wid, 1, 10.0f, 160.0f, 45.0f);
	break;
    case 2:
	pfuWidgetRange(wid, 1, 10.0f, 320.0f, 90.0f);
	break;
    default:
	pfuWidgetRange(wid, 1, 10.0f, 360.0f, PF_MIN2(NumChans * 45.0f, 360.0f));
	break;
    }
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_FOV] = wid;

    y -= PFUGUI_SLIDER_YINC;

    /* Slider to set far clip plane distance */
    wid = pfuNewWidget(panel, PFUGUI_SLIDER_LOG, GUI_FAR);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Far Clip");
    pfuWidgetRange(wid, 1, ViewState->sceneSize / 10.0f, 
			   ViewState->far * 8.0f, 
			   ViewState->far);
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_FAR] = wid;

    y -= PFUGUI_SLIDER_YINC;

    /* Slider to set near clip plane distance */
    wid = pfuNewWidget(panel, PFUGUI_SLIDER_LOG, GUI_NEAR);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Near Clip");
    pfuWidgetRange(wid, 1, 
		   PF_MIN2(.1f, ViewState->sceneSize / 100.0f),
		   PF_MAX2(ViewState->near, ViewState->sceneSize),
		   ViewState->near);
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_NEAR] = wid;
    
    if (ViewState->guiFormat == GUI_HORIZONTAL)
    {
	xo += 2*(PFUGUI_BUTTON_VLONG_XINC);
	y = ys;
    }

    x = xo;
    ySize = PFUGUI_BUTTON_YSIZE;
    y -= PFUGUI_BUTTON_YINC;

    /* preset fogs */
    wid = pfuNewWidget(panel, PFUGUI_MENU_BUTTON, GUI_FOG);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Fog Type");
    pfuWidgetFunc(wid, controlPanel);
    if (ViewState->realityGraphics)
	pfuWidgetSelections(wid, fog_strs, fog_vals, NULL, 
	    sizeof(fog_strs)/sizeof(fog_strs[0]));
    else
	pfuWidgetSelections(wid, fog_strs, fog_vals, NULL, 
	    sizeof(fog_strs)/sizeof(fog_strs[0]) - 1);
    pfuWidgetDefaultValue(wid, ViewState->fog);
    ViewState->guiWidgets[GUI_FOG] = wid;
    
    y -= PFUGUI_SLIDER_YINC;
    ySize = PFUGUI_SLIDER_YSIZE;

    /* Slider for fog near range */
    wid = pfuNewWidget(panel, PFUGUI_SLIDER_LOG, GUI_NEAR_FOG_RANGE);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Near Fog");
    pfuWidgetRange(wid, 1, 0.0f, ViewState->far*8.0f, ViewState->nearFogRange);
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_NEAR_FOG_RANGE] = wid;

    y -= (PFUGUI_SLIDER_YINC );

    /* Slider for fog far range */
    wid = pfuNewWidget(panel, PFUGUI_SLIDER_LOG, GUI_FAR_FOG_RANGE);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Far Fog");
    pfuWidgetRange(wid, 1, 0.0f, ViewState->far*8.0f, ViewState->farFogRange);
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_FAR_FOG_RANGE] = wid;

    y -= (PFUGUI_SLIDER_YINC );

    /* Slider for Time of Day */
    wid = pfuNewWidget(panel, PFUGUI_SLIDER, GUI_TOD);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Time Of Day");
    pfuWidgetRange(wid, 1, 0.0f, 1.0f, 0.8f);
    pfuWidgetFunc(wid, controlPanel);
    ViewState->guiWidgets[GUI_TOD] = wid;
    
    ySize = PFUGUI_BUTTON_YSIZE;
    y -= (PFUGUI_BUTTON_YINC );
    
    /* cull mode */
    wid = pfuNewWidget(panel, PFUGUI_MENU_BUTTON, GUI_CULL);
    pfuWidgetDim(wid, x, y, xSize, ySize);
    pfuWidgetLabel(wid, "Cull Mode");
    pfuWidgetFunc(wid, controlPanel);
    pfuWidgetSelections(wid, Cull_strs, Cull_vals, NULL, 13);
    pfuWidgetDefaultValue(wid, ViewState->cullMode);
    ViewState->guiWidgets[GUI_CULL] = wid;
    
    
    if (ViewState->realityGraphics)
    {
	
	if (ViewState->guiFormat == GUI_HORIZONTAL)
	{
	    xo += 2*(PFUGUI_BUTTON_VLONG_XINC);
	    y = ys;
	}

	x = xo;
	y -= (PFUGUI_BUTTON_YINC );
	xSize = PFUGUI_BUTTON_VLONG_XSIZE;
	ySize = PFUGUI_BUTTON_YSIZE;
	/* toggle Anti-Aliasing*/
	wid = pfuNewWidget(panel, PFUGUI_SWITCH, GUI_AA);
	pfuWidgetLabel(wid,"Anti-Alias");
	pfuWidgetDim(wid, x, y, xSize, ySize);
	pfuWidgetFunc(wid, controlPanel);
    	pfuWidgetDefaultValue(wid, ViewState->aa);

	x += (PFUGUI_BUTTON_VLONG_XINC);

	/* LOD Fade button */
	wid = pfuNewWidget(panel, PFUGUI_SWITCH, GUI_LOD_FADE);
	pfuWidgetDim(wid, x, y, xSize, ySize);
	pfuWidgetLabel(wid, "Fade LOD");
	pfuWidgetFunc(wid, controlPanel);
    	pfuWidgetDefaultValue(wid, ViewState->fade);
	ViewState->guiWidgets[GUI_LOD_FADE] = wid;

	x = xo;
	y -= (PFUGUI_SLIDER_YINC);

	xSize = PFUGUI_BUTTON_VLONG_XSIZE +(PFUGUI_BUTTON_VLONG_XINC);
	ySize = PFUGUI_SLIDER_YSIZE;
	wid = pfuNewWidget(panel, PFUGUI_SLIDER, GUI_LOD_FADE_RANGE);
	pfuWidgetDim(wid, x, y, xSize, ySize);
	pfuWidgetLabel(wid, "Lod Fade Range");
	pfuWidgetRange(wid, 1, 0.0f, 1000.0f, ViewState->fadeRange);
	pfuWidgetFunc(wid, controlPanel);
	ViewState->guiWidgets[GUI_LOD_FADE_RANGE] = wid;
    }

    /* Reset GUI so that it will come up with the specified default values */
    pfuResetGUI();

    return(panel);
}
