/*
iscan.h
Darrell Anderson 5/96

ISCAN communication code.  Assumes sgi is connected to the ISCAN
via ISCAN_DEVICE (/dev/[rs232] defined below). 

The interface is modelled after that presented in flock.c, and polhemus.c
in order to be (somewhat) interchangable. */

#ifndef __ISCAN_H__
#define __ISCAN_H__

#define ISCAN_DEVICE "/dev/ttyd2"

#define ISCAN_BAUD 19200

#define ISCAN_TIMEOUT 0

/* defining this to 0 makes it work sometimes.. */
#define SYNC_SERIAL 1

/* the data format we request from the iscan: */
typedef struct iscan_record {
	
  /* record header info (filled in with 0x4444) */
  unsigned char header[2];

  /* eye position */
  short h,v;

} iscan_record;

/*----------------------------------------------------------------
initialize the ISCAN (called by P_iscan_Start)
----------------------------------------------------------------*/
void iscan_initialize(char *device, short baud, short timeout );

/*----------------------------------------------------------------
send numBytes bytes from (*data) to the iscan.
returns 0 on fail, 1 otherwise
----------------------------------------------------------------*/
int iscan_send( char *data, short numBytes );

/*----------------------------------------------------------------
read numBytes from the iscan, storing it in (*data)
returns 0 on fail, 1 otherwise
----------------------------------------------------------------*/
int iscan_read( char *data, short numBytes );

/*----------------------------------------------------------------
  fetch the 'freshest' complete record from the iscan
  ----------------------------------------------------------------*/
void iscan_fetch_record( iscan_record *data );

/*----------------------------------------------------------------
update global matrices, called as a pre-frame hook 
P_iscan_Start must have been previously called for this to work.
----------------------------------------------------------------*/
void iscan_c_hook(void);

/*----------------------------------------------------------------
master iscan initialization (and activation) funtion, called from  
scheme Iscan_Init must have been previously called for this to work.
----------------------------------------------------------------*/
Object P_iscan_Start(void);

/*----------------------------------------------------------------
master iscan kill(suspend really) function, called from scheme 
P_iscan_Start must have been previously called for this to work.
----------------------------------------------------------------*/
Object P_iscan_Stop(void);

/*----------------------------------------------------------------
make the global variables and above P_ functions available 
----------------------------------------------------------------*/
void iscan_init(void);

/*----------------------------------------------------------------
reverse the byte order in a word X86->'normal' (:
----------------------------------------------------------------*/
void reverse_byte_order2( unsigned char *word );

#endif /* __ISCAN_H__ */




