#include "kart.h"
#include "persia-scm.h"
#include "persia.h"

#include <stdio.h>
#include <unistd.h>
#include </usr/vision/include/analog.h>

#define MB_STEER_OFFSET		5000 /* 'straight ahead' val from kart */
#define SUCCESS			0
#define	FAIL			(-1)

static const int brakeChannel = 0;
static const int gasChannel = 1;
static const int steeringChannel = 2;

float kart_steering, kart_gas, kart_brake;

static analog_input_handle_t handle;

extern analog_input_handle_t analog_input_open(int offset);

#if DEBUG
#define debug_print(S) fprintf(stderr, S "\n")
#else
#define debug_print(S)
#endif

void
kart_initialize()
{
  int baseAddress = ANALOG_INPUT_ACTUAL_POSITION_BASE;

  debug_print("kart_initialize");
  if ((handle = analog_input_open(baseAddress)) == NULL) 
    {
      fprintf(stderr, "kart_initialize: analog_input_open() failed\n");
      exit(1);
    }
}

void
kart_get_data(float *s, float *g, float *b)
{
  *s = kart_steering;
  *g = kart_gas;
  *b = kart_brake;
}

void
kart_update_data1()
{
  debug_print("kart_update_data1");
  /* steering must be  a value between -1.0 and 1.0
     raw steering values: full left = 124, full right = 3774, center = 1953 */
  kart_steering = ANALOG_INPUT(handle, steeringChannel);
  kart_steering = -((kart_steering - 1953) / 1825);
  /* printf("steering = %f\n", eventData.steering); */
  if (kart_steering < -1.0)
    kart_steering = -1.0;
  else if (kart_steering > 1.0)
    kart_steering = 1.0;

  /* raw gas values: no gas = 3910, full gas = 1592 */
  kart_gas = ANALOG_INPUT(handle, gasChannel);
  kart_gas = -((kart_gas - 3910) / 2318);
  /* printf("gas = %f\n", eventData.gas); */
  if (kart_gas < 0.0)
    kart_gas = 0;
  else if (kart_gas > 1.0)
    kart_gas = 1.0;

  /* raw brake values: no brake = 2133, full brake = 299 */
  kart_brake = ANALOG_INPUT(handle, brakeChannel);
  kart_brake = -((kart_brake - 2133) / 1834);
  /*  printf("brake = %f\n", eventData.brake); */
  if (kart_brake < 0.0)
    kart_brake = 0.0;
  else if (kart_brake > 1.0)
    kart_brake = 1.0;
}

void
kart_update_data ()
{
  /* Set the Scheme kart variables */
  static int kart_initialized = 0;
  static int kart_last_frame_count = -1;

  debug_print("kart_update_data");
  if (!kart_initialized)
    {
      kart_initialize();
      kart_initialized = 1;
    }

  if (kart_last_frame_count != persia_frame_count)
    {
      debug_print("getting kart data\n");
      kart_update_data1();
      kart_last_frame_count = persia_frame_count;
    }
}

Object P_KartGas()
{
  if (!ViewState)
    Primitive_Error("ps:init-cameras not yet called");
  kart_update_data();
  return Make_Flonum(kart_gas);
}

Object P_KartBrake()
{
  if (!ViewState)
    Primitive_Error("ps:init-cameras not yet called");
  kart_update_data();
  return Make_Flonum(kart_brake);
}

Object P_KartSteering()
{
  if (!ViewState)
    Primitive_Error("start-graphics not yet called");
  kart_update_data();
  return Make_Flonum(kart_steering);
}

Object P_DoNothing()
{
  return False;
}

void kart_init ()
{
#if 0
  Define_Primitive(P_DoNothing, "ps:kart-gas", 0, 0, EVAL);
  Define_Primitive(P_DoNothing, "ps:kart-brake", 0, 0, EVAL);
  Define_Primitive(P_DoNothing, "ps:kart-steering", 0, 0, EVAL);
#else
  Define_Primitive(P_KartGas, "ps:kart-gas", 0, 0, EVAL);
  Define_Primitive(P_KartBrake, "ps:kart-brake", 0, 0, EVAL);
  Define_Primitive(P_KartSteering, "ps:kart-steering", 0, 0, EVAL);
#endif
}
