/* light.c - Scheme interface to pfLightSource functions */

#include "persia-scm.h"
#include "persia.h"

Object P_NewLSource ()
{
  return Make_Pfo((pfObject*) pfNewLSource());
}

Object P_LightAmbient (Object lsource, Object r, Object g, Object b)
{
  Check_Pfo_Class(lsource, PFTYPE_LIGHT);
  Check_Type(r, T_Flonum);
  Check_Type(g, T_Flonum);
  Check_Type(b, T_Flonum);
  
  pfLightAmbient(PFO(lsource), 
		 FLONUM(r)->val,
		 FLONUM(g)->val,
		 FLONUM(b)->val);
  return Void;
}

Object P_LightColor (Object lsource, Object r, Object g, Object b)
{
  Check_Pfo_Class(lsource, PFTYPE_LIGHT);
  Check_Type(r, T_Flonum);
  Check_Type(g, T_Flonum);
  Check_Type(b, T_Flonum);
  
  pfLightColor(PFO(lsource), 
	       FLONUM(r)->val,
	       FLONUM(g)->val,
	       FLONUM(b)->val);
  return Void;
}

Object P_SpotLightDir (Object lsource, Object x, Object y, Object z)
{
  Check_Pfo_Class(lsource, PFTYPE_LIGHT);
  Check_Type(x, T_Flonum);
  Check_Type(y, T_Flonum);
  Check_Type(z, T_Flonum);
  
  pfSpotLightDir(PFO(lsource), 
		 FLONUM(x)->val,
		 FLONUM(y)->val,
		 FLONUM(z)->val);
  return Void;
}

Object P_SpotLightCone (Object lsource, Object f1, Object f2)
{
  Check_Pfo_Class(lsource, PFTYPE_LIGHT);
  Check_Type(f1, T_Flonum);
  Check_Type(f2, T_Flonum);
  
  pfSpotLightCone(PFO(lsource), 
		  FLONUM(f1)->val,
		  FLONUM(f2)->val);
  return Void;
}

Object P_LightPos (Object lsource, Object x, Object y, Object z, Object w)
{
  Check_Pfo_Class(lsource, PFTYPE_LIGHT);
  Check_Type(x, T_Flonum);
  Check_Type(y, T_Flonum);
  Check_Type(z, T_Flonum);
  Check_Type(w, T_Flonum);
  
  pfLightPos(PFO(lsource), 
	     FLONUM(x)->val,
	     FLONUM(y)->val,
	     FLONUM(z)->val,
	     FLONUM(w)->val);
  return Void;
}

Object P_LightOn (Object lsource)
{
  Check_Pfo_Class(lsource, PFTYPE_LIGHT);
  pfLightOn(PFO(lsource));
  return Void;
}

Object P_LightOff (Object lsource)
{
  Check_Pfo_Class(lsource, PFTYPE_LIGHT);
  pfLightOff(PFO(lsource));
  return Void;
}

Object P_IsLightOn (Object lsource)
{
  Check_Pfo_Class(lsource, PFTYPE_LIGHT);
  return Make_Fixnum(pfIsLightOn(PFO(lsource)));
}

void light_init ()
{
  Define_Primitive(P_NewLSource, "pf:new-lsource", 0, 0, EVAL);
  Define_Primitive(P_LightAmbient, "pf:light-ambient", 4, 4, EVAL);
  Define_Primitive(P_LightColor, "pf:light-color", 4, 4, EVAL);
  Define_Primitive(P_SpotLightDir, "pf:spot-light-dir", 4, 4, EVAL);
  Define_Primitive(P_SpotLightCone, "pf:spot-light-cone", 3, 3, EVAL);
  Define_Primitive(P_LightPos, "pf:light-pos", 5, 5, EVAL);
  Define_Primitive(P_LightOn, "pf:light-on", 1, 1, EVAL);
  Define_Primitive(P_LightOff, "pf:light-off", 1, 1, EVAL);
  Define_Primitive(P_IsLightOn, "pf:is-light-on", 1, 1, EVAL);
}
