/* material.c - Scheme interface to pfMaterial functions */

#include "persia-scm.h"
#include "persia.h"

Object P_NewMtl ()
{
  return Make_Pfo((pfObject*) pfNewMtl(pfGetSharedArena()));
}

Object P_MtlAlpha (Object mtl, Object alpha)
{
  Check_Pfo_Class(mtl, PFTYPE_MATERIAL);
  Check_Type(alpha, T_Flonum);

  pfMtlAlpha((pfMaterial*) PFO(mtl), FLONUM(alpha)->val);
  return Void;
}

Object P_MtlShininess (Object mtl, Object shininess)
{
  Check_Pfo_Class(mtl, PFTYPE_MATERIAL);
  Check_Type(shininess, T_Flonum);

  pfMtlAlpha((pfMaterial*) PFO(mtl), FLONUM(shininess)->val);
  return Void;
}

Object P_MtlColor (Object mtl, Object color, Object r, Object g, Object b)
{
  Check_Pfo_Class(mtl, PFTYPE_MATERIAL);
  Check_Type(color, T_Fixnum);
  Check_Type(r, T_Flonum);
  Check_Type(g, T_Flonum);
  Check_Type(b, T_Flonum);

  pfMtlColor((pfMaterial*) PFO(mtl), 
	     FIXNUM(color),
	     FLONUM(r)->val,
	     FLONUM(g)->val,
	     FLONUM(b)->val);
  return Void;
}

Object P_ApplyMtl (Object mtl)
{
  Check_Pfo_Class(mtl, PFTYPE_MATERIAL);

  pfApplyMtl((pfMaterial*) PFO(mtl));
  return Void;
}

Object P_GetCurMtl (Object side)
{
  pfMaterial *m;
  long s;

  Check_Type(side, T_Fixnum);
  s = FIXNUM(side);
  fprintf(stderr, "about to getcurmtl %d\n", s);
  m = pfGetCurMtl(s);
  fprintf(stderr, "finished getcurmtl\n");

  return Make_Pfo((pfObject*) m);
}

static Object V_PFMTL_AMBIENT;
static Object V_PFMTL_DIFFUSE;
static Object V_PFMTL_EMISSION;
static Object V_PFMTL_SPECULAR;

static Object V_PFMTL_FRONT;
static Object V_PFMTL_BACK;

void material_init ()
{
  Define_Variable(&V_PFMTL_AMBIENT, "pf:*mtl-ambient*", 
		  Make_Fixnum(PFMTL_AMBIENT));
  Define_Variable(&V_PFMTL_DIFFUSE, "pf:*mtl-diffuse*",
		  Make_Fixnum(PFMTL_DIFFUSE));
  Define_Variable(&V_PFMTL_EMISSION, "pf:*mtl-emission*",
		  Make_Fixnum(PFMTL_EMISSION));
  Define_Variable(&V_PFMTL_SPECULAR, "pf:*mtl-specular*",
		  Make_Fixnum(PFMTL_SPECULAR));

  Define_Variable(&V_PFMTL_FRONT, "pf:*mtl-front*",
		  Make_Fixnum(PFMTL_FRONT));
  Define_Variable(&V_PFMTL_BACK, "pf:*mtl-back*",
		  Make_Fixnum(PFMTL_BACK));
#if 0
  This doesn't work because it ps:init-cameras hasn't yet been called.
  Define_Variable(&V_global_mtl, "ps:*default_mtl*",
		  Make_Pfo((pfObject*) pfGetCurMtl(PFMTL_FRONT)));
#endif

  Define_Primitive(P_NewMtl, "ps:new-mtl", 0, 0, EVAL);
  Define_Primitive(P_MtlAlpha, "pf:mtl-alpha", 2, 2, EVAL);
  Define_Primitive(P_MtlShininess, "pf:mtl-shininess", 2, 2, EVAL);
  Define_Primitive(P_MtlColor, "pf:mtl-color", 5, 5, EVAL);
  Define_Primitive(P_ApplyMtl, "pf:apply-mtl", 1, 1, EVAL);
  Define_Primitive(P_GetCurMtl, "pf:get-cur-mtl", 1, 1, EVAL);
}
