;;; New York Driving Obstacle Car Support

;;; Program to implement mouse-controlled driving down a 4-lane
;;;  freeway containing other cars at lower velocity.

;;; todo: 
;;;  car birth/death dependant on agent-car position *** done
;;;  fix road texture *** done????
;;;  get aerostar working, scaled, etc.... *** done
;;;  put cars on road, not grass *** done
;;;  make a function for converstion mph->m/s *** done
;;;  collison detection, do something when crash

;;; new todo:
;;;  make cars appear as type1 and type2 *** done
;;;  make cars appear independant of where they were previously.
;;;  fix road texture again (4 lane, not 2x2) *** done
;;;  increase frame-rate?
;;;  remove cars, make boxes (red ones). *** done

;;;  for-each-vec
;;;  This function will allow the simpler use of iteration up through
;;;  the vector applying the function "func" to each of the elements
;;;  of the vector.  Should be usable for applying functions to each
;;;  of the vectors representing a lane if the supplied vector is
;;;  "road".

(define for-each-vec
  (lambda (func vector)
    (begin
      (define loopend (vector-length vector))
      (define i 0)
      (do ((i 0 (+ i 1)))
	  ((= i loopend) vector)
	(func (vector-ref vector i)))
      )))

(define fast-car-prob .5)
(define new-car-prob 0.5)
(define double-car-prob 0.0)

(define frame-count 0)
(define honking 0)
(define agent-slow-crash 0)

;;; other cars
; obstacle-car-attributes:
;  (x-pos, y-pos, x-offset, y-offset, motion-func, geode, velocity, type)
;  (0    , 1    , 2       , 3       , 4          , 5    , 6       , 7)
(define num-obstacle-car-attributes 8)

(define obstacle-car-type1-velocity 8)
(define agent-car-velocity 16)
(define obstacle-car-type2-velocity 24)
(define obstacle-car-type1-width 2.4) ;;; was 2.9
(define obstacle-car-type1-length 3.35)
(define obstacle-car-type1-height 3.)

(define obstacle-car-type1-init-y 0.)
(define obstacle-car-type1-init-x 0.)

(define road-vec (make-vector 4 '()))

(vector-set! road-vec 0
	     (make-vector 17 (make-vector num-obstacle-car-attributes '()))) 
(vector-set! road-vec 1
	     (make-vector 17 (make-vector num-obstacle-car-attributes '()))) 
(vector-set! road-vec 2
	     (make-vector 17 (make-vector num-obstacle-car-attributes '()))) 
(vector-set! road-vec 3
	     (make-vector 17 (make-vector num-obstacle-car-attributes '()))) 


;;;; obstacle-car-motion-func
;  new function for generic motion along road.  It no longer does any
;  testing as all of the testing/removal is done using the 20th frame
;  traversal.  Just moves the cars along their forward paths...

(define (obstacle-car-motion-func atts)
  (vector-set! atts 1 (+ (vector-ref atts 1)
			 (* (vector-ref atts 6) delta-seconds))))

;; filename was "esprit.flt"
;; filename was "truck.flt"
;; filename was "truck_plain.obj"
;; filename was "dumptruck.obj"
(define cab-percent 0.35)
(define (ps:make-dcs-truck width depth height . attributes)
  (let* ((the-dcs (pf:new-dcs))
;	 (body
;	  (let
;	      ((body-group (pf:new-dcs))
;	       (chasis (apply ps:make-geode-cube 
;			      (append (list (* width (/ 7 8)) (* depth (- 1.0 cab-percent)) height)
;				      attributes))))
;	    (pf:add-child body-group chasis)
;	    body-group))
	 (body (apply ps:make-geode-cube
		      (append (list (* width (/ 7 8))
				    (* depth (- 1.0 cab-percent)) height)
			      attributes)))
	 (wheel-width (/ width 8))
	 (wheel-radius (/ height 3))
	 (wheel-to-end (/ depth 16))
	 (body-clearance (* wheel-radius .7))
	 (wheel (ps:make-geode-cube
		 wheel-width (/ height 3) (/ height 3)
		 (cons 'gstate (ps:make-gstate '(color . #(.0 .0 .0))))))
	 (cab-inset-width (/ width 8))
	 (cab-height (* height .6))
	 (cab (apply ps:make-geode-cube
		     (append (list (- (* width (/ 7 8)) (* 2 cab-inset-width))
				   (* depth cab-percent)
				   cab-height)
			     attributes)))
	 (windshield (ps:make-geode-plane
		      (- (- (* width (/ 7 8)) (* 2 cab-inset-width)) .1)
;		      10.
		      (* cab-height .3)
		      (cons 'gstate re-white)))
	 (body-dcs (ps:place-node body
				  (- (/ width 2))
				  (- (/ depth 2))
				  body-clearance
				  the-dcs))
	 (windshield-dcs (ps:place-node windshield
					(+ cab-inset-width .05)
					(+ depth 0.03)
					(- cab-height .15)
					body-dcs)))
    (ps:place-node cab
		   cab-inset-width 
		   (- depth (* depth cab-percent))
;		   body-clearance body-dcs)
		   0. body-dcs)
    (pf:dcs-rot windshield-dcs 0. -90. 0.)
    (ps:place-node wheel 
		   (+ cab-inset-width (/ wheel-width -2))
;		   (/ wheel-width -2) 
		   wheel-to-end
		   (- body-clearance)
		   body-dcs)
    (ps:place-node wheel
		   (+ (- (* width (/ 7 8)) cab-inset-width) (/ wheel-width -2))
;		   (+ (* width (/ 7 8)) (/ wheel-width -2)) 
		   wheel-to-end
		   (- body-clearance)
		   body-dcs)
    (ps:place-node wheel
		   (+ cab-inset-width (/ wheel-width -2))
		   (- depth
		      wheel-to-end wheel-radius)
		   (- body-clearance)
		   body-dcs)
    (ps:place-node wheel
		   (+ (- (* width (/ 7 8)) cab-inset-width) (/ wheel-width -2))
		   (- depth
		      wheel-to-end wheel-radius)
		   (- body-clearance)
		   body-dcs)
    the-dcs))

(define red-box
  (ps:make-geode-cube 12. .1 3.
		      (cons 'gstate red)))

(define esprit-geode-cache #f)
(define (get-esprit-dcs)
  (let ((the-geode (begin (if (not esprit-geode-cache)
			      (set! esprit-geode-cache
				    (ps:make-node-from-file 
				     "dumptruck.obj" red)))
			  esprit-geode-cache))
	(the-dcs (pf:new-dcs))
	(scale 1.))
;    (pf:add-child the-dcs red-box)
    (pf:add-child the-dcs the-geode)
    (pf:dcs-rot the-dcs 90. 0. 0.)
;    (pf:dcs-rot the-dcs 90. 90. 0.)
    (pf:dcs-scale the-dcs 1.)
    (pf:dcs-trans the-dcs (* scale 1.) 0. 1.)
    the-dcs))

(define (get-esprit-dcs color)
  (ps:make-dcs-truck obstacle-car-type1-width obstacle-car-type1-length obstacle-car-type1-height (cons 'gstate color)))

(define (type1-body color)
  (if doing-car-file
      (cond (doing-aerostar
	     (let ((aero-body (ps:make-node-from-file car-file))
		   (adj-dcs (pf:new-dcs)); adjustment dcs
		   (scale .035)) 
	       (pf:add-child adj-dcs aero-body)
	       (pf:dcs-rot adj-dcs 90. 0. 0.)
	       (pf:dcs-scale adj-dcs scale)
	       (pf:dcs-trans adj-dcs (* -561. scale) (* 2964.5 scale)
			     (* 381. scale))
	       adj-dcs))
	    (doing-esprit
	     (get-esprit-dcs color)))
      (ps:make-geode-cube car-width car-depth obstacle-car-type1-height
			  (cons 'gstate color))))

(define car-body-red (type1-body red))
(define car-body-blue (type1-body blue))
(define car-body-yellow (type1-body yellow))
(define car-body-white (type1-body white))
(define car-body-gray (type1-body gray))
(define car-body-tan (type1-body tan))

;;; This should take color as an argument, because color needs
;;; to be in the attribute vector.
(define (car-with-random-color)
  (let ((color-int (random-int 0 5)))
    (cond
     ((= color-int 0) car-body-red)
     ((= color-int 1) car-body-blue)
     ((= color-int 2) car-body-yellow)
     ((= color-int 3) car-body-white)
     ((= color-int 4) car-body-gray)
     (#t car-body-tan))))

(define (obstacle-car-type1-make) ; create the car's body
  (ps:place-node (car-with-random-color) 0. 0. 0.))

(define car-type1 obstacle-car-type1-velocity)
(define car-type2 obstacle-car-type2-velocity)

(define (obstacle-car-type1-car-new)
  (let ((attributes-vec (make-vector num-obstacle-car-attributes)))
    (vector-set! attributes-vec 0 obstacle-car-type1-init-x)
    (vector-set! attributes-vec 1 obstacle-car-type1-init-y)
    (vector-set! attributes-vec 2 (/ obstacle-car-type1-width 2))
    (vector-set! attributes-vec 3 (/ obstacle-car-type1-length 2))
    (vector-set! attributes-vec 4 obstacle-car-motion-func)
    (vector-set! attributes-vec 5 (obstacle-car-type1-make))
    (vector-set! attributes-vec 6 obstacle-car-type1-velocity)
    (vector-set! attributes-vec 7 car-type1)
    attributes-vec))

(define (obstacle-car-type2-car-new)
  (let ((attributes-vec (make-vector num-obstacle-car-attributes)))
    (vector-set! attributes-vec 0 obstacle-car-type1-init-x)
    (vector-set! attributes-vec 1 obstacle-car-type1-init-y)
    (vector-set! attributes-vec 2 (/ obstacle-car-type1-width 2))
    (vector-set! attributes-vec 3 (/ obstacle-car-type1-length 2))
    (vector-set! attributes-vec 4 obstacle-car-motion-func)
    (vector-set! attributes-vec 5 (obstacle-car-type1-make))
    (vector-set! attributes-vec 6 obstacle-car-type2-velocity)
    (vector-set! attributes-vec 7 car-type2)
   attributes-vec))

;;;;
;;;;
;;;;
;;;;

;;;; agent-car-move
;;
;;     Note that this will not work with non integer lane locations.
;;
(define agent-car-move
  (lambda (atts)
    (begin
;Forward motion first
      (vector-set! atts 1 (+ (vector-ref atts 1)
			     (* (vector-ref atts 6) delta-seconds)))
      ;Then horizontal motion
      (define move-per-frame (/ lane-width (- (/ 1 delta-seconds) 1)))
      (cond
       ((= curr-offset 0) '())
       ((> curr-offset 0)
	(cond
	 ((> curr-offset move-per-frame)
	  (begin
	    (set! curr-offset (- curr-offset move-per-frame))
	    (pf:get-dcs-mat agent-car-dcs ps:tmp-pfmatrix)
	    (vector-set!
	     atts 0
	     (+ (ps:pfmatrix-ref ps:tmp-pfmatrix 3 0) move-per-frame))))
	 ((< curr-offset move-per-frame)
	  (begin
	    (pf:get-dcs-mat agent-car-dcs ps:tmp-pfmatrix)
	    (vector-set!
	     atts 0
	     (exact->inexact (inexact->exact
			      (- (ps:pfmatrix-ref ps:tmp-pfmatrix 3 0)
				 (abs curr-offset)))))
	    (set! curr-offset 0)))))
       ((< curr-offset 0)
	(cond
	 ((> (abs curr-offset) move-per-frame)
	  (begin
	    (set! curr-offset (+ curr-offset move-per-frame))
	    (pf:get-dcs-mat agent-car-dcs ps:tmp-pfmatrix)
	    (vector-set!
	     atts 0
	     (- (ps:pfmatrix-ref ps:tmp-pfmatrix 3 0)
		move-per-frame))))
	 ((< (abs curr-offset) move-per-frame)
	  (begin
	    (pf:get-dcs-mat agent-car-dcs ps:tmp-pfmatrix)
	    (vector-set!
	     atts 0
	     (exact->inexact (inexact->exact
			      (- (ps:pfmatrix-ref ps:tmp-pfmatrix 3 0)
				 (abs curr-offset)))))
	    (set! curr-offset 0)))))))))

(define (agent-car-attr)
  (let
      ((attributes-vec
	(make-vector num-obstacle-car-attributes)))
    (vector-set! attributes-vec 0 agent-car-x)
    (vector-set! attributes-vec 1 agent-car-y)
    (vector-set! attributes-vec 2 (/ obstacle-car-type1-width 2))
    (vector-set! attributes-vec 3 (/ obstacle-car-type1-length 2))
    (vector-set! attributes-vec 4 agent-car-move)
    (vector-set! attributes-vec 5 agent-car-dcs)
    (vector-set! attributes-vec 6 agent-car-velocity)
    (vector-set! attributes-vec 7 agent-car-velocity)
    attributes-vec))

(define agent-car-attr-list (agent-car-attr))

;;;; This sets the agent car's initial lane.  (X coordinate.)
(vector-set! agent-car-attr-list 0 7.)

;;; Replace the following field offsets with a agent-car offset

(define slow-car-create-y
  (lambda ()
    (+ (vector-ref agent-car-attr-list 1) 72)))

(define-macro fast-car-create-y
  (lambda ()
    (- (vector-ref agent-car-attr-list 1) 72)))

(define (which-lane speed)
  (let ((lane-chose (random-int 0 9)))
    (cond
     ((= speed 1)
      (cond
       ((< lane-chose 4) 3)
       ((< lane-chose 7) 2)
       ((< lane-chose 9) 1)
       (#t 0)))
     ((= speed -1)
      (cond
       ((< lane-chose 4) 0)
       ((< lane-chose 7) 1)
       ((< lane-chose 9) 2)
       (#t 3))))))

;;; new-fast-car
;;; looks in all four lanes for a blank spot to put the car.  If none
;;; found, doesn't create a car.  Makes sense, no?  actually checks
;;; four random lanes, not all of them...

(define (new-fast-car passed recurs)
  (if (> 20 recurs)
      (if (null? (vector-ref (vector-ref
			      (vector-ref road-vec passed) 0) 5))
	  (begin
	    (vector-set! (vector-ref road-vec passed)
			 0 (obstacle-car-type2-car-new))
	    (vector-set! (vector-ref (vector-ref road-vec passed) 0) 0
			 (exact->inexact (- 7. (* 4 passed))))
	    (vector-set! (vector-ref (vector-ref road-vec passed) 0) 1
			 (fast-car-create-y)))
	  (new-fast-car (which-lane 1) (+ 1 recurs)))
      '()))

;;; new-slow-car
;;; looks in all four lanes for a blank spot to put the car.  If none
;;; found, doesn't create a car.  Makes sense, no?

(define new-slow-car
  (lambda (passed recurs)
    (if (> 20 recurs)
	(if (null? (vector-ref (vector-ref (vector-ref road-vec
						       passed) 16) 5))
	    (begin
	      (vector-set! (vector-ref road-vec passed)
			   16 (obstacle-car-type1-car-new))
	      (vector-set! (vector-ref (vector-ref road-vec passed) 16) 0
			   (exact->inexact (- 7. (* 4 passed))))
	      (vector-set! (vector-ref (vector-ref road-vec passed) 16) 1
			   (slow-car-create-y)))
	    (new-slow-car (which-lane -1) (+ 1 recurs)))
	'())))

(define new-car
  (lambda ()
    (if (< (random-float 0.0 1.0) fast-car-prob)
	(new-fast-car (which-lane 1) 0)
	(new-slow-car (which-lane -1) 0))
  ))

(define add-cars
  (lambda ()
    (if (<= (random-float 0.0 1.0) new-car-prob)
	(if (<= (random-float 0.0 1.0) double-car-prob)
	    (begin
	      (new-car)
	      (new-car)))
	(new-car))))

;;; The following function is intended for use in 19 out of every 20
;;; frames.  It should result in a faster frame rate, due to its
;;; increased simplicity as compared to the 20th frame refresh engine,
;;; which follows.
(define (motion curr-car)
  (if
   (not (null? (vector-ref curr-car 4)))
   (begin
     ((vector-ref curr-car 4) curr-car)
     (pf:dcs-trans (vector-ref curr-car 5)
		   (vector-ref curr-car 0)
		   (vector-ref curr-car 1)
		   0.))))

(define (update-obstacle-car-vec car-vec)
  (for-each-vec motion car-vec))

;;;;
;;;;
;;;;
;;;;

;;; update field, obstacle car, positions
(ps:add-pre-frame-hook
 (lambda ()
   (let ((agent-y (begin
		    (pf:get-dcs-mat agent-car-dcs ps:tmp-pfmatrix)
		    (ps:pfmatrix-ref ps:tmp-pfmatrix 3 1))))
     (if (> (- agent-y field-y) cycle-pt)
	 (begin                    ; move field forward
	   (format #t "Cycling, frames: ~s\n" frame-count)
	   (set! field-y (- agent-y (/ cycle-pt 2)))
	   (pf:dcs-trans field field-x field-y field-z))))))

(define curr-offset 0)

(define agent-car-trans
  (lambda (offset)
    (cond
     ((= offset lane-width)
      (set! curr-offset offset))
     ((= offset (- 0 lane-width))
      (set! curr-offset offset))
     ((= offset (+ road-xpos 2))
      (set! curr-offset offset)))))

;;; The following function is intended for use in 19 out of every 20
;;; frames.  It should result in a faster frame rate, due to its
;;; increased simplicity as compared to the 20th frame refresh engine,
;;; which follows.

(define update-decide-obstacle-cars
  (lambda ()
    (begin
      (for-each-vec fast-car-update road-vec)
      (for-each-vec slow-car-update road-vec)
    )))


;;; All collision detection for the agent is done here.  Updated every
;;; 20 frames (one tick).  This catches sideswipes as well as
;;; collisions with slow cars from behind.  Removed a special case
;;; from the slow car motion algorithm.

(define find-crash
  (lambda ()
    (if
     (not (null? (vector-ref
		  (vector-ref
		   (vector-ref road-vec (lane-no agent-car-attr-list))
		   8)
		  7)))
     (set! agent-slow-crash 1)
     '())))

;;;; one-tick-update
;  Top-level function in charge of updating data-structures every 20th frame.
;  Order of the function calls is _very_ important.

(define (one-tick-update)
  (find-crash)
  (format #t "Perc-vector: ~s Reward: ~s\n"
	  (calc-perc) (calc-award))
  (action-rlkit (calc-perc) (calc-award))
  (set! honking 0)
  (set! agent-slow-crash 0)
  (update-decide-obstacle-cars)
  (add-cars))

;;;; lane-no
;  Handy function to determine which lane a given vector is located in.
;  Lanes are ordered as:  3 2 1 0

(define (lane-no attr-list)
  (inexact->exact (/ (- 6 (vector-ref attr-list 0)) 4)))

;;;; general-motion
;  Called each frame, it decides whether to call a generic update
;  function, or to call the specific function for updating data
;  structures.

;;;; 2up
;  A quick function that acts as a bounds check for the
;  fast-car-update function, when dealing with large increments.

(define 2up
  (lambda (i)
    (if
     (< i 15)
     (+ i 2)
     (+ i 1))))

;;;; change-loc
;  change-loc was a means to increasing the readability of the slow
;  and fast car update functions by taking out repetitive bounds
;  checks and common code out.  Also acts to remove those obstacle
;  cars that have moved out of visual range.

(define change-loc
  (lambda (curr lane-vec change)
    (cond
     ((= change 1)
      (cond
       ((= curr 16)
	(begin
	  (pf:remove-child ps:*default-parent*
			   (vector-ref (vector-ref lane-vec curr) 5))
	  (vector-set! lane-vec curr
		       (make-vector num-obstacle-car-attributes '()))))
       (#t
	(begin
	  (vector-set! lane-vec (+ curr change) (vector-ref lane-vec curr))
	  (vector-set! lane-vec curr
		     (make-vector num-obstacle-car-attributes '()))))))
     ((= change -1)
      (cond
       ((= curr 0)
	(begin
	  (pf:remove-child ps:*default-parent*
			   (vector-ref (vector-ref lane-vec curr) 5))
	  (vector-set! lane-vec curr
		       (make-vector num-obstacle-car-attributes '()))))
       (#t
	(begin
	  (vector-set! lane-vec (+ curr change) (vector-ref lane-vec curr))
	  (vector-set! lane-vec curr
		       (make-vector num-obstacle-car-attributes '()))))))
     (#t '())
     )))

;;;; slow-car-update
;  slow-car-update is a rather simple function, only serving to move
;  slow cars through the data-structure.  Since the logic for all
;  other motion anticipates the motion here, there is not much need
;  for error checking or special cases.

(define slow-car-update
  (lambda (lane-vec)
    (begin
      (if
       (and
	(not (null? (vector-ref (vector-ref lane-vec 0) 0)))
	(= (vector-ref (vector-ref lane-vec 0) 7) car-type1))
       (change-loc 0 lane-vec -1)
       '())
      (do ((i 1 (+ i 1)))
	  ((> i 16) lane-vec)
	(if
	 (and
	  (not (null? (vector-ref (vector-ref lane-vec i) 0)))
	  (= (vector-ref (vector-ref lane-vec i) 7) car-type1)
	  (null? (vector-ref (vector-ref lane-vec (- i 1)) 0)))
	 (change-loc i lane-vec -1)
	 '()))
      )))

;;; Only do initial loop from 15 to 0, since 16 gets removed and is a
;;; special case.  Since the looping is done top down, no fast car can
;;; get stuck behind other fast car unless other fast car is stuck
;;; behind agent.  Need to make sure that forward searches don't go
;;; out of range.  Done.

; Potentially the most complex and confusing function in the entire
; endeavor to date, fast-car-update probably has several hidden bugs
; and almost definetly has room for rethinking the logic, but it
; appears to work for now.  (Ie: no known bugs... it may have
; pointless statements though...)

(define fast-car-update
  (lambda (lane-vec)
    (begin
      (if
       (and (not (null? (vector-ref (vector-ref lane-vec 16) 0)))
	    (= (vector-ref (vector-ref lane-vec 16) 7) car-type2))
       (change-loc 16 lane-vec 1)
       '())
      (do ((i 15 (- i 1)))
	  ((< i 0) lane-vec)
	(if
	 (and
	  (not (null? (vector-ref (vector-ref lane-vec i) 0)))
	  (= (vector-ref (vector-ref lane-vec i) 7) car-type2))
	 (cond
	  ((and
	    (= i 7)
	    (=
	     (lane-no (vector-ref lane-vec i))
	     (lane-no agent-car-attr-list))
	    (null? (vector-ref (vector-ref lane-vec (+ i 1)) 0)))
	   (begin
	     (set! honking 1)
	     (vector-set! (vector-ref lane-vec i) 6
			  agent-car-velocity)))
     ;;; Can only get fast car in front of fast car if first car is
     ;;;     blocked by the agent car, so change to agent car speed.
	  ((and
	    (not (null? (vector-ref (vector-ref lane-vec (+ i 1)) 0)))
	    (= (vector-ref (vector-ref lane-vec (+ i 1)) 7) car-type2))
	   (vector-set! (vector-ref lane-vec i) 6
			agent-car-velocity))
	  ((and
	    (not (null? (vector-ref (vector-ref lane-vec (+ i 1)) 0)))
	    (= (vector-ref (vector-ref lane-vec (+ i 1)) 7) car-type1))
	   (begin
	     (vector-set! (vector-ref lane-vec i) 6 
			  obstacle-car-type1-velocity)
	     (vector-set! (vector-ref lane-vec i) 7
			  car-type1)))
	  ((null? (vector-ref (vector-ref lane-vec (+ i 1)) 0))
	   (cond
	    ((null? (vector-ref (vector-ref lane-vec (2up i)) 0))
	     (begin
	       (vector-set! (vector-ref lane-vec i) 6
			    obstacle-car-type2-velocity)
	       (change-loc i lane-vec 1)))
	    ((= (vector-ref (vector-ref lane-vec (2up i)) 7)
		car-type2)
	     (cond
	      ((= (vector-ref (vector-ref lane-vec (2up i)) 6)
		  obstacle-car-type2-velocity)
	       (begin
		 (vector-set! (vector-ref lane-vec i) 6
			      obstacle-car-type2-velocity)
		 (change-loc i lane-vec 1)))
	      ((= (vector-ref (vector-ref lane-vec (2up i)) 6)
		  agent-car-velocity)
	       (begin
		 (vector-set! (vector-ref lane-vec i) 6
			      obstacle-car-type2-velocity)
		 (change-loc i lane-vec 1)))
	      ))
	    ((= (vector-ref (vector-ref lane-vec (2up i)) 7)
		car-type1)
	     (vector-set! (vector-ref lane-vec i) 6
			  agent-car-velocity))
	    ((= (vector-ref (vector-ref lane-vec (2up i)) 7)
		agent-car-velocity)
	     (change-loc i lane-vec 1))
	  ))))))))


;;; The next function can be removed when the driving package has
;;; been modified to do only constant forward motion, and no longer
;;; uses mouse input for driving.

(define (ps:update-car-with-mouse v accel-control)
  (begin
    (ps:update-vehicle-vector
     v
     0.0
     0.0
     0.1
     ps:*delta-seconds*)))

(provide 'nyd-cars)
