;;; New York Driving Sensory-Motor system for agent

(define using-utpolicy #f)

(if using-utpolicy
    (begin
      (ps:dlopen "policy.so")
      (ps:policy-init)))

;;; pass in a vector of integers indicating perception,
;;; and a float for reward
;;; return an integer indicating action
(define (rl-choose-action new-vector prev-reward)
  (if using-utpolicy
      (ps:policy-choose-action new-vector 0.0)
      (random-int 0 4)))

;;; Start eye processing
(if (not making-snapshot) 
    (begin
      (ps:eye-start)
      (ps:eye-set-camera 0 second-driver-headrest)))
;(display second-driver-headrest)
;(ps:add-pre-frame-hook
; (lambda ()
;   (ps:eye-center-head 0.5)))

;;; Add ray
(define eye-ray-width 0.2)
(define eye-ray-geode
  (ps:make-geode-plane eye-ray-width 200.
		       (cons 'gstate (ps:make-gstate '(color . #(1. 1. 1.))))))
(define eye-ray-dcs
  (ps:place-node eye-ray-geode 
		 (- (/ eye-ray-width 2)) 0. 40.
		 second-driver-headrest))

(define rand-lane-color
  (lambda ()
    (+ (random-int 0 3) 2)))

;;;; find-obj
;  Starts one away from agent, follows gaze to horizon looking for
;  objects.  If no objects are found before the last location is
;  reached, the last location is then sent on to the make-vector-list
;  function.  make-vector-list then returns the proper (mostly filled
;  in) percept vector.  This vector can then be passed out.

(define find-obj
  (lambda (lane)
    (if
     (or (= lane -1)
	 (= lane 4))
     (begin
       (define out-vec (make-vector 8))
       (vector-set! out-vec 1 2)     ;;; Shoulder
       (vector-set! out-vec 4 (if (= for-back 1) 0 1))
       (vector-set! out-vec 5 2)     ;;; Far
       (vector-set! out-vec 6 0)     ;;; Far-half (horizon)
       (vector-set! out-vec 7 (rand-lane-color))   ;;; Lanes have random colors
       out-vec)
     (do
	 ((location (+ 8 for-back) (+ location for-back))
	  (object
	   (vector-ref 
	    (vector-ref (vector-ref road-vec lane) (+ 8 for-back)) 0)
	   (cond
	    ((> 17 location)
	     (vector-ref (vector-ref road-vec lane) location))
	    ((= 17 location)
	     '())))
	  )
	 ((or (or (= location 16)
		  (= location 0))
	      (not (null?
		    (vector-ref
		     (vector-ref
		      (vector-ref road-vec lane) location) 0))))
	  (make-vector-list
	   (list
	    (vector-ref (vector-ref road-vec lane) location) location)))
       '()))))

;;;; make-vector-list
;  This function creates a percept vector from a lane-location and an
;  integer representing the current location in the lane.

(define make-vector-list
  (lambda (in-list)
    (begin
      (define out-vec (make-vector 8))
      (vector-set! out-vec 1
		   (if (null? (vector-ref (car in-list) 0))
		       0 1))
      (vector-set! out-vec 4
		   (if (null? (vector-ref (car in-list) 0))
		       (if (= for-back 1)
			   0 1)
		       (if (> (cadr in-list) 8)
			   (if (= (vector-ref (car in-list) 6) 8)
			       0 1)
			   (if (= (vector-ref (car in-list) 6) 8)
			       1 0))))
      (vector-set! out-vec 5
		   (cond
		    ((< (abs (- 8 (cadr in-list))) 3)
		     0)
		    ((< (abs (- 8 (cadr in-list))) 5)
		     1)
		    (#t 2)))
      (vector-set! out-vec 6
		   (cond
		    ((member (abs (- 8 (cadr in-list))) '(1 3 5 6))
		     1)
		    (#t 0)))
      (vector-set! out-vec 7
		   (if (null? (vector-ref (car in-list) 0))
		       (rand-lane-color)   ;;; Lanes have random colors
		       (find-color in-list)))                 ;;; All cars are currently red.
      out-vec)))

(define find-color
  (lambda (in-list)
    (begin
      (define car-geode (pf:get-child (vector-ref (car in-list) 5) 0))
      (cond
       ((equal? car-geode car-body-red) 0)
       ((equal? car-geode car-body-blue) 1)
       ((equal? car-geode car-body-yellow) 2)
       ((equal? car-geode car-body-white) 3)
       ((equal? car-geode car-body-gray) 4)
       (#t 5)))))

(define make-vector-list-screwed-up-by-andrew
  (lambda (in-list)
    (begin
      (define out-vec (make-vector 8))
      ;; hear horn
      ;;(vector-set! out-vec 7 0)
      ;; gaze object
      (vector-set! out-vec 6
		   (if (null? (vector-ref (car in-list) 0))
		       0 1))
      ;; gaze side
      ;;(vector-set! out-vec 5 0)
      ;; gaze direction
      ;;(vector-set! out-vec 4 0)
      ;; gaze speed
      (vector-set! out-vec 3
		   (if (null? (vector-ref (car in-list) 0))
		       0
		       (if (> (cadr in-list) 8)
			   (if (= (vector-ref (car in-list) 6) 8)
			       0 1)
			   (if (= (vector-ref (car in-list) 6) 8)
			       1 0))))
      ;; gaze distance
      (vector-set! out-vec 2
		   (cond
		    ((< (abs (- 8 (cadr in-list))) 3)
		     0)
		    ((< (abs (- 8 (cadr in-list))) 5)
		     1)
		    (#t 2)))
      ;; gaze distance refined
      (vector-set! out-vec 1
		   (cond
		    ((member (abs (- 8 (cadr in-list))) '(1 3 5 6))
		     1)
		    (#t 0)))
      ;; gaze color
      (vector-set! out-vec 0
		   (if (null? (vector-ref (car in-list) 0))
		       (rand-lane-color)   ;;; Lanes have random colors
		       0))                 ;;; All cars are currently red.

      out-vec)))

;;;; calc-perc
;  calc-perc creates the complete perception vector by using some
;  global variables as well as a call to find-obj.

(define calc-perc
  (lambda ()
    (begin
      (define out-vec (make-vector 8))
      (set! out-vec (find-obj (+ (lane-no agent-car-attr-list) looking)))
      (vector-set! out-vec 0 honking)
      (vector-set! out-vec 2 (cond ((= looking 1) 0)
				   ((= looking 0) 1)
				   (#t 2)))
      (vector-set! out-vec 3 (if (> for-back 0) 0 1))
      out-vec)))

;;;; calc-award
;  Meant to be called every 20 frames, calc-award returns:
;    -10 if collision and no honking
;    -11 if collision and honking
;    -1 if no collision and honking
;    0.1 if no collision and no honking

(define calc-award
  (lambda ()
    (if
     (or (= agent-slow-crash 1)
	 (= honking 1))
     (+ (* -10 agent-slow-crash) (* -1 honking))
     0.1)))

;;;; action-rlkit
;  Interprets the meaning of the single integer returned by rl-choose-action.

(define action-rlkit
 (lambda (perc-vector prev-award)
   (begin
     (define curr-action (rl-choose-action perc-vector prev-award))
     (cond
      ((= curr-action 0)
       (gaze-left))
      ((= curr-action 1)
       (gaze-forward))
      ((= curr-action 2)
       (gaze-right))
      ((= curr-action 3)
       (gaze-backward))
      ((= curr-action 4)
       (move-to-gaze-lane))))))

(define lane-angle 5.)
(define gaze-heading-dest 0.)
(define gaze-heading-source 0.)

(define (gaze-left)    ;;;; Means gaze-FORWARD-left
  (set! looking 1)
  (set! for-back 1)
  (set! gaze-heading-source gaze-heading-dest)
  (set! gaze-heading-dest lane-angle))

(define (gaze-right)   ;;;; Means gaze-FORWARD-right
  (set! looking -1)
  (set! for-back 1)
  (set! gaze-heading-source gaze-heading-dest)
  (set! gaze-heading-dest (- lane-angle)))

(define (gaze-forward)
  (set! looking 0)
  (set! for-back 1)
  (set! gaze-heading-source gaze-heading-dest)
  (set! gaze-heading-dest 0.))

(define (gaze-backward)
  (set! gaze-heading-source gaze-heading-dest)
  (set! for-back -1)
  (cond
   ((= looking -1)
    (set! gaze-heading-dest (- lane-angle 180.)))
   ((= looking 1)
    (set! gaze-heading-dest (- 180. lane-angle)))
   ((= looking 0)
    (set! gaze-heading-dest 180.))))

(define sacade-frames 12)
(define (do-sacade)
  (let* ((alpha1 (modulo frame-count 20))
	 (alpha (if (<= alpha1 sacade-frames) (/ alpha1 sacade-frames) 1.)))
    (pf:dcs-rot second-driver-headrest 
		(+ (* gaze-heading-source (- 1. alpha))
		   (* gaze-heading-dest alpha))
		0. 0.)))

(define move-to-gaze-lane
  (lambda ()
    (cond
     ((= looking -1)
      (move-right))
     ((= looking 1)
      (move-left)))
    (gaze-forward)))

(define move-left
  (lambda ()
    (begin
      (if
       (< (lane-no agent-car-attr-list) 3)
       (agent-car-trans (- 0 lane-width))
       (set! agent-slow-crash 1)))))

(define move-right
  (lambda ()
    (begin
      (if
       (> (lane-no agent-car-attr-list) 0)
       (agent-car-trans lane-width)
       (set! agent-slow-crash 1)))))

;;;  The first cond won't be needed for the final implementation, due
;;;  to the fact that the off-set change will only be usable every 20
;;;  frames (when the motion has completed)



;;; if in doubt, consult /u/mccallum/research/src/rlkit/HighwayWorldSM.[hm]

;;; reward
;;; -10 if overlapping another car, or if just tried to go to shoulder
;;; -1  if getting beeped at
;;; .1  otherwise

;;; 0 GAZE_FORWARD_LEFT,
;;; 1 GAZE_FORWARD_CENTER,
;;; 2 GAZE_FORWARD_RIGHT,
;;; 3 GAZE_BACKWARD,    Look back in lane currently looking in
;;; 4 GO_TO_GAZE_LANE,  This implies gaze-forward-center afterwards 

;;; 7 hear horn (yes, no)
;;; 6 gaze object (road, car, shoulder)
;;; 5 gaze side (left, center, right)
;;; 4 gaze direction (forward, backward)
;;; 3 gaze speed (looming, receding)
;;; 2 gaze distance (nose, near, far)
;;; 1 gaze distance refined (close-half, far-half)
;;; 0 gaze color ()

(provide 'nyd-sm)
