;;; New York Driving field and road creation

;;; globals
(define field-y 0.)               ; will move when agent gets to edge
(define looking 0)
(define for-back 1)

(define doing-vr #f)
(define doing-second-car #t)
(define doing-vulture #t)
;(define doing-vulture #f)
(if making-snapshot
    (define vulture-window-position '(525 500 150 450))
    (define vulture-window-position '(1150 10 150 450)))
(define second-car-x 0.)
(define second-car-y 0.)
(if making-snapshot
    (define second-car-window-position '(1 500 525 450))
    (define second-car-window-position '(650 10 499 450)))
(define second-car-color '#(.5 .1 .5))
(define second-car-color '#(.5 .5 .5))
(define first-car-x 0.)
(define first-car-y 0.)
(define use-default-field #f)
(define car-width 1.5)
(define car-depth 2.5)
(define car-height .75)
(define doing-car-file #t)
(define doing-aerostar #f)        ; note BIG bottleneck (approx. 84%)
;(define car-file "aerostar1.1.obj")
(define doing-esprit #t)
(define car-file "esprit.flt")

(define-macro delta-seconds (/ 1. 20))
;; with update based on time:
;;(define-macro delta-seconds ps:*delta-seconds*)

(require 'driving)
;(ps:set-fog #t)

;;(ps:dlopen "mouse_down.so")       ; mouse-button routines

(define agent-car-dcs-mat (ps:make-pfmatrix))

(define-macro agent-car-dcs (if doing-second-car 
				second-car-dcs
				first-car-dcs))

(pf:get-dcs-mat agent-car-dcs agent-car-dcs-mat)

(define agent-car-x (ps:pfmatrix-ref agent-car-dcs-mat 3 0))
(define agent-car-y 500.)

;;; DEFINITIONS
;;; utility functions
(define (mph->m/s speed) (* speed .46))

;;; define some useful g-states
(define green   (ps:make-gstate '(color . #(0.15 0.4 0.15))))
(define brown   (ps:make-gstate '(color . #(0.3 0.1 0.1))))
(define red     (ps:make-gstate '(color . #(0.5 0.1 0.1))))
(define blue    (ps:make-gstate '(color . #(0.1 0.1 0.5))))
(define yellow  (ps:make-gstate '(color . #(0.2 0.2 0.05))))
(define purple  (ps:make-gstate '(color . #(0.3 0.1 0.3))))
(define white   (ps:make-gstate '(color . #(0.4 0.2 0.0))))  ; actually orange
(define re-white (ps:make-gstate '(color . #(0.5 0.5 0.5))))
(define gray    (ps:make-gstate '(color . #(0.15 0.15 0.15))))
(define green    (ps:make-gstate '(color . #(0.1 0.3 0.1))))
(define tan     (ps:make-gstate '(color . #(0.4 0.3 0.1))))
(define road-tex (ps:make-gstate '(color . #(1. 1. 1.))
				 '(texture-file . "highway4.rgb")))
(define wall-tex (ps:make-gstate '(color . #(1. 1. 1.))
				 '(texture-file . "brick.rgba")))

;;; the field
(define field-width 400.)
(define field-depth 2000.)
(define field-z -.1)
(define field-x (- (/ field-width 2)))

(define field 
  (ps:place-node
   (ps:make-geode-plane field-width field-depth
			(cons 'gstate
			      (ps:make-gstate '(color . #(1. 1. 1. .4))
					      '(texture-file . "grass2.rgb")))
			(cons 'texture-scale 8.))
   field-x 0. field-z))

;;; field cycling
(define cycle-pt 1000.)           ; when past here, driver go back
; when past here, obstacle cars go back
;(define recycle-pt-front field-depth)
(define recycle-pt-front 200000000)
; when behind here, obstacle cars move forward
(define recycle-pt-back (- 50000000.))

;;; the road
(define num-lanes 4)
;(define lane-width 3.6)
(define lane-width 4.0)
(define road-tex-scale 15.9)    ; trial and error
; road-xpos * relative to world coords
(define road-xpos (+ (- (* num-lanes (/ lane-width 2))) .75))
(define road
  (ps:place-node
   (ps:make-geode-plane
    (* num-lanes lane-width) field-depth
    (cons 'gstate road-tex)
    (cons 'texture-scale road-tex-scale))
   (+ (/ field-width 2) road-xpos) 0. 0.4  ; was 0.5
   field))

(define doing-walls #t)
(if doing-walls
    (begin
      (define wall-tex-scale 3.0)    ; trial and error
      (define wall-height 6.0)
      (define wall-x (+ (/ field-width 2) road-xpos (* lane-width 8)))
      (define wall-1
	(pf:dcs-rot
	 (ps:place-node
	  (ps:make-geode-plane
	   field-depth wall-height
	   (cons 'gstate wall-tex)
	   (cons 'texture-scale wall-tex-scale))
	  (- wall-x (* lane-width 12)) 0. 0. ;wall-height
	  field)
	 0. 90. 90.))
      (define wall-2
	(pf:dcs-rot
	 (ps:place-node
	  (ps:make-geode-plane
	   field-depth wall-height
	   (cons 'gstate wall-tex)
	   (cons 'texture-scale wall-tex-scale))
	  wall-x 0. wall-height
	  field)
	 0. -90. -90.))
))

; (file-name . scale), or #f if no trees
;(define tree-object #f)
(define tree-object '("tree2.flt" . .5))

;;; trees
(if tree-object
    (begin
      (define tree-geode (ps:make-node-from-file (car tree-object) '()))
      (define num-trees 16)
      (define tree-near 5.)
      (define tree-far 5.1)
      (define (tree-pick-x)
	; move to center of field
	(+ (/ field-width 2)
	   ; pick left or right of road
	   (* (- (* 2 (random-int 0 1)) 1)
	      ; pick distance from road's edge
	      (+ (* (/ num-lanes 2.) lane-width)
		 (random-float tree-near tree-far)))))
      (define (tree-pick-y)
	; uniform distribution over [0, cycle-pt] along road
	(random-float 0. cycle-pt))
      (letrec
	  ((helper
	    (lambda (val)
	      (if (< 0 val)
		  (let ((tree-x (tree-pick-x))
			(tree-y (tree-pick-y)))
		    ; Any tree placed at dist. alpha from 0. has a twin
		    ;  placed at dist. alpha from cyle-pt, 2*cycle-pt &
		    ;  3*cycle-pt to avoid flicker.
		    (pf:dcs-scale
		     (ps:place-node tree-geode tree-x tree-y 0. field)
		     (cdr tree-object))
		    (pf:dcs-scale
		     (ps:place-node tree-geode tree-x (+ cycle-pt tree-y)
				    0. field)
		     (cdr tree-object))
		    (pf:dcs-scale
		     (ps:place-node tree-geode tree-x
				    (+ (* 2 cycle-pt) tree-y) 0. field)
		     (cdr tree-object))
		    (pf:dcs-scale
		     (ps:place-node tree-geode tree-x
				    (+ (* 6 cycle-pt) tree-y) 0. field)
		     (cdr tree-object))
;		    (m-display "tree placed" #\newline)
		    (helper (- val 1)))))))
	(helper (/ num-trees 4)))))

(provide 'nyd-world)
