/*
 * Copyright (c) 1993 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that (i) the above copyright notices and this
 * permission notice appear in all copies of the software and related
 * documentation, and (ii) the name of Silicon Graphics may not be
 * used in any advertising or publicity relating to the software
 * without the specific, prior written permission of Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

/* 
 * file: perfly.h	
 * --------------
 * $Revision: 1.7 $
 * $Date: 1995/11/22 21:46:53 $
 */

#ifndef __PERFLY_H__
#define __PERFLY_H__

/******************************************************************************
 *			    Constants           
 *******************************************************************************/

#include "generic.h"
#include "env.h"

#define UPDATE_MODES	0x1
#define REDRAW_WINDOW	0x2

/* tokens for resetPosition */
#define POS_ORIG	0
#define POS_CENTER	1

#define DEVLIST_LENGTH 	9
#define BOUND		1000

/* limits and sizes */
#define	MAX_LAMPS	8

#if MCCALLUM
#define PERSIA_MAX_NUM_C_HOOKS 50
#endif

/*
 *  Structure for data shared between processes
 */
typedef struct
{
	/* required by common/main.c */
    pfChannel	    *masterChan;	/* Main channel */
    long	    phase;		/* phase: FREE, FLOAT, or LOCK */
    float	    frameRate;		/* target frame rate	 */
    ulong	    procLock;		/* flags to lock down processors */

	/* required by common/generic.c */
    long	    input;		/* input type: GL or X */
    long	    gui;		/* ui mode flag */
    long	    exitFlag;		/* exit application flag */
    long	    updateChannels;	/* update channel configuration */
    float	    near;		/* near clipping plane  */
    float	    far;		/* far clipping plane  */
    float	    sceneSize;		/* maximum scene bound */
    long	    stats;		/* diag graph on/ff flag */
    pfList	    *texList;		/* texture list to download */

	/* required by common/env.c */
    pfEarthSky	    *eSky;		/* current EarthSky model */
    long	    earthSkyMode;	/* Earth-Sky mode */
    long	    fog;		/* fog on/off */
    float	    farFogRange;	/* far fog range */
    float	    nearFogRange;	/* near fog range */
    pfLightSource   *sun;		/* sun light source */


	/* used for libpfutil input handling */
    pfuWidget       *guiWidgets[PFUGUI_MAXWIDGETS];
    long	    guiFormat;		/* GUI vertical, or horizontal */
    pfuGLXWindow    xWinInfo;		/* GLX window info */
    pfuMouse	    mouse;		/* mouse structure */
    pfuEvent	    events;		/* event structure */
    pfuXformer	    *xformer;		/* interactive movement model */
    long	    xformerMode;	/* Fly, Drive, or Trackball   */

	/* used in perfly to define light sources */
    pfVec3	    lampXyz[MAX_LAMPS];	/* light source direction */
    pfVec3	    lampRgb[MAX_LAMPS];	/* light source color */
    int		    lamps;		/* number of light sources */

	/* Viewpoint and modes control */
    pfScene	    *scene;		/* visual database */
    pfDCS	    *sceneDCS;		/* DCS for trackball interface */ 
    pfGroup	    *sceneGroup;	/* root of all loaded files */ 
    long	    realityGraphics;	/* flag for RE */
    pfCoord	    viewCoord;		/* current view position, direction */ 
    pfCoord	    initView;		/* initial view position, direction */ 
    long	    resetPos;		/* reset position flag */
    long	    drawFlags[MAX_PIPES];/* Flags for draw process(es) */
    long	    redrawOverlay;	/* signal to redraw overlay */
    long	    stress;		/* Enable/disable stress */
    long	    stressMax;		/* stress bound */
    float	    lowLoad; 		/* stress parameters- low bound */
    float	    highLoad;		/* high bound */
    float	    stressScale;	/* stress scale amount */
    float	    LODscale;		/* LOD scale value */
    float	    fadeRange;		/* LOD fade range- RE Graphics */
    long	    fade;		/* Fade on/off -   RE Graphics */
    long	    collideMode;	/* collision & terrain following flag*/
    long	    statsEnable;	/* diag graph mode flag */
    long	    printStats;		/* flag to print stats to stderr */
    long 	    wireframe;		/* wireframe mode flag */
    long 	    aa;			/* anti-alias flag */
    long 	    backface;		/* backface culling flag */
    long	    lighting;		/* lighting enable/disable */
    long	    texture;		/* texture enable/disable */
    long	    cullMode;		/* culling mode */
    float	    explode;		/* explode files */
    long	    snapImage;		/* save screen to file */
    
#if MCCALLUM

    /* mccallum - multi-window stuff */
    int num_cameras;
    pfChannel *chans[MAX_CHANS];
    pfDCS *camera_pos[MAX_CHANS];
#if 1 || ELK_WINDOW_POS
    int window_positions[MAX_CHANS][4];
    int window_num;
#endif

    /* C hook functions */
    void (*c_hooks[PERSIA_MAX_NUM_C_HOOKS])();
    int num_c_hooks;

    /* frame count */
    int frame_count;

    /* eye processing */
    int doing_eye_processing;
    int eye_camera_num;
    pfDCS *eye_head_dcs;

#endif /* MCCALLUM */
    
} SharedViewState;

/* Structures visible to all processes */
extern SharedViewState	*ViewState;   /* holds data shared between processes */

/* 
 * Function prototypes  - called by common 
 */

    /* Application process */
extern long initSceneGraph(pfScene *scene);
extern void initViewState(int);
extern void initView(pfScene *scene);
extern void updateView(pfChannel *chan);
extern void resetPosition(long pos);
extern void initXformer(void);
extern void updateSim(pfChannel *chan);
extern void xformerMode(long mode, long collideMode);
extern void drawModesChanged(void);

    /* Draw process */
extern void initPipe(pfPipe *p);
extern void drawUpdate(pfChannel *chan, void *data);
#endif
