;;; Definitions for "Persia"

(require 'cscheme)

;;; Don't clutter the screen the Elk's garbage collection notices
(set! garbage-collect-notify? #f)

;;; MATRIX

(define ps:tmp-pfmatrix (ps:make-pfmatrix))

;;; NODE HELPING FUNCTIONS

(define ps:*scene-group* #f)		; just until elk.c can set it
(define ps:*default-parent* ps:*scene-group*)
(define (ps:set-default-parent a-node)
  (set! ps:*default-parent* a-node))

(define (ps:place-node the-node x y z . parent)
  (begin
    (let ((the-group (if (null? parent) ps:*default-parent* (car parent)))
	  (the-dcs (pf:new-dcs)))
      (pf:add-child the-dcs the-node)
      (pf:add-child the-group the-dcs)
      (pf:dcs-trans the-dcs x y z)
      the-dcs)))

(define _ps:*planting-scs* #f)
(define (_ps:planting-scs)
  (if (not _ps:*planting-scs*)
      (define _ps:*planting-scs* (ps:make-pfmatrix)))
  _ps:*planting-scs*)

;;; or "ps:put-node" ?
(define (ps:plant-node the-node . args)
  (if (> (length args) 2)
      (apply ps:plane-node-1 (cons the-node args))    ; matrix version
      (apply ps:plane-node-2 (cons the-node args))))  ; x/y/z version

(define (ps:plant-node-1 the-node mat . parent)
  (begin
    (let ((the-group (if (null? parent) ps:*default-parent* (car parent)))
	  (the-scs (pf:new-scs mat)))
      (pf:add-child the-scs the-node)
      (pf:add-child the-group the-scs)
      the-scs)))

(define (ps:plant-node-2 the-node x y z . parent)
  (begin
    (let ((the-group (if (null? parent) ps:*default-parent* (car parent)))
	  (the-scs (begin
		     (pf:make-trans-matrix (_ps:planting-scs) x y z)
		     (pf:new-scs _ps:*planting-scs*))))
      (pf:add-child the-scs the-node)
      (pf:add-child the-group the-scs)
      the-scs)))

(define (ps:list-children the-group)
  (let ((num-kids (pf:get-num-children the-group))
	(kids nil))
    (do ((i (1- num-kids) (1- i)))
	((> i 0))
      (setq kids (cons (pf:get-child the-group i))))
    kids))
    

;;; HOOKS

;;; The pre-frame hook functions

;; This is required by the frame loop, and their types must match
;; those expected in elk.c
(define ps:*pre-frame-hooks* '())

(define (ps:*pre-frame-function*)
  (map (lambda (hook) (apply hook '())) ps:*pre-frame-hooks*))

;;; Add the function ps:set-delta-seconds-hook as a pre-frame hook 
;;; if you want these variables to be kept up-to-date.
(define ps:*second-count* 0.0)
(define ps:*prev-second-count* 0.0)
(define ps:*delta-seconds* 0.0)
(define (ps:set-delta-seconds-hook)
  (set! ps:*second-count* (ps:second-count))
  (set! ps:*delta-seconds* (- ps:*second-count* ps:*prev-second-count*))
  (set! ps:*prev-second-count* ps:*second-count*))

(define (ps:add-pre-frame-hook lamb)
  (set! ps:*pre-frame-hooks* (cons lamb ps:*pre-frame-hooks*)))

(define (ps:remove-pre-frame-hook lamb)
  (set! ps:*pre-frame-hooks* (remove! lamb ps:*pre-frame-hooks*)))

;;; The post-init-cameras hook functions

(define ps:*post-init-cameras-hooks* '())

(define (ps:*post-init-cameras-function*)
  (map (lambda (hook) (apply hook '())) ps:*post-init-cameras-hooks*))

(define (ps:add-post-init-cameras-hook lamb)
  (set! ps:*post-init-cameras-hooks* 
	(cons lamb ps:*post-init-camers-hooks*)))

(define (ps:remove-pre-frame-hook lamb)
  (set! ps:*post-init-cameras-hooks* 
	(remove! lamb ps:*post-init-cameras-hooks*)))

(define ps:*screen-width* (ps:mouse-x-max ))
(define ps:*screen-height* (ps:mouse-y-max))

;;; GSTATE CREATION WITH ATTRIBUTES
(define (ps:make-gstate . args)
  (let* ((tf (assq 'texture-file args))
	 (gs (if tf 
		 (ps:make-textured-gstate (cdr tf))
		 (ps:make-plain-gstate))))
    (begin
     (mapcar (lambda (pair)
	       (cond 
		((eq? (car pair) 'color)
		 (ps:gstate-color gs (cdr pair)))
		((eq? (car pair) 'texture-file)
		 #t)
		(else "bad attribute label")))
	     args)
     gs)))

;;; CREATING PLANES

(define (ps:make-geode-plane width depth . attributes)
  (let* ((gstate (assq 'gstate attributes))
	 (tex-scale (assq 'texture-scale attributes))
	 (geode (ps:make-geode-plane1 width depth)))
    (if gstate (pf:gset-gstate (pf:get-gset geode 0) (cdr gstate)))
    (if tex-scale (ps:plane-scale-texture geode (cdr tex-scale)))
    geode))

;;; CREATING CUBES

(define (ps:make-geode-cube width depth height . attributes)
  (let* ((gstate (assq 'gstate attributes))
	 (tex-scale (assq 'texture-scale attributes))
	 (geode (ps:make-geode-cube1 width depth height)))
    (if gstate (pf:gset-gstate (pf:get-gset geode 0) (cdr gstate)))
    (if tex-scale (ps:cube-scale-texture geode (cdr tex-scale)))
    geode))

;;; CREATING SPHERES

(define (ps:make-geode-sphere radius gen . attributes)
  (let* ((gstate (assq 'gstate attributes))
	 (tex-scale (assq 'texture-scale attributes))
	 (geode (ps:make-geode-sphere1 radius gen)))
    (if gstate 
	(do ((i 0 (1+ i)))
	    ((> i 7)) ; There are 8 gset's in a sphere geode
	  (pf:gset-gstate (pf:get-gset geode i) (cdr gstate))))
    (if tex-scale (ps:sphere-scale-texture geode (cdr tex-scale)))
    geode))

;;; CREATING ROCKS

(define (ps:make-geode-rock width depth height gen . attributes)
  (let* ((gstate (assq 'gstate attributes))
	 (tex-scale (assq 'texture-scale attributes))
	 (geode (ps:make-geode-rock1 width depth height gen)))
    (if gstate 
	(do ((i 0 (1+ i)))
	    ((> i 3)) ; There are 4 gset's in a rock geode
	  (pf:gset-gstate (pf:get-gset geode i) (cdr gstate))))
    (if tex-scale (ps:rock-scale-texture geode (cdr tex-scale)))
    geode))

(define (ps:make-node-from-file filename . gstate)
  (if (> (length gstate) 0)
      (ps:make-node-from-file1 filename (car gstate))
      (ps:make-node-from-file1 filename nil)))

;;; RANDOM NUMBER GENERATION UTILITIES

(define (random-float low high)
  (let ((rand (/ (random) 2147483647))	; max random is 2^31-1
	(range (- high low)))
    (+ (* rand range) low)))

(define (random-int low high)
  (+ (modulo (random) (+ (- high low) 1)) low))

;;; RANDOM PLACEMENT

(define (ps:random-place-node the-node min-x max-x min-y max-y z)
  (ps:place-node the-node
		 (random-float min-x max-x)
		 (random-float min-y max-y)
		 z))

;;; RANDOM COLOR
(define (ps:make-random-color-vector red-min red-max
				     green-min green-max
				     blue-min blue-max
				     alpha-min alpha-max)
  (let ((color-vector (make-vector 4)))
    (vector-set! color-vector 0 (random-float red-min red-max))
    (vector-set! color-vector 1 (random-float green-min green-max))
    (vector-set! color-vector 2 (random-float blue-min blue-max))
    (vector-set! color-vector 3 (random-float alpha-min alpha-max))
    color-vector))
  

;;; RANDOM ROCKS

(define (ps:make-random-geode-rock)
  (ps:make-geode-rock
   (random-float 50 300)
   (random-float 50 300)
   (random-float 10 50)
   (random-int 1 2)
   (cons 'gstate 
	 (ps:make-gstate (cons 'color 
			       (ps:make-random-color-vector .7 .4
							    .2 .3
							    .2 .3
							    .5 .5))
			 (cons 'texture-file "grass2.rgb")))
   (cons 'texture-scale 5.)))

(define (ps:make-and-place-random-rock min-x max-x min-y max-y z)
  (ps:place-node (ps:make-random-geode-rock)
		   (random-float min-x max-x)
		   (random-float min-y max-y)
		   z))

;;; VULTURE CAMERAS

(define (ps:make-vulture-camera camera-num the-node height)
  (let ((vulture-dcs (pf:new-dcs)))
    (pf:add-child ps:*default-parent* vulture-dcs)
    (pf:dcs-rot vulture-dcs 0. -90. 0.)
    (ps:attach-camera camera-num vulture-dcs)
    (ps:add-pre-frame-hook (lambda () 
			  (pf:get-dcs-mat the-node ps:tmp-pfmatrix)
			  (pf:dcs-trans vulture-dcs
				     (ps:pfmatrix-ref ps:tmp-pfmatrix 3 0)
				     (ps:pfmatrix-ref ps:tmp-pfmatrix 3 1)
				     height)))))

;;; VEHICLE VECTOR STRUCTURE

;;; This vector must match what is expected by update_vehicle.c.
(define (ps:make-vehicle-vector-with-dcs the-dcs)
  (let ((v (make-vector 8)))
    (vector-set! v 0 the-dcs)
    (vector-set! v 1 0.0)		; velocity
    (vector-set! v 2 50.0)		; mass
    (vector-set! v 3 150.0)		; engine
    (vector-set! v 4 70.0)		; brakes
    (vector-set! v 5 1.0)		; dynamic friction
    ; b/c brakes are always non-zero
    (vector-set! v 6 10.0)		; wheelbase
    (vector-set! v 7 4.0)		; width
    v))
;;; This value will depend on the values in the vector above.
;;; This value is used below when setting the engine speed sound.
(define ps:*max-velocity* 66.0)

(define (ps:make-mouse-car-with-dcs the-dcs accel-control)
  (let ((v (ps:make-vehicle-vector-with-dcs the-dcs)))
    (ps:add-pre-frame-hook (lambda () (ps:update-car-with-mouse v accel-control)))
    v))

(define (ps:update-car-with-mouse v accel-control)
  (begin
    ;;(format #t "velocity ~s\n" (vector-ref v 1))
    (ps:update-vehicle-vector 
     v

;     (- (- (/ (ps:mouse-x) (ps:mouse-x-max)) .5)) ; for linear control
;     (let ((x (* 2 (- (/ (ps:mouse-x) (ps:mouse-x-max)) .5)))) ; for cubic control
;       (- (* x x x)))
;     (let ((mouse-sensitivity .1)) ; for desensitized linear control
;       (* mouse-sensitivity (- (- (/ (ps:mouse-x) (ps:mouse-x-max)) .5))))

     (let ((x (* 2 (- .5 (/ (ps:mouse-x) (ps:mouse-x-max))))) ; for trig control
	   (pi-div-2 1.5707963)
	   (mouse-sensitivity .5))
       (* mouse-sensitivity (sin (* pi-div-2 x))))

     (if (number? accel-control)
	 accel-control
	 (/ (ps:mouse-y) (ps:mouse-y-max)))
     .1
     ps:*delta-seconds*)))

(define (ps:make-kart-car-with-dcs the-dcs)
  (let ((v (ps:make-vehicle-vector-with-dcs the-dcs)))
    (ps:add-pre-frame-hook
     (lambda () (ps:update-car-with-kart v (if (bound? first-car-control)
					       (if (equal? 'constant first-car-control)
						   (if (bound? first-car-accel)
						       first-car-accel
						       .5)
						   'accelerator)
					       'accelerator))))
    v))

(define ps:*doing-kart-sound* #t)

(define (ps:update-car-with-kart v accel-control)
  (begin
    ;;(format #t "velocity ~s *delta-seconds* ~s\n" (vector-ref v 1) ps:*delta-seconds*)
    (ps:update-vehicle-vector v
;			      (ps:kart-steering)   ; for linear control
			      (let ((x (ps:kart-steering))
				    (steer-sensitivity .3)
				    (pi-div2 1.5707963))
				(* steer-sensitivity (sin (* pi-div2 x))))
			      (if (number? accel-control)
				  accel-control
				  (let ((accelerator-sensitivity .5))
				    (* accelerator-sensitivity (ps:kart-gas))))
			      (if (number? accel-control)
				  0.
				  (ps:kart-brake))
			      ps:*delta-seconds*)
    (let ((vel (vector-ref v 1)))
      (if ps:*doing-kart-sound*
	  (ps:engine-sound (/ vel ps:*max-velocity*))))))

;;; CAR

(define (ps:make-dcs-car width depth height . attributes)
  (let* ((the-dcs (pf:new-dcs))
	 (body
	  (let
	      ((body-group (pf:new-dcs))
	       (chasis (apply ps:make-geode-cube 
			      (append (list width depth (* .75 height))
				      attributes)))
	       (upper-body (apply ps:make-geode-cube
				  (append (list width (* .75 depth)
						(* .25 height)) attributes)))
	       (hood (apply ps:make-geode-cube
			    (append (list (* .75 width) (* .2 depth)
					  (* .25 height)) attributes))))
	    (pf:add-child body-group chasis)
	    (ps:place-node upper-body 0. 0. (* .75 height) body-group)
	    (ps:place-node hood (* .125 width) (* .75 depth) (* .75 height) body-group)
	    body-group))
	 (ornament-size (/ width 35.))
	 (ornament (ps:make-geode-cube 
		    ornament-size ornament-size ornament-size
		    (cons 'gstate (ps:make-gstate '(color . #(.9 .9 .9))))))
	 (wheel-width (/ width 4))
	 (wheel-radius (/ height 3))
	 (wheel-to-end (/ depth 8))
	 (body-clearance (* wheel-radius .3))
	 (wheel (ps:make-geode-cube
		 wheel-width (/ height 2) (/ height 2)
		 (cons 'gstate (ps:make-gstate '(color . #(.0 .0 .0))))))
	 (cab-inset-width (* width .1))
	 (cab (apply ps:make-geode-cube
		     (append (list (- width (* 2 cab-inset-width))
				   (* depth .5)
				   (* height .7))
			     attributes)))
	 (body-dcs (ps:place-node body 
				  (- (/ width 2))
				  (- (/ depth 2))
				  body-clearance
				  the-dcs)))
    (ps:place-node ornament
		   (- (/ width 2) (/ ornament-size 2))
		   (- (* .95 depth) ornament-size) 
		   height 
		   body-dcs)
    (ps:place-node cab
		   cab-inset-width (* depth .2) height body-dcs)
    (ps:place-node wheel 
		   (/ wheel-width -2) 
		   wheel-to-end
		   (- body-clearance)
		   body-dcs)
    (ps:place-node wheel 
		   (+ width (/ wheel-width -2)) 
		   wheel-to-end
		   (- body-clearance)
		   body-dcs)
    (ps:place-node wheel 
		   (/ wheel-width -2)
		   (- depth wheel-to-end wheel-radius)
		   (- body-clearance)
		   body-dcs)
    (ps:place-node wheel 
		   (+ width (/ wheel-width -2)) 
		   (- depth wheel-to-end wheel-radius)
		   (- body-clearance)
		   body-dcs)
    the-dcs))


;;; WALLED FIELD
(define (ps:make-geode-wall width depth height)
  (ps:make-geode-cube width depth height
		      (cons 'gstate (ps:make-gstate '(color . #(.1 .1 .0))))))
;;; .4 .4 .4 1. no-tex
;;;.7 .4 .4 .2 brick-tex

(define (ps:make-dcs-walled-field width depth height . attributes)
  (let* ((field (pf:new-dcs))
	 (geode-field
	  (apply ps:make-geode-plane
		 (append (list width depth)
			 attributes)))
	 (wall-thickness 1.0)
	 (wall-height 2.0))
    (ps:place-node (ps:make-geode-wall wall-thickness depth wall-height)
		     0. 0. 0. field)
    (ps:place-node (ps:make-geode-wall wall-thickness depth wall-height)
		     (- width wall-thickness) 0. 0. field)
    (ps:place-node (ps:make-geode-wall width wall-thickness wall-height)
		     0. 0. 0. field)
    (ps:place-node (ps:make-geode-wall width wall-thickness wall-height)
		     0. (- depth wall-thickness) 0. field)
    (pf:add-child ps:*scene-group* field)
    (ps:set-default-parent field)
    (ps:place-node geode-field 0. 0. 0.)
    (pf:dcs-trans field 0. 0. height)
    field))

;;; GOOFING AROUND

(define (u)
  (ps:frame))

(define (spin-heading dcs)
  (do ((i 0.0 (1+ i)))
      ((> i 360))
    (pf:dcs-rot dcs i 0.0 0.0)
    (ps:tick)))


;;; RECORDING AND PLAYBACK

(define (ps:record-dcs the-vector dcs ticknum)
  (let ((m (ps:make-pfmatrix)))
    (pf:get-dcs-mat dcs m)
    (vector-set! the-vector ticknum m)))

(define (ps:play-dcs the-vector dcs ticknum)
  (pf:dcs-mat dcs (vector-ref the-vector ticknum)))


;;;
;;; SOME USEFUL, BUT NON-STANDARD OBJECTS.
;;; THESE FUNCTIONS MAY CHANGE IN THE FUTURE.
;;;

;;; SLABS
(define (ps:make-random-geode-slab)
  (ps:make-geode-cube (random-float 10 100)
		       (random-float 10 100)
		       (random-float 5 20)
		       (cons 'gstate
			     (ps:make-gstate 
			      (cons 'color (ps:make-random-color-vector
					    .6 .3
					    .0 .2
					    .0 .2
					    1. 1.))))))
(define (ps:make-and-place-random-slab ground-width ground-depth)
  (ps:random-place-node (ps:make-random-geode-slab) 
			  0. ground-width 0. ground-depth 0.))

;;; ARCHES
(define (ps:make-dcs-arch separation width depth height . attributes)
  (let ((the-dcs (pf:new-dcs))
	(pillar (apply ps:make-geode-cube 
		       (append (list width depth height)
			       attributes)))
	(cap (apply ps:make-geode-cube
		    (append (list (+ separation (* 2 width)) 
				  depth width)
			    attributes))))
    (ps:place-node pillar 0. 0. 0. the-dcs)
    (ps:place-node pillar (+ separation width) 0. 0. the-dcs)
    (ps:place-node cap 0. 0. height the-dcs)
    the-dcs))

(define the-standard-node-arch #f)
(define (standard-node-arch)
  (if (not the-standard-node-arch)
      (set! the-standard-node-arch
	    (ps:make-dcs-arch 12. 3. 5. 20. 
			      (cons 'gstate 
				    (ps:make-gstate 
				     '(color . #(.4 .4 .1)))))))
  the-standard-node-arch)
(define (ps:make-and-randomly-place-arch min-x max-x min-y max-y z)
  (let ((the-dcs
	 (ps:place-node (standard-node-arch)
			(random-float min-x max-x)
			(random-float min-y max-y)
			z)))
    (pf:dcs-rot the-dcs (random-float 0. 180.) 0. 0.)
    the-dcs))

;;; LOAD STANDARD PERSIA SHARED OBJECT FILES
(ps:dlopen "geodes.so")
(ps:dlopen "gset.so")
(ps:dlopen "gstate.so")
(ps:dlopen "plane.so")
(ps:dlopen "cube.so")
(ps:dlopen "sphere.so")
(ps:dlopen "rock.so")
(ps:dlopen "time.so")
(ps:dlopen "tree.so")
(ps:dlopen "light.so")
(ps:dlopen "material.so")
(ps:dlopen "pfmatrix.so")
(ps:dlopen "scs.so")
(ps:dlopen "engine_sound.so")

(provide 'persia)
