/* Planes
   written by:  Chandler Chao from SGI provided code.

   Description:  Herein are provided some basic routines to construct a
   plane.

  Planes are created by: MakePlane(width, depth, red, green, blue,
                                    transparency, geostate);
             it returns a geoset.
*/

#include "persia-scm.h"
#include "geodes.h"

pfGeoSet*
MakePlane(float width, float depth)
{
    pfGeoSet *gset;
    void *arena;
    pfVec3 *scoords;
    pfVec4 *scolor;
    ushort *scindex;
    pfTexture *tex;
    pfTexEnv *tenv;
    pfVec2 *tcoords;
    pfVec3 *snorms;
    ushort *snindex;
    ushort *svindex;
    ushort *tindex;

    /*
     * The pfGeoSet has been allocated out of shared memory, but notice that
     * the geometry data (above) has not. Since the data is static, in an
     * MP application, all processes will have a copy of the data as
     * initialized before they are forked.
     * However, changes made to the geometry data will only be visible to 
     * the process the changes are made in.
     */
    arena = pfGetSharedArena();
    gset = pfNewGSet(arena);

    scoords = (pfVec3 *) pfMalloc (sizeof(pfVec3) * 4, arena);
    scolor = (pfVec4 *) pfMalloc (sizeof(pfVec4), arena);
    scindex = (ushort *) pfMalloc (sizeof(ushort), arena);
    tcoords = (pfVec2 *) pfMalloc (sizeof(pfVec2) * 4, arena);
    snorms = (pfVec3 *) pfMalloc (sizeof(pfVec3), arena);
    snindex = (ushort *) pfMalloc (sizeof(ushort), arena);
    svindex = (ushort *) pfMalloc (sizeof(ushort) * 4, arena);
    tindex = (ushort *) pfMalloc (sizeof(ushort) * 4, arena);

    scoords[0][0] = 0.0; scoords[0][1] = 0.0; scoords[0][2] = 0;
    scoords[1][0] = width; scoords[1][1] = 0.0; scoords[1][2] = 0;
    scoords[2][0] = width; scoords[2][1] = depth; scoords[2][2] = 0;
    scoords[3][0] = 0.0; scoords[3][1] = depth; scoords[3][2] = 0;

    tcoords[0][0] = 0.0f; tcoords[0][1] = 0.0f;
    tcoords[1][0] = width; tcoords[1][1] = 0.0f;
    tcoords[2][0] = width; tcoords[2][1] = depth;
    tcoords[3][0] = 0.0f; tcoords[3][1] = depth;

    /* xxx What's the deal here?  Can we just remove the
       pfGSetAttr PFGS_COLOR4 call below? */
/*    scolor[0][0] = red; scolor[0][1] = green;
    scolor[0][2] = blue; scolor[0][3] = trans;
*/    scindex[0] = 1;

    snorms[0][0] = 0.0f; snorms[0][1] = 0.0f; snorms[0][2] = -1.0f;

    snindex[0] = 0;

    svindex[0] = 0; svindex[1] = 1; svindex[2] = 2; svindex[3] = 3;

    tindex[0] = 0; tindex[1] = 1; tindex[2] = 2; tindex[3] = 3;

    pfGSetAttr(gset, PFGS_COORD3, PFGS_PER_VERTEX, scoords, svindex);
    pfGSetAttr(gset, PFGS_NORMAL3, PFGS_PER_PRIM, snorms, snindex);
    pfGSetAttr(gset, PFGS_COLOR4, PFGS_OVERALL, scolor, scindex);
    pfGSetAttr(gset, PFGS_TEXCOORD2, PFGS_PER_VERTEX, tcoords, tindex);
    pfGSetPrimType(gset, PFGS_QUADS);
    pfGSetNumPrims(gset, 1);
    
    return gset;
}

Object P_MakePlane(Object s_width,
		   Object s_depth)
{
  pfGeode *geode;
  pfGeoSet *gset;
  float width, depth;

  Check_Type(s_width, T_Flonum);
  Check_Type(s_depth, T_Flonum);
  gset = MakePlane(FLONUM(s_width)->val,
		   FLONUM(s_depth)->val);
  geode = pfNewGeode();
  pfAddGSet(geode, gset);
  return Make_Pfo((pfObject*)geode);
}

Object
P_PlaneSetTexScale(Object geode, Object scale)
{
  pfVec2 *tcoords;
  ushort *tindex;
  float scal;
  pfGeode *g;
  pfGeoSet *geoset;
  void *arena;
  pfVec3 *scoords;
  ushort *svindex;
  float width, depth;

  Check_Pfo_Class(geode, PFTYPE_GEODE);
  Check_Type(scale, T_Flonum);

  g = (pfGeode*) PFO(geode);
  geoset = pfGetGSet(g, 0);
  scal = 1.0f / FLONUM(scale)->val;

  pfGetGSetAttrLists(geoset, PFGS_TEXCOORD2, (void**)&tcoords, &tindex);
  pfGetGSetAttrLists(geoset, PFGS_COORD3, (void**)&scoords, &svindex);
  width = scoords[1][0];
  depth = scoords[2][1];

  tcoords[1][0] = width*scal;
  tcoords[2][0] = width*scal; tcoords[2][1] = depth*scal;
  tcoords[3][1] = depth*scal;

  return geode;
}

#if 0
float plane_get_depth (pfGeode *plane)
{
  GeoSet *geoset;
  pfVec3 *scoords;
  geoset = pfGetGSet (plane, 0);

  pfGetGSetAttrLists(geoset, PFGS_COORD3, (void**)&scoords, &svindex);
  return scoords[2][1];
}
#endif

void plane_init()
{
  Define_Primitive(P_MakePlane, "ps:make-geode-plane1", 2, 2, EVAL);
  Define_Primitive(P_PlaneSetTexScale, "ps:plane-scale-texture", 2, 2, EVAL);
}
