/* policy.m - Scheme interface UTPolicy.m */

#include "persia-scm.h"
#include "persia.h"
#include <assert.h>

#if 1

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#define NAME "/tmp/rug_c"
#define RNAME "/tmp/rug_s"

static int sock;
static struct sockaddr_un name;
static struct sockaddr_un rname;

static int npd = 8;

Object P_InitPolicy()
{
  /* Make local socket */
  if ((sock = socket (AF_UNIX, SOCK_DGRAM, 0)) < 0)
    perror("creating socket");
  name.sun_family = AF_UNIX;
  strcpy(name.sun_path, NAME);
  unlink(NAME);
  if (bind(sock, &name, sizeof(struct sockaddr_un)))
    perror("bind");
  fprintf(stderr, "new socket(), fd=%d\n", sock);

  /* Setup for remote sending */
  rname.sun_family = AF_UNIX;
  strcpy(rname.sun_path, RNAME);
  return Void;
}

#define BUFS 128

Object P_ChooseAction (Object new_perceptions,
		       Object prev_reward)
{
  int i;
  int a;
  Object entry;
  char buf[BUFS];
  int bufp = 0;
  int rlen;
  struct sockaddr_un rn;

  Check_Type(new_perceptions, T_Vector);
  Check_Type(prev_reward, T_Flonum);
  for (i = npd-1; i >= 0; i--)
    {
      entry = P_Vector_Ref(new_perceptions, Make_Fixnum(i));
      Check_Type(entry, T_Fixnum);
      buf[bufp++] = FIXNUM(entry);
      printf("%d ", (int)buf[bufp-1]);
    }
#if 0
  buf[bufp++] = (int) (100 * (FLONUM(prev_reward)->val + 10.0));
#else
  buf[bufp++] = 0;
#endif
  printf(", %d/10\n", buf[bufp-1]);
  if ((i = sendto(sock, buf, bufp, 0,
		  &rname, sizeof(struct sockaddr_un)))
      < 0)
    perror("sendto");
  assert(i == bufp);

  if (recvfrom(sock, buf, BUFS, 0, &rn, &rlen) < 0)
    perror("recvfrom");
  a = buf[0];
  return Make_Fixnum(a);
}

#elif 0

#define Object ObjcObject
#define String ObjcString
#include <objects/SocketPort.h>
#undef Object
#undef String

id local_port;
id remote_port;
int npd;

Object P_InitPolicy(Object socknum)
{
  Check_Type(socknum, T_Fixnum);
  local_port = [SocketPort newLocal];
  remote_port = [SocketPort newRemoteWithNumber:3
			    onHost:@"localhost"];
}

#define BUFS 100

Object P_ChooseAction (Object prev_perceptions,
		       Object new_perceptions,
		       Object action,
		       Object reward)
{
  int i;
  int a;
  Object entry;
  char buf[BUFS];
  int bufp = 0;
  int len;
  id rp;

  Check_Type(prev_perceptions, T_Vector);
  Check_Type(new_perceptions, T_Vector);
  Check_Type(action, T_Fixnum);
  Check_Type(reward, T_Flonum);
  for (i = 0; i < npd; i++)
    {
      entry = P_Vector_Ref(prev_perceptions, Make_Fixnum(i));
      Check_Type(entry, T_Fixnum);
      buf[bufp++] = FIXNUM(entry);
      entry = P_Vector_Ref(new_perceptions, Make_Fixnum(i));
      Check_Type(entry, T_Fixnum);
      buf[bufp++] = FIXNUM(entry);
    }
  buf[bufp++] = FIXNUM(action);
  buf[bufp++] = (int) (10 * FLONUM(reward)->val);
  [local_port sendPacket:buf length:bufp
	      toPort:remote_port
	      timeout:-1];
  len = [local_port receivePacket:buf length:BUFS
		    fromPort:&rp
		    timeout:-1];
  assert (len != -1);
  a = buf[0];
  return Make_Fixnum(a);
}
#else /* if 1 */


#define Object ObjcObject
#define String ObjcString
#include "Agent.h"
#include "HighwayWorldSM.h"
//#include "UTPolicy.h"
#include <objects/TextCoder.h>
#include <objects/StdioStream.h>
#undef Object
#undef String

static id policy;
static id sm;
static id agent;
static int npd;


Object P_InitPolicy(Object filename)
{
  id stream, coder;
  const char *fn;

  Check_Type(filename, T_String);
  fn = STRING(filename)->data;
  stream = [[StdioStream alloc] initWithFilename:fn fmode:"r"];
  coder = [[TextCoder alloc] initDecodingOnStream:stream];
  [coder decodeObjectAt:&policy withName:NULL];

  agent = [Agent new];
  sm = [HighwayWorldSM new];
  npd = [sm numPerceptionDimensions];
  [agent setPolicy:policy];
  [agent setSm:sm];

  [policy setupForAgent:agent];

  [coder release];
  
  printf("Read policy with %d actions\n", [policy numActions]);
  return Void;
}

Object P_ChooseAction (Object prev_perceptions,
		       Object new_perceptions,
		       Object reward)
{
  int i;
  static int prev[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int new[npd];
  int a;
  float r;
  Object entry;

  Check_Type(prev_perceptions, T_Vector);
  Check_Type(new_perceptions, T_Vector);
  Check_Type(action, T_Fixnum);
  Check_Type(reward, T_Flonum);
  
  for (i = 0; i < npd; i++)
    {
      prev[i] = new[i];
      entry = P_Vector_Ref(new_perceptions, Make_Fixnum(i));
      Check_Type(entry, T_Fixnum);
      new[i] = FIXNUM(entry);
    }
  a = FIXNUM(action);
  r = FLONUM(reward)->val;
  [policy updatePolicyFromPerceptions:prev
	  toPerceptions:new
	  usingActions:&a
	  withRewards:&r];

  [policy chooseActions:&a
	  fromPerceptions:new];
  return Make_Fixnum(a);
}


#endif /* if 1 */


void
init_policy()
{
  fprintf(stderr, "init_policy()\n");
  Define_Primitive(P_InitPolicy, "ps:policy-init", 0, 0, EVAL);
  Define_Primitive(P_ChooseAction, "ps:policy-choose-action", 2, 2, EVAL);
}

