#include "persia-scm.h"
#include "rock.h"

/*this function generates and returns a random integer to be used in xyz*/
/* adjustment;*/
int rnd( void )
{
  int value;
  int mod;

  value = rand() % 100;
  mod = value % 4;
  /* decide whether a positive or negative value */
  /* change mod value to bias the building of the rock, whether it is
     more caved inward or puffed outward */
  if (mod < 3) value = value;
  else if (mod == 3) value = -value;

  return value;
}

/*Adjusts a 3-D point by a random amount
  The if statements are put in there for sides 2 and 3 which are built
  in a negative direction */
void adjust_xyz( double potential_height, pnt_t *pt )
{
  double r;

  r = (double) rnd()/ 100.;
  r *= potential_height;
  if (wrkngon_side2 == TRUE) pt->x -= r;
  else pt->x += r;

  r = (double) rnd() / 100.;
  r *= potential_height;
  if (wrkngon_side3 == TRUE) pt->y -= r;
  else pt->y += r;

  r = (double) rnd() / 100.;
  r *= potential_height;
  pt->z += r;
  if (pt->z < 0) pt->z = 0.2;

  return;
}

/*Adjusts a 2-D point by a random amount;*/
void adjust_xy( double potential_height, pnt_t *pt )
{
  double r;

  r = (double) rnd() / 100.;
  r *= potential_height;
  if (wrkngon_side2 == TRUE) pt->x -= r;
  else pt->x += r;

  r = (double) rnd() / 100.;
  r *= potential_height;
  if (wrkngon_side3 == TRUE) pt->y -= r;
  else pt->y += r;

  return;
}

/* Don't touch this routine */
int find_top_left( int n )
{
  int r;

  if (n == 0) 
    return 0;
  else if (n == 1)
    return 3;
  else
    r = find_top_left(n - 1) + 3 * ((int)((pow(2, n - 2)) + 1.0) * (pow(2, n - 3)));
  
  return r;
}


double interpolate( double a, double b, double u)
{
  return ((1 - u) * a + u * b);
}

void intp_point( pnt_t p1, pnt_t p2, pnt_t *p, double u)
{
  p->x = interpolate(p1.x, p2.x, u);
  p->y = interpolate(p1.y, p2.y, u);
  p->z = interpolate(p1.z, p2.z, u);

  return;
}

/* Used to skip points that have already been calculated */
int skip ( int n )
{
  return (((n + 3) / 2) * ((n + 1) / 2) + (((n + 2) / 2) * (n / 2) / 2));
}

/* For each level in a triangle, figures out which points are in that
   row, both top and bottom */
void get_row_indices( int n, int j, int *indices)
{
  int i, k, jj;
  int next_point[10];

  next_point[0] = (j == 0) ? 0 : 1;
  for (i = 1; i <= n; i++) {
    next_point[i] = find_top_left(i);
    jj = (j - 1) / (1 << (n - i));
    next_point[i] += skip(jj);
  }

  for (i = 0; i <= j; i++) {
    for (k = n; k >= 0; k--) {
      if ( !(j % (1 << k)) && !(i % (1 << k))) {
	k = n - k;
	indices[i] = next_point[k];
	next_point[k]++;
	break;
      }
    }
  }
  
  return;
}

void copy_point (pnt_t *to, pnt_t *from)
 {
  to->x = from->x;
  to->y = from->y;
  to->z = from->z;

  return;
}

/*********************************************************************
/* Routines that may change */

/* All initializiations done in here */
void initRock(int n)
{
  int count;   /* Number of vertices in a side when drawing */
  int i;
  int size;    /* Number of vertices in a side when not drawing */

  generation = n;
  count = 0;

  size = find_top_left(n + 1);

  arena = pfGetSharedArena();

  /* Initialized from NULL because they don't need to be shared
     in performer */
  side1 = (pnt_t *) pfMalloc (sizeof (pnt_t) * size, NULL);
  side2 = (pnt_t *) pfMalloc (sizeof (pnt_t) * size, NULL);
  side3 = (pnt_t *) pfMalloc (sizeof (pnt_t) * size, NULL);
  side4 = (pnt_t *) pfMalloc (sizeof (pnt_t) * size, NULL);

  top = (int *) pfMalloc (sizeof (int) * (1 << n), NULL);
  bot = (int *) pfMalloc (sizeof (int) * (1 << n), NULL);

  /* Initialized with arena because they need to be shared in performer */
  for (i = 1; i <= (1 << n); i++) count += i * 2 + 1;
  vindex = (ushort*) pfMalloc (sizeof(ushort) * count, arena);
  nindex = (ushort*) pfMalloc (sizeof(ushort) * count, arena);

  for (i = 0; i < count; i++) nindex[i] = i;

  norms = (pfVec3 **) pfMalloc (sizeof(pfVec3) * 4, arena);
  for (i = 0; i < 4; i++)
    norms[i] = (pfVec3*) pfMalloc (sizeof(pfVec3) * count, arena);

  side = (pfVec3 **)pfMalloc(sizeof(pfVec3) * 4, arena);
  for (i = 0; i < 4; i++)
    side[i] = (pfVec3*) pfMalloc (sizeof(pfVec3) * size, arena);

  lengths = (long *)pfMalloc (sizeof(long) * 1 << generation, arena);

  scolor = (pfVec4 *) pfMalloc (sizeof(pfVec4) * 1, arena);
  scindex = (ushort *) pfMalloc (sizeof(ushort) * 1, arena);

  tcoords = (pfVec2 *) pfMalloc (sizeof(pfVec2) * 3, arena);
  tindex = (ushort *) pfMalloc (sizeof(ushort) * count, arena);

  for (i = 0; i < 4; i++)
    gset[i] = pfNewGSet(arena);

  return;
}

/* This is the routine that is responsible for generating all the points
   within a side */
void generate_side(pnt_t *plist)
{
  int n = generation;
  int x, y, k;
  int next;
  double potential_height;   /* Height to adjust by */

  double dx, dy, dz;
  
  for (k = 0; k < n; k++) {
    bot[0] = 0;
    next = find_top_left(k + 1);
    for (y = 1; y <= (1 << k); y++) {
      tmp_row_pointer = top;
      top = bot;
      bot = tmp_row_pointer;
      get_row_indices (k, y, bot);
      
      for (x = 0; x < y; x++) {

	/* DETERMINE THE HEIGHT TO OFFSET BY.  You may at times want to
	   adjust this to alter the jaggedness of the rock.  If so adjust
	   the denominator in the potential_height equation below.  The
	   range should stay between 5.0 and 8.0.  5.0 is very jagged, and
	   8.0 is more smooth and pyramid like */
	dx = plist[top[x]].x - plist[bot[x + 0]].x;
	dy = plist[top[x]].y - plist[bot[x + 0]].y;
	dz = plist[top[x]].z - plist[bot[x + 0]].z;
	potential_height = (double) sqrt(pow(dx, 2) + pow(dy, 2) +\
					 pow(dz, 2)) / 6.5;

	/* All these if statements find out if the point has already been
	   calculated for a differt side (edge point).  Therefore we don't
	   want to recalculate it, just copy it. */
	if (x == 0 && wrkngon_side4 == TRUE)
	  copy_point(&plist[next + 2 * x + 0], &side1[next + 2 * y - 1]);
	else {
	  intp_point (plist[top[x]], plist[bot[x + 0]], &plist[next + 2 * x + 0], 0.5);
	  adjust_xyz(potential_height, &plist[next + 2 * x + 0]);
	}

	/* find out if point already exists, if so set pointer to it */
	if (x == (y - 1) && wrkngon_side2 == TRUE)
	  copy_point(&plist[next + 2 * x + 1], &side1[next]);
	else if (x == (y - 1) && wrkngon_side3 == TRUE)
	  copy_point(&plist[next + 2 * x + 1], &side2[next]);
	else if (x == (y - 1) && wrkngon_side4 == TRUE)
	  copy_point(&plist[next + 2 * x + 1], &side3[next]);
	else {
	  intp_point (plist[top[x]], plist[bot[x + 1]], &plist[next + 2 * x + 1], 0.5);
	  adjust_xyz(potential_height, &plist[next + 2 * x + 1]);
	}

	intp_point (plist[bot[x + 0]], plist[bot[x + 1]], & plist[next + 2 * y + x], 0.5);

	if (y != (1 << k))
	  adjust_xyz(potential_height, &plist[next + 2 * y + x]);
	else
	  adjust_xy(potential_height, &plist[next + 2 * y + x]);
      }

      next += 3 * y;
    }
  }
  return;
}

/*this routine sets up the vindex for the GeoSet which is done by strip
note that this vindex is uniform no matter which side you are on*/
void generate_vindex()
{
  int i, j, k, c;
  int n;
  double u, v;
  int *last_row, *this_row, *tmp;

  n = generation;

  last_row = (int *) pfMalloc (sizeof (int) * (1 << 6), NULL);
  this_row = (int *) pfMalloc (sizeof (int) * (1 << 6), NULL);

  for (i = 1; i <= n; i++) {
    next_point[i] = find_top_left(i);
  }

  c = 0;
  last_row[0] = 0;
  for (j = 1; j <= (1 << n); j++){
    for (i = 0; i <= j; i++) {
      for (k = n; k >= 0; k--) {
	if ( !(j % (1 << k)) && !(i % (1 << k))) {
	  k = n - k;
	  u = ((double) i) / ((double) j);
	  v = ((double) j) / ((double) (1 << n));
	  if (v == 1.0 && u == 0.0)
	      this_row[i] = 1;
	  else if (v == 1.0 && u == 1.0)
	      this_row[i] = 2;
	  else {
	    this_row[i] = next_point[k];
	    next_point[k]++;
	  }
	  break;
	}
      }
    }

    for (i = 0; i < j; i++) {
      vindex[c++] = this_row[i];
      vindex[c++] = last_row[i];
    }
    vindex[c++] = this_row[i];

    tmp = this_row;
    this_row = last_row;
    last_row = tmp;
  }
  
  free(this_row);
  free(last_row);
  return;
}

/*This routine sets the normals for the GeoSet;*/
void generate_normals ()
{
  int i;

  for (i = 0; i < 4; i++) {
      set_normals(side[i], norms[i]);
  }

  return;
}

/* Calculates the normals for the geoset */
void set_normals(pfVec3 *sideN, pfVec3 *normsN)
{
  int j;
  pfVec3 vec0, vec1;
  float *v0, *v1, *v2;
  int read;
  int level;
  int length;

  read = 0;

  for (level = 1; level <= (1 << generation); level++){
    length = level * 2 + 1;

    v0 = sideN[vindex[read]];
    v1 = sideN[vindex[read + 1]];
    v2 = sideN[vindex[read + 2]];
    PFSUB_VEC3(vec0, v1, v0);
    PFSUB_VEC3(vec1, v2, v1);
    
    pfCrossVec3(normsN[read], vec0, vec1);
    pfNormalizeVec3(normsN[read]);
    pfCopyVec3(normsN[read+1], normsN[read]);
    pfCopyVec3(normsN[read+2], normsN[read]);

    for(j = read + 3; j < read + length; j++) {
      v0 = sideN[vindex[j]];
      v1 = sideN[vindex[j - 1]];
      v2 = sideN[vindex[j - 2]];
      PFSUB_VEC3(vec0, v1, v0);
      PFSUB_VEC3(vec1, v2, v1);

      pfCrossVec3(normsN[j], vec0, vec1);
      pfNormalizeVec3(normsN[j]);
      j++;

      if(j < read + length) {
	v0 = sideN[vindex[j - 2]];
	v1 = sideN[vindex[j - 1]];
	v2 = sideN[vindex[j]];
	PFSUB_VEC3(vec0, v1, v0);
	PFSUB_VEC3(vec1, v2, v1);
	pfCrossVec3(normsN[j], vec0, vec1);
	pfNormalizeVec3(normsN[j]);
      }
    }
    read += length;
  }
 
  return;
}

/*This routine is the main body of the rock generation routine;*/
pfGeoSet **draw_rock(float width, float depth, float height, int n)
{
  int i;
  struct timeval tp;

  initRock(n);

  /* seed the random number generator */
  gettimeofday(&tp, NULL);
  srand((int)(tp.tv_usec));

  /*  If you want to change the basic appearance of the pyramid drawn
      from that alter some of these ABCDE's below */
  A.x = 0; A.y = 0; A.z = height;
  B.x = -width/2; B.y = depth/2; B.z = 0;
  C.x = width/2; C.y = depth/2; C.z = 0;
  D.x = -width/2; D.y = -depth/2; D.z = 0;
  E.x = width/2; E.y = -depth/2; E.z = 0;

  side1[0] = A; side1[1] = B; side1[2] = C;
  side2[0] = A; side2[1] = D; side2[2] = B;
  side3[0] = A; side3[1] = E; side3[2] = D;
  side4[0] = A; side4[1] = C; side4[2] = E;

  generate_side(side1);

  wrkngon_side2 = TRUE;
  generate_side(side2);
  wrkngon_side2 = FALSE;

  wrkngon_side3 = TRUE;
  generate_side(side3);
  wrkngon_side3 = FALSE;
 
  wrkngon_side4 = TRUE;
  generate_side(side4);
  wrkngon_side4 = FALSE;

  generate_vindex();
  generate_tindex();
  copy_verts();
  generate_normals();
  set_lengths();

  drawSelf();      /* pass in a geoState */

  return gset;
}

/* Copies all vertex from side? to side[][][].  The reason for this
   is that performer wants its vertex in the form of a pfVec3 type */
void copy_verts()
{
  int size;
  int i;

  size = find_top_left(generation + 1);


  for (i = 0; i < size; i++) {
    side[0][i][0] = (float)side1[i].x;
    side[0][i][1] = (float)side1[i].y;
    side[0][i][2] = (float)side1[i].z;
    
    side[1][i][0] = (float)side2[i].x;
    side[1][i][1] = (float)side2[i].y;
    side[1][i][2] = (float)side2[i].z;
    
    side[2][i][0] = (float)side3[i].x;
    side[2][i][1] = (float)side3[i].y;
    side[2][i][2] = (float)side3[i].z;

    side[3][i][0] = (float)side4[i].x;
    side[3][i][1] = (float)side4[i].y;
    side[3][i][2] = (float)side4[i].z;
  }

  return;
}

/* Set lengths of the primitives (tri-strips) to be drawn, i.e. the number
   of vertex in a primitive */
void set_lengths()
{
  int i;

  for (i = 1; i <= (1 << generation); i++)
    lengths[i-1] = (2 * i) + 1;

  return;
}

/* set the color of the pyramid */
void set_color(float r, float g, float b, float t)
{
  int i;
  
  scolor[0][0] = r;
  scolor[0][1] = g;
  scolor[0][2] = b;
  scolor[0][3] = t;

  scindex[0] = 0;
  return;
}

/* generates the texture indexes and coordinates */
void generate_tindex() 
{
  int i, j;
  int c = 0;
  int vert = 0;

  for (i = 1; i <= (1 << generation); i++) {
    for (j = 0; j < (i * 2 + 1); j++) {
      tindex[c++] = vert++;
      if (vert > 2) vert = 0;
    }
  }

  tcoords[0][0] = 0.0; tcoords[0][1] = 0.0;
  tcoords[1][0] = 0.5; tcoords[1][1] = 1.0;
  tcoords[2][0] = 1.0; tcoords[2][1] = 0.0;

  return;
}

/* This small routine just puts everything into their geoSets */
void drawSelf()
{
  int i;

  i = 0;

  for (i = 0; i < 4; i ++) {
    pfGSetAttr(gset[i], PFGS_COORD3, PFGS_PER_VERTEX, side[i], vindex);
    pfGSetAttr(gset[i], PFGS_NORMAL3, PFGS_PER_VERTEX, norms[i], nindex);
    pfGSetAttr(gset[i], PFGS_COLOR4, PFGS_OVERALL, scolor, scindex);
    pfGSetAttr(gset[i], PFGS_TEXCOORD2, PFGS_PER_VERTEX, tcoords, tindex);
    pfGSetPrimType(gset[i], PFGS_TRISTRIPS);
    pfGSetNumPrims(gset[i], (1 << generation));
    pfGSetPrimLengths(gset[i], lengths);
  }

  return;
}

/* free things at the end */
void freeStuff()
{
  /* free whatever you malloc'ed in initFrame... here */
  free(side1); free(side2); free(side3); free(side4);
  free(top); free(bot);
  pfFree(vindex);
  pfFree(norms);
  pfFree(side);
  pfFree(lengths);

  return;
}

Object P_MakeRock(Object width,
		  Object depth,
		  Object height,
		  Object fractal_depth)
{
  pfGeode *geode;
  pfGeoSet **sides;
  Check_Type(width, T_Flonum);
  Check_Type(depth, T_Flonum);
  Check_Type(height, T_Flonum);
  Check_Type(fractal_depth, T_Fixnum);

  sides = draw_rock(FLONUM(width)->val,
		    FLONUM(depth)->val,
		    FLONUM(height)->val,
		    FIXNUM(fractal_depth));
  geode = pfNewGeode(); 
  pfAddGSet(geode, sides[0]);
  pfAddGSet(geode, sides[1]);
  pfAddGSet(geode, sides[2]);
  pfAddGSet(geode, sides[3]);
  return Make_Pfo((pfObject*)geode);
}

Object
P_RockScaleTexture(Object geode, Object scale)
{
  pfVec2 *tcoords;
  ushort *tindex;
  float scal;
  pfGeode *g;
  pfGeoSet *geoset;
  void *arena;
  float coord1;
  float coord2;

  Check_Pfo_Class(geode, PFTYPE_GEODE);
  Check_Type(scale, T_Flonum);

  g = (pfGeode*) PFO(geode);
  geoset = pfGetGSet(g, 0);
  scal = 1.0f / FLONUM(scale)->val;

  pfGetGSetAttrLists(geoset, PFGS_TEXCOORD2, (void**)&tcoords, &tindex);

  coord1 = 0.5;
  coord2 = 1.0;

#if 0
  {
    int i;
    for (i = 1; i < 8; i++)
      fprintf(stderr, "tcoords[%d][0] = %f tcoords[%d][1] = %f\n",
	      i, tcoords[i][0], i, tcoords[i][1]);
  }
#endif

  tcoords[0][0] = 0.0; tcoords[0][1] = 0.0;
  tcoords[1][0] = coord1 * scal; tcoords[1][1] = coord2 * scal;
  tcoords[2][0] = coord2 * scal; tcoords[2][1] = 0.0;

#if 0
  {
    int i;
    for (i = 1; i < 8; i++)
      fprintf(stderr, "tcoords[%d][0] = %f tcoords[%d][1] = %f\n",
	      i, tcoords[i][0], i, tcoords[i][1]);
  }
#endif

  return geode;
}

void rock_init ()
{
  Define_Primitive(P_MakeRock, "ps:make-geode-rock1", 4, 4, EVAL); 
  Define_Primitive(P_RockScaleTexture, "ps:rock-scale-texture", 2, 2, EVAL); 
}
