/* FRACTAL ROCKS
   algorithm by:  Andy Forsberg
   code by:  Chandler Chao

   This is a fractal rock primitive for the IRIS Performer drawing program.

   Basic Concept:  Begin with a pyramid.  Each side of the pyramid is a 
   triangle.  To create the fractal rock, each triangular side is bisected
   in half, then bisected in half, and so on, up until the number of
   generations.  These points are then randomly adjusted on the x,y,z
   axis to create the fractal effect.

   Using this Routine:
   Fractal rocks are called witht he following function call.
           draw_rock(width, depth, height, red,
	              green, blue, transparency, generation, geostate);
   
   It will return 4 geostates in an array.  Each geostate should be
   bound to a geode.

   Maintenance:  There is not much room for maintenance in this program.
   Things you may want to adjust are "potential_height", which determines
   the jaggedness of the rock.  Adjusting it is explained herein.  You
   may also want to adjust the concavity bias which is found in the
   routine "rnd".  Other than that you shouldn't need to change
   anything at all, unless you wanted to put more sides in the pyramid
   making it more conical, but that's rediculous.  If you really wanted
   to do that it would take some fairly major alterations in the code.

*/
   
#include <stdlib.h>
#include <sys/time.h>
#include <Performer/pf.h>
#include "pfsgi.h"
#include <math.h>

#define TRUE 1
#define FALSE 0

typedef struct point {
    double x, y, z;
} pnt_t;

static int generation;
static pnt_t A, B, C, D, E;
static pnt_t *side1, *side2, *side3, *side4;
static int next_point[20];
static int *top, *bot, *tmp_row_pointer;
static int wrkngon_side2, wrkngon_side3, wrkngon_side4;
static char *textureName;

pfVec3 **side;
pfVec3 **norms;
pfVec4 *scolor;
pfVec2 *tcoords;
ushort *vindex;
ushort *nindex;
ushort *tindex;
long *lengths;
ushort *scindex;
void *arena;
pfGeoSet *gset[4];

int rnd( void );
void adjust_xyz( double potential_height, pnt_t *pt );
void adjust_xy( double potential_height, pnt_t *pt );
int find_top_left( int n );
double interpolate( double a, double b, double u);
void intp_point( pnt_t p1, pnt_t p2, pnt_t *p, double u);
int skip ( int n );
void get_row_indices( int n, int j, int *indices);
void copy_point (pnt_t *to, pnt_t *from);
void initRock(int n);
void generate_side(pnt_t *plist);
void generate_vindex();
void generate_normals ();
void set_normals(pfVec3 *sideN, pfVec3 *normsN);
void generate_tindex();
pfGeoSet **draw_rock(float width, float depth, float height,
		     int n);

void copy_verts();
void set_lengths();
void set_color(float r, float g, float b, float t);
void drawSelf();
void freeStuff();
