#include "persia-scm.h"
#include "sphere.h"

pfGeoSet **draw_sphere(float rad, int n)
{
  int i;
  pfVec3 A, B, C, D, E, F;
  initSphere(n);

  radius = rad;

  A[X] = radius; A[Y] = radius; A[Z] = 2*radius;
  B[X] = 0; B[Y] = radius; B[Z] = radius;
  C[X] = radius; C[Y] = 2*radius; C[Z] = radius;
  D[X] = radius; D[Y] = 0; D[Z] = radius;
  E[X] = 2*radius; E[Y] = radius; E[Z] = radius;
  F[X] = radius; F[Y] = radius; F[Z] = 0;

  pfCopyVec3(side[0][0],A); pfCopyVec3(side[0][1],B); pfCopyVec3(side[0][2],C);
  pfCopyVec3(side[1][0],A); pfCopyVec3(side[1][1],D); pfCopyVec3(side[1][2],B);
  pfCopyVec3(side[2][0],A); pfCopyVec3(side[2][1],E); pfCopyVec3(side[2][2],D);
  pfCopyVec3(side[3][0],A); pfCopyVec3(side[3][1],C); pfCopyVec3(side[3][2],E);
  pfCopyVec3(side[4][0],F); pfCopyVec3(side[4][1],C); pfCopyVec3(side[4][2],B);
  pfCopyVec3(side[5][0],F); pfCopyVec3(side[5][1],B); pfCopyVec3(side[5][2],D);
  pfCopyVec3(side[6][0],F); pfCopyVec3(side[6][1],D); pfCopyVec3(side[6][2],E);
  pfCopyVec3(side[7][0],F); pfCopyVec3(side[7][1],E); pfCopyVec3(side[7][2],C);

  for (i = 0; i < 8; i++)
    generate_side(side[i]);

  generate_vindex();
  generate_tindex();
  generate_normals();
  set_lengths();
  set_color();
  
  drawSelf();

  return sphereGset;
}

void initSphere(int n)
{
  int count;   /* Number of vertices in a side when drawing */
  int i;
  int size;    /* Number of vertices in a side when not drawing */

  generation = n;
  count = 0;

  size = find_top_left(n + 1);

  arena = pfGetSharedArena();

  /* Initialized from NULL because they don't need to be shared
     in performer */
  top = (int *) pfMalloc (sizeof (int) * (1 << n), NULL);
  bot = (int *) pfMalloc (sizeof (int) * (1 << n), NULL);

  /* Initialized with arena because they need to be shared in performer */
  for (i = 1; i <= (1 << n); i++) count += i * 2 + 1;
  vindex = (ushort*) pfMalloc (sizeof(ushort) * count, arena);
  nindex = (ushort*) pfMalloc (sizeof(ushort) * count, arena);

  for (i = 0; i < count; i++) nindex[i] = i;

  norms = (pfVec3 **) pfMalloc (sizeof(pfVec3) * 8, arena);
  for (i = 0; i < 8; i++)
    norms[i] = (pfVec3*) pfMalloc (sizeof(pfVec3) * count, arena);

  side = (pfVec3 **)pfMalloc(sizeof(pfVec3) * 8, arena);
  for (i = 0; i < 8; i++)
    side[i] = (pfVec3*) pfMalloc (sizeof(pfVec3) * size, arena);

  lengths = (long *)pfMalloc (sizeof(long) * 1 << generation, arena);

  scolor = (pfVec4 *) pfMalloc (sizeof(pfVec4) * 1, arena);
  scindex = (ushort *) pfMalloc (sizeof(ushort) * 1, arena);

  tcoords = (pfVec2 *) pfMalloc (sizeof(pfVec2) * 3, arena);
  tindex = (ushort *) pfMalloc (sizeof(ushort) * count, arena);

  for (i = 0; i < 8; i++)
    sphereGset[i] = pfNewGSet(arena);

  return;
}

/* This is the routine that is responsible for generating all the points
   within a side */
void generate_side(pfVec3 *plist)
{
  int n = generation;
  int x, y, k;
  int next;

  for (k = 0; k < n; k++) {
    bot[0] = 0;
    next = find_top_left(k + 1);
    for (y = 1; y <= (1 << k); y++) {
      tmp_row_pointer = top;
      top = bot;
      bot = tmp_row_pointer;
      get_row_indices (k, y, bot);
      
      for (x = 0; x < y; x++) {
	intp_point (plist[top[x]], plist[bot[x+0]], \
		    plist[next + 2 * x + 0], 0.5);
	adjust_xyz(plist[next + 2 * x + 0]);

	intp_point (plist[top[x]], plist[bot[x + 1]], \
		    plist[next + 2 * x + 1], 0.5);
	adjust_xyz(plist[next + 2 * x + 1]);

	intp_point (plist[bot[x + 0]], plist[bot[x + 1]], \
		    plist[next + 2 * y + x], 0.5);
	adjust_xyz(plist[next + 2 * y + x]);
      }

      next += 3 * y;
    }
  }
  return;
}

float interpolate( float a, float b, float u)
{
  return ((1 - u) * a + u * b);
}

void intp_point( pfVec3 p1, pfVec3 p2, pfVec3 p, float u)
{
  p[X] = interpolate(p1[X], p2[X], u);
  p[Y] = interpolate(p1[Y], p2[Y], u);
  p[Z] = interpolate(p1[Z], p2[Z], u);

  return;
}

/* Normalize vector, then scale it by the radius to find the corresponding
   point on the sphere */
void adjust_xyz(pfVec3 pt)
{
  pfVec3 center, vec;
  
  center[X] = radius;
  center[Y] = radius;
  center[Z] = radius;

  pfSubVec3(vec, pt, center);

  pfNormalizeVec3(vec);
  pfScaleVec3(vec, radius, vec);
  pfAddVec3(pt, vec, center);

  return;
}

/*this routine sets up the vindex for the GeoSet which is done by strip
note that this vindex is uniform no matter which side you are on*/
void generate_vindex()
{
  int i, j, k, c;
  int n;
  double u, v;
  int *last_row, *this_row, *tmp;

  n = generation;

  last_row = (int *) pfMalloc (sizeof (int) * (1 << 6), NULL);
  this_row = (int *) pfMalloc (sizeof (int) * (1 << 6), NULL);

  for (i = 1; i <= n; i++) {
    next_point[i] = find_top_left(i);
  }

  c = 0;
  last_row[0] = 0;
  for (j = 1; j <= (1 << n); j++){
    for (i = 0; i <= j; i++) {
      for (k = n; k >= 0; k--) {
	if ( !(j % (1 << k)) && !(i % (1 << k))) {
	  k = n - k;
	  u = ((double) i) / ((double) j);
	  v = ((double) j) / ((double) (1 << n));
	  if (v == 1.0 && u == 0.0)
	      this_row[i] = 1;
	  else if (v == 1.0 && u == 1.0)
	      this_row[i] = 2;
	  else {
	    this_row[i] = next_point[k];
	    next_point[k]++;
	  }
	  break;
	}
      }
    }

    for (i = 0; i < j; i++) {
      vindex[c++] = this_row[i];
      vindex[c++] = last_row[i];
    }
    vindex[c++] = this_row[i];

    tmp = this_row;
    this_row = last_row;
    last_row = tmp;
  }
  
  free(this_row);
  free(last_row);
  return;
}

/* Used to skip points that have already been calculated */
int skip ( int n )
{
  return (((n + 3) / 2) * ((n + 1) / 2) + (((n + 2) / 2) * (n / 2) / 2));
}

/* For each level in a triangle, figures out which points are in that
   row, both top and bottom */
void get_row_indices( int n, int j, int *indices)
{
  int i, k, jj;
  int next_point[10];

  next_point[0] = (j == 0) ? 0 : 1;
  for (i = 1; i <= n; i++) {
    next_point[i] = find_top_left(i);
    jj = (j - 1) / (1 << (n - i));
    next_point[i] += skip(jj);
  }

  for (i = 0; i <= j; i++) {
    for (k = n; k >= 0; k--) {
      if ( !(j % (1 << k)) && !(i % (1 << k))) {
	k = n - k;
	indices[i] = next_point[k];
	next_point[k]++;
	break;
      }
    }
  }
  
  return;
}

/*This routine sets the normals for the GeoSet;*/
void generate_normals ()
{
  int i;

  for (i = 0; i < 8; i++) {
      set_normals(side[i], norms[i]);
  }

  return;
}

/* Calculates the normals for the geoset */
void set_normals(pfVec3 *sideN, pfVec3 *normsN)
{
  int j;
  pfVec3 vec0, vec1;
  float *v0, *v1, *v2;
  int read;
  int level;
  int length;

  read = 0;

  for (level = 1; level <= (1 << generation); level++){
    length = level * 2 + 1;

    v0 = sideN[vindex[read]];
    v1 = sideN[vindex[read + 1]];
    v2 = sideN[vindex[read + 2]];
    PFSUB_VEC3(vec0, v1, v0);
    PFSUB_VEC3(vec1, v2, v1);
    
    pfCrossVec3(normsN[read], vec0, vec1);
    pfNormalizeVec3(normsN[read]);
    pfCopyVec3(normsN[read+1], normsN[read]);
    pfCopyVec3(normsN[read+2], normsN[read]);

    for(j = read + 3; j < read + length; j++) {
      v0 = sideN[vindex[j]];
      v1 = sideN[vindex[j - 1]];
      v2 = sideN[vindex[j - 2]];
      PFSUB_VEC3(vec0, v1, v0);
      PFSUB_VEC3(vec1, v2, v1);

      pfCrossVec3(normsN[j], vec0, vec1);
      pfNormalizeVec3(normsN[j]);
      j++;

      if(j < read + length) {
	v0 = sideN[vindex[j - 2]];
	v1 = sideN[vindex[j - 1]];
	v2 = sideN[vindex[j]];
	PFSUB_VEC3(vec0, v1, v0);
	PFSUB_VEC3(vec1, v2, v1);
	pfCrossVec3(normsN[j], vec0, vec1);
	pfNormalizeVec3(normsN[j]);
      }
    }
    read += length;
  }
 
  return;
}

/* Set lengths of the primitives (tri-strips) to be drawn, i.e. the number
   of vertex in a primitive */
void set_lengths()
{
  int i;

  for (i = 1; i <= (1 << generation); i++)
    lengths[i-1] = (2 * i) + 1;

  return;
}

/* set the color of the pyramid */
void set_color()
{
/* Matt's trial code: */

  scindex[0] = 1;
  return;
}

/* generates the texture indexes and coordinates */
void generate_tindex() 
{
  int i, j;
  int c = 0;
  int vert = 0;

  for (i = 1; i <= (1 << generation); i++) {
    for (j = 0; j < (i * 2 + 1); j++) {
      tindex[c++] = vert++;
      if (vert > 2) vert = 0;
    }
  }

  tcoords[0][0] = 0.0; tcoords[0][1] = 0.0;
  tcoords[1][0] = 0.5; tcoords[1][1] = 1.0;
  tcoords[2][0] = 1.0; tcoords[2][1] = 0.0;

  return;
}

/* This small routine just puts everything into their geoSets */
void drawSelf()
{
  int i;

  i = 0;

  for (i = 0; i < 8; i ++) {
    pfGSetAttr(sphereGset[i], PFGS_COORD3, PFGS_PER_VERTEX, side[i], vindex);
    pfGSetAttr(sphereGset[i], PFGS_NORMAL3, PFGS_PER_VERTEX, norms[i], nindex);
    pfGSetAttr(sphereGset[i], PFGS_COLOR4, PFGS_OVERALL, scolor, scindex);
    pfGSetAttr(sphereGset[i], PFGS_TEXCOORD2, PFGS_PER_VERTEX, tcoords, tindex);
    pfGSetPrimType(sphereGset[i], PFGS_TRISTRIPS);
    pfGSetNumPrims(sphereGset[i], (1 << generation));
    pfGSetPrimLengths(sphereGset[i], lengths);
  }

  return;
}

int find_top_left( int n )
{
  int r;

  if (n == 0) 
    return 0;
  else if (n == 1)
    return 3;
  else
    r = find_top_left(n - 1) + 3 * ((int)((pow(2, n - 2)) + 1.0) * (pow(2, n - 3)));
  
  return r;
}

Object P_MakeSphere(Object radius,
		    Object fractal_depth)
{
  pfGeode *geode;
  pfGeoSet **sides;
  Check_Type(radius, T_Flonum);
  Check_Type(fractal_depth, T_Fixnum);

  sides = draw_sphere(FLONUM(radius)->val,
		      FIXNUM(fractal_depth));
  geode = pfNewGeode(); 
  pfAddGSet(geode, sides[0]);
  pfAddGSet(geode, sides[1]);
  pfAddGSet(geode, sides[2]);
  pfAddGSet(geode, sides[3]);
  pfAddGSet(geode, sides[4]);
  pfAddGSet(geode, sides[5]);
  pfAddGSet(geode, sides[6]);
  pfAddGSet(geode, sides[7]);
  return Make_Pfo((pfObject*)geode);
}

Object
P_SphereScaleTexture(Object geode, Object scale)
{
  pfVec2 *tcoords;
  ushort *tindex;
  float scal;
  pfGeode *g;
  pfGeoSet *geoset;
  void *arena;
  float coord1;
  float coord2;

  Check_Pfo_Class(geode, PFTYPE_GEODE);
  Check_Type(scale, T_Flonum);

  g = (pfGeode*) PFO(geode);
  geoset = pfGetGSet(g, 0);
  scal = 1.0f / FLONUM(scale)->val;

  pfGetGSetAttrLists(geoset, PFGS_TEXCOORD2, (void**)&tcoords, &tindex);

  coord1 = 0.5;
  coord2 = 1.0;

#if 0
  {
    int i;
    for (i = 1; i < 8; i++)
      fprintf(stderr, "tcoords[%d][0] = %f tcoords[%d][1] = %f\n",
	      i, tcoords[i][0], i, tcoords[i][1]);
  }
#endif

  tcoords[0][0] = 0.0; tcoords[0][1] = 0.0;
  tcoords[1][0] = coord1 * scal; tcoords[1][1] = coord2 * scal;
  tcoords[2][0] = coord2 * scal; tcoords[2][1] = 0.0;

#if 0
  {
    int i;
    for (i = 1; i < 8; i++)
      fprintf(stderr, "tcoords[%d][0] = %f tcoords[%d][1] = %f\n",
	      i, tcoords[i][0], i, tcoords[i][1]);
  }
#endif

  return geode;
}

void sphere_init ()
{
  Define_Primitive(P_MakeSphere, "ps:make-geode-sphere1", 2, 2, EVAL);
  Define_Primitive(P_SphereScaleTexture, "ps:sphere-scale-texture", 2, 2, EVAL);
}
