;;; Definitions to be set before loading driving.scm
(define ground-width 800.)
;;(define ground-depth 2000.)
(define ground-depth 1400.)
(define doing-vr #t)
(define doing-vr-head #t)
(define doing-vr-eyes #f)
(define doing-kart #t)
(define doing-second-car #f)
(define doing-vulture #f)
(define car-width 2.5)
(define car-depth 5.)
(define car-height 1.)

; For testing with Darrell
;(set! doing-vr-head #f)
;(set! doing-vr-eyes #t)
;(set! doing-kart #f)
;(set! doing-vulture #t)

(require 'driving)

(define gazelle-free-edge-width 200.)

(define doing-across #f)
(define num-gazelles 20)
(define max-leap-height 3.)
(define leap-heights (vector
		      (vector 0.)
		      (vector 0.)
		      (vector 0. 1.)
		      (vector 0. 2. 1.)
		      (vector 0. 2. 2.5 1.7)
		      (vector 0. 2. 3. 2. 1.)
		      (vector 0. 2. 3. 3. 2. 1.)
		      (vector 0. 2.0 2.5 3.0 2.5 2.0 1.0)
		      (vector 0. 1.9 2.7 3.0 3.0 2.7 1.9 1.0)
		      ))

;;; These speed values should be integers
(define max-speed 2)
(define min-speed 1)

(if doing-across
    (begin
      (define gazelle-width 7.)
      (define gazelle-depth 4.)
      (define gazelle-height 4.))
    (begin
;      (define gazelle-width 4.0)
;      (define gazelle-depth 5.5)
;      (define gazelle-height 4.5)
      (define gazelle-width 1.7)
      (define gazelle-depth 2.2)
      (define gazelle-height 2.1)
      ))

(define gazelle-leg-height 0.7)

(define gazelle-body
  (ps:make-geode-cube 
   gazelle-width gazelle-depth gazelle-height
   (cons 'gstate (ps:make-gstate '(color . #(1. .2 .2))
				 '(texture-file . "bark.rgb")))
   (cons 'texture-scale 100.)))

(define gazelle (pf:new-dcs))
(pf:add-child gazelle gazelle-body)
(ps:place-node (ps:make-geode-cube
		  (* .5 gazelle-width)
		  (* .7 gazelle-depth)
		  (* .5 gazelle-height)
		  (cons 'gstate (ps:make-gstate '(color . #(.3 .0 .0))
						'(texture-file . "bark.rgb"))))
	       (* .25 gazelle-width) (* gazelle-depth .7) (* gazelle-height .8)
	       gazelle)

(define tower-wall-dist 250.)

(define gazelle-list '())

(do ((i 0 (1+ i)))
    ((> i num-gazelles))
  (set! gazelle-list 
	(cons 
	 (list 
	  ;; the dcs
	  (ps:place-node gazelle 
			 (random-float gazelle-free-edge-width
				       (- ground-width 
					  gazelle-free-edge-width))
			 (random-float 0. ground-depth)
			 gazelle-leg-height)
	  ;; the speed
	  (random-int min-speed max-speed)
	  ;; sideways motion
	  (random-float -1. 1.))
	 gazelle-list)))

;;; rotate gazelle bodies to match direction of motion.
(if (not doing-across)
    (map (lambda (gl) 
	   (let ((rot (* (/ (caddr gl) (cadr gl)) -90.)))
	     (begin
	       ;;	   (format #t "rotated ~s\n" rot)
	       (pf:dcs-rot (car gl) rot 0. 0.))))
	 gazelle-list))

(define gazelle-tick-count 0)

(define (move-gazelle-x gaz-list)
  (let ((g (car gaz-list)))
    (begin 
      (pf:get-dcs-mat g ps:tmp-pfmatrix)
      (let ((x (ps:pfmatrix-ref ps:tmp-pfmatrix 3 0))
	    (y (ps:pfmatrix-ref ps:tmp-pfmatrix 3 1))
	    (speed (cadr gaz-list)))
	(if (> x ground-width)
	    (pf:dcs-trans g 0. y 0.)
	    (if (> (modulo gazelle-tick-count speed) 0)
		(pf:dcs-trans g (+ x speed)
			      y
			      (* max-leap-height (/ speed max-speed)))
		(pf:dcs-trans g (+ x speed)
			      y
			      0.)))))))

(define (move-gazelle-y gaz-list)
  (let ((g (car gaz-list)))
    (begin 
      (pf:get-dcs-mat g ps:tmp-pfmatrix)
      (let* ((speed (cadr gaz-list))
	     (tmp-x (+ (ps:pfmatrix-ref ps:tmp-pfmatrix 3 0) (caddr gaz-list)))
	     (x (if (> tmp-x (- ground-width gazelle-free-edge-width))
		    gazelle-free-edge-width 
		    (if (< tmp-x gazelle-free-edge-width)
			(- ground-width gazelle-free-edge-width)
			tmp-x)))
	     (y (ps:pfmatrix-ref ps:tmp-pfmatrix 3 1))
	     (leap-vector-length (* speed 4))
	     (leap-index (modulo gazelle-tick-count leap-vector-length)))
	(if (> y ground-depth)
	    (pf:dcs-trans g x 0. 0.)
	    (pf:dcs-trans g x 
			  (+ y speed) 
			  (+ gazelle-leg-height 
			     (vector-ref (vector-ref leap-heights 
						     leap-vector-length) 
					 leap-index))))))))

(if doing-across
    (define (move-gazelles)
      (set! gazelle-tick-count (1+ gazelle-tick-count))
      (map move-gazelle-x gazelle-list))
    (define (move-gazelles)
      (set! gazelle-tick-count (1+ gazelle-tick-count))
      (map move-gazelle-y gazelle-list)))

(ps:add-pre-frame-hook (lambda () (move-gazelles)))

;;; The towers
(define watertower-node (ps:make-node-from-file "watertower.flt"))
(define tower1
  (pf:dcs-scale (ps:place-node watertower-node
				 (/ ground-width 2) tower-wall-dist 0.)
		5.0))
(define tower2
  (pf:dcs-scale (ps:place-node (ps:make-node-from-file "house2.flt")
				 (/ ground-width 2)
				 (- ground-depth tower-wall-dist)
				 0.)
		5.0))

(define tree1-node (ps:make-node-from-file "tree1.flt"))
(define tree2-node (ps:make-node-from-file "tree2.flt"))
(define tree3-node (ps:make-node-from-file "tree3.flt"))

(define tree-scale 4.)
;;; at the four corners
(pf:dcs-scale (ps:place-node tree1-node 50. 100. 0.) tree-scale)
(pf:dcs-scale (ps:place-node tree1-node 730. 1850. 0.) tree-scale)
(pf:dcs-scale (ps:place-node tree2-node 40. 1300. 0.) tree-scale)
(pf:dcs-scale (ps:place-node tree3-node 750. 90. 0.) tree-scale)

;;; some towards the middle
(pf:dcs-scale (ps:place-node tree1-node 250. 1200. 0.) tree-scale)
(pf:dcs-scale (ps:place-node tree2-node 600. 900. 0.) tree-scale)
(pf:dcs-scale (ps:place-node tree1-node 300. 500. 0.) tree-scale)
;;; a tree dead-center in the field
(pf:dcs-scale (ps:place-node tree3-node 
			     (* ground-width .5) 
			     (* ground-depth .5) 
			     0.)
	      tree-scale)
;;; a tree behind the house
(pf:dcs-scale (ps:place-node tree2-node 
			     (/ ground-width 2)
			     (- ground-depth (/ tower-wall-dist 2))
			     0.)
	      tree-scale)
;;; a tree behind the watertower
(pf:dcs-scale (ps:place-node tree1-node 
			     (/ ground-width 2)
			     (* tower-wall-dist .8)
			     0.)
	      tree-scale)

;;; some rocks
(ps:place-node
 (ps:make-geode-rock 200. 300. 20. 3
		     (cons 'gstate (ps:make-gstate
				    '(color . #(.5 .2 .2 .5))
				    '(texture-file . "grass2.rgb"))))
 0. (* ground-depth 0.8)  0.)
(ps:place-node
 (ps:make-geode-rock 200. 250. 20. 3
		     (cons 'gstate (ps:make-gstate
				    '(color . #(.5 .3 .3 .5))
				    '(texture-file . "grass2.rgb"))))
 ground-width 500. 0.)

;;; watertower scale change
(define tower-is-double-scale #f)
(define (watch-mouse-for-watertower-scaling) 
   (let ((mouse-is-left (< (ps:mouse-x) (- (ps:mouse-x-max) 20))))
     (if (and tower-is-double-scale mouse-is-left)
	 (begin
	   (pf:dcs-scale tower1 5.0)
	   (set! tower-is-double-scale #f)))
     (if (and (not tower-is-double-scale) (not mouse-is-left))
	 (begin
	   (pf:dcs-scale tower1 10.0)
	   (set! tower-is-double-scale #t)))))
;; Temporarily commented out for speed.
;;(ps:add-pre-frame-hook watch-mouse-for-watertower-scaling)

;;; arches
;(ps:place-node (make-dcs-arch 20. 2. 2. 20. 1. .1 .1 1.)
;		 450. 700. 0.)
;(ps:place-node (make-dcs-arch 20. 2. 2. 20. .1 .1 1. 1.)
;		 350. 1000. 0.)
;(ps:place-node (make-dcs-arch 20. 2. 2. 20. 1. .1 .1 1.)
;		 450. 1300. 0.)
;(ps:place-node (make-dcs-arch 20. 2. 2. 20. .1 .1 1. 1.)
;		 350. 1600. 0.)

(ps:run)
