/* tree.c - Scheme interface to Performer scene tree manipulation */

#include "persia-scm.h"
#include "persia.h"

Object P_AddChild(Object x, Object y)
{
  Check_Pfo(x);
  Check_Pfo(y);
  pfAddChild(PFO(x), PFO(y));
  return Void;
}

Object P_RemoveChild(Object x, Object y)
{
  Check_Pfo(x);
  Check_Pfo(y);
  if (pfRemoveChild(PFO(x), PFO(y)))
    return True;
  else
    return False;
}

Object P_GetChild(Object group, Object index)
{
  Check_Pfo_Class(group, PFTYPE_GROUP);
  Check_Type(index, T_Fixnum);
  return Make_Pfo((pfObject*)pfGetChild(PFO(group), FIXNUM(index)));
}

Object P_GetNumChildren(Object group)
{
  Check_Pfo_Class(group, PFTYPE_GROUP);  
  return Make_Fixnum(pfGetNumChildren(PFO(group)));
}

Object P_GetNumParents (Object node)
{
  Check_Pfo_Class(node, PFCLASS_NODE);  
  return Make_Fixnum(pfGetNumParents(PFO(node)));
}

Object P_GetParent (Object node, Object index)
{
  Check_Pfo_Class(node, PFCLASS_NODE);  
  Check_Type(index, T_Fixnum);
  return Make_Pfo((pfObject*) pfGetParent(PFO(node), FIXNUM(index)));
}

void tree_init ()
{
  Define_Primitive(P_AddChild, "pf:add-child", 2, 2, EVAL);
  Define_Primitive(P_RemoveChild, "pf:remove-child", 2, 2, EVAL);
  Define_Primitive(P_GetChild, "pf:get-child", 2, 2, EVAL);
  Define_Primitive(P_GetNumChildren, "pf:get-num-child", 1, 1, EVAL);
  Define_Primitive(P_GetNumParents, "pf:get-num-parents", 1, 1, EVAL);
  Define_Primitive(P_GetParent, "pf:get-parent", 2, 2, EVAL);
}
