#include "persia-scm.h"

#define PI 3.14159265359
#define DEG2RAD(a) ((a) / (180.0 / PI))
#define RAD2DEG(a) ((a) * (180.0 / PI))
#define ABS(x) (((x) > 0) ? (x) : -(x))

Object P_UpdateVehicle(Object vehicle_vector_o,
		       Object steering_o, Object gas_o, Object brake_o,
		       Object delta_seconds_o)
{
  float steering, gas, brake, delta_seconds;
  float velocity, mass, engine, brakes, dfriction, wheelbase;
  pfDCS *car_dcs;
  pfMatrix car_matrix, car_delta_matrix;
  pfCoord car_coord;
  pfCoord car_delta_coord;
  float heading, kinetic_energy, arc_angle, distance, 
    displacement_angle, width;

  Check_Type(steering_o, T_Flonum);
  Check_Type(gas_o, T_Flonum);
  Check_Type(brake_o, T_Flonum);
  Check_Type(delta_seconds_o, T_Flonum);
  Check_Type(vehicle_vector_o, T_Vector);
  
  steering = FLONUM(steering_o)->val;
  gas = FLONUM(gas_o)->val;
  brake = FLONUM(brake_o)->val;
  delta_seconds = FLONUM(delta_seconds_o)->val;

  /* printf("steering %f gas %f brake %f\n", steering, gas, brake); */

  car_dcs = PFD(P_Vector_Ref(vehicle_vector_o, V_INTS(0)));
  pfGetDCSMat(car_dcs, car_matrix);
  pfGetOrthoMatCoord(car_matrix, &car_coord);
  heading = car_coord.hpr[0];
  velocity = FLONUM(P_Vector_Ref(vehicle_vector_o, V_INTS(1)))->val;
  mass = FLONUM(P_Vector_Ref(vehicle_vector_o, V_INTS(2)))->val;
  engine = FLONUM(P_Vector_Ref(vehicle_vector_o, V_INTS(3)))->val;
  brakes = FLONUM(P_Vector_Ref(vehicle_vector_o, V_INTS(4)))->val;
  dfriction = FLONUM(P_Vector_Ref(vehicle_vector_o, V_INTS(5)))->val;
  wheelbase = FLONUM(P_Vector_Ref(vehicle_vector_o, V_INTS(6)))->val;
  width =  FLONUM(P_Vector_Ref(vehicle_vector_o, V_INTS(7)))->val;

  kinetic_energy = velocity * mass;
  kinetic_energy += (gas * engine) - (velocity * dfriction) - (brake * brakes);
  if (kinetic_energy < 0) 
    kinetic_energy = 0;
  velocity = kinetic_energy / mass;

  if (steering != 0)
    {
      float radius = wheelbase / sinf(steering);
      arc_angle = (velocity * delta_seconds) / radius;
      distance = 2 * radius * sinf(arc_angle/2);
    }
  else
    {
      arc_angle = 0;
      distance = (velocity * delta_seconds);
    }

  /* xxx Collision detection should go here */

  car_delta_coord.xyz[0] = 0.0;
  car_delta_coord.xyz[1] = distance;
  car_delta_coord.xyz[2] = 0.0;
  car_delta_coord.hpr[0] = RAD2DEG(arc_angle/2);
  car_delta_coord.hpr[1] = 0.0;
  car_delta_coord.hpr[2] = 0.0;
  pfMakeCoordMat(car_delta_matrix, &car_delta_coord);
  pfPreMultMat(car_matrix, car_delta_matrix);
  pfDCSMat(car_dcs, car_matrix);

  /* Update vector velocity */
  P_Vector_Set(vehicle_vector_o, V_INTS(1), Make_Flonum(velocity));

  return Void;
}

void
update_vehicle_init()
{
  /* printf("running update_vehicle_init()\n"); */
  Define_Primitive(P_UpdateVehicle, "ps:update-vehicle-vector", 5, 5, EVAL);
}
